/*
 *  audio.c
 *
 *  Copyright (C) Gerhard Monzel - October 2003
 *
 *  This file is part of lve, a linux mpeg video editor
 *      
 *  lve tools are free software; you can redistribute it and/or modify
 *  it under the terms of the GNU General Public License as published by
 *  the Free Software Foundation; either version 2, or (at your option)
 *  any later version.
 *   
 *  lve is distributed in the hope that it will be useful,
 *  but WITHOUT ANY WARRANTY; without even the implied warranty of
 *  MERCHANTABILITY or FITNESS FOR A PARTICULAR PURPOSE.  See the
 *  GNU General Public License for more details.
 *   
 *  You should have received a copy of the GNU General Public License
 *  along with GNU Make; see the file COPYING.  If not, write to
 *  the Free Software Foundation, 675 Mass Ave, Cambridge, MA 02139, USA. 
 *
 */
//============================================================================
//==
//============================================================================

#include <malloc.h>
#include "audio.h"

static SDL_sem     *au_sem = NULL;
/*
static SDL_cond    *au_cond  = NULL;
static SDL_mutex   *au_mutex = NULL;
static int         au_count  = 0;
*/

static int      cb_audio_on    = 0;
static double   cb_mult        = 1.0;
static uint32_t *cb_sample_buf = NULL;



//-- audio callback routine to fill play buffer --
//------------------------------------------------ 
static void a_callback(void *userdata, Uint8 *stream, int len)
{
  static double samples_cnt[3];
  
  T_AUDIO_CTX *audio_ctx = (T_AUDIO_CTX *)userdata;
  
  uint32_t *dst   = (uint32_t *)stream;
  int n           = 0;
  
  if (SDL_SemValue(au_sem) > 0) return;
  
  //-- silence, if no data available --
  if (!cb_audio_on)
  {
    memset(stream, 0, MIN_OUT_SAMPLES * 4);
  }
  else if (audio_ctx->a_decoder)
  {
    samples_cnt[0] = MIN_OUT_SAMPLES;

    //-- read audio data for 1 video-frame starting at position of last --
    //-- decoded video frame (will be detected inside readSamples())    -- 
    n = audio_ctx->a_decoder->readSamples(cb_sample_buf, samples_cnt);
  }

  //-- need simple sample-rate conversion, because SDL supports --
  //-- only output of 2^n bytes to the audio device :-(.        --
  if (n>0)
  {
    for(int i=0; i<MIN_OUT_SAMPLES; i++)
      dst[i] = cb_sample_buf[(int)(i*cb_mult)];
  }   

  //-- unlock main thread, which's waiting audio out is armed --
  SDL_SemPost(au_sem);
}


//-- Constructor creates base resoures to work with SDL-Audio --
//--------------------------------------------------------------
T_AUDIO_OUT::T_AUDIO_OUT()
{
  audio_open          = 0;
  audio_ctx.a_decoder = NULL;
  
  cb_sample_buf = (uint32_t *)memalign(32, (MAX_SAMPLE_BUF+1)*sizeof(int16_t)*2);
  audio_spec    = (SDL_AudioSpec *)memalign(32, sizeof(SDL_AudioSpec) );

  if (cb_sample_buf) memset(cb_sample_buf, 0, MAX_SAMPLES_PVF*4);

  //SDL_InitSubSystem(SDL_INIT_AUDIO);
  
  char dn[16] = "unknown";
  //SDL_AudioInit("artsc");
  //SDL_AudioInit("alsa");
  SDL_AudioInit("dsp");
  if (!SDL_AudioDriverName(dn, 12)) 
  {
    SDL_AudioInit("alsa");   // force use of OSS
    SDL_AudioDriverName(dn, 12);
  }  
  fprintf(stderr, "audio driver is (\"%s\")\n", dn);

}

//-- Destructor frees resources --
//--------------------------------
T_AUDIO_OUT::~T_AUDIO_OUT()
{
  close();
  cleanup();
}
  
//-- helper to cleanup resources --
//---------------------------------  
void T_AUDIO_OUT::cleanup()  
{
  if (audio_spec) free(audio_spec);  
  if (cb_sample_buf) free(cb_sample_buf);  
  
  cb_sample_buf = NULL;
  audio_spec    = NULL;

  audio_open  = 0;
  cb_audio_on = 0;
}

//-- close SDL-audio device --
//----------------------------
void T_AUDIO_OUT::close(void)
{
  if (au_sem)
  {
    SDL_DestroySemaphore(au_sem);
    au_sem = NULL;
  }
  
  if (audio_open)
  {
    SDL_PauseAudio(1); 
    SDL_LockAudio();
    SDL_CloseAudio();
    SDL_Delay(100);
  }
  audio_open = 0;
}    

//-- open SDL-audio device (silent mode) --
//-----------------------------------------
int T_AUDIO_OUT::open(void)
{
  SDL_AudioSpec req_audio_spec;
  SDL_AudioSpec *desired = &req_audio_spec;
  
  //-- silence-mode --
  cb_audio_on = 0;

  //-- create semaphore to control --
  //-- video frame timing in lve.c --
  if (!au_sem)
  {
    if ( (au_sem = SDL_CreateSemaphore(0)) == NULL)
    {
      fprintf(stderr, "T_AUDIO_OUT - cannot trigger audio semapahore\n");   
      close();
      return 0;
    }
  }  

  if (cb_sample_buf && !audio_open)
  {
    //-- using same parameters for all audio formats. --
    //-- audio decoders must deliver always 16 Bit    --
    //-- stereo with 48 KHz or 44.1 KHz. samples will --
    //-- be adjusted before writing to audio-device ! --  
    desired->channels = 2;
    desired->freq     = MAX_SAMPLE_RATE;
    desired->format   = AUDIO_S16SYS;   
    desired->samples  = MIN_OUT_SAMPLES;
    desired->size     = MIN_OUT_SAMPLES * 4;
    desired->callback = a_callback;     
    desired->userdata = &audio_ctx;

    fprintf(stderr, "T_AUDIO_OUT - buffer / samples wanted (%d / %d)\n", desired->size, desired->samples);

    if ( SDL_OpenAudio(desired, audio_spec) < 0 )
    {
      fprintf(stderr, "T_AUDIO_OUT - couldn't open audio: %s\n", SDL_GetError());
      return 0;
    }
    
    fprintf(stderr, "T_AUDIO_OUT - buffer size (%d), out samples per frame (%d)\n", audio_spec->size, audio_spec->samples);

    if (audio_spec->size != (4 * MIN_OUT_SAMPLES) )
    {
      fprintf(stderr, "T_AUDIO_OUT - WARNING: cannot obtain buffer size wanted!\n"); 
      fprintf(stderr, "              sometimes \"artsd\" will cause this trouble, please reduce\n");
      fprintf(stderr, "              idle timeout to 5 seconds or deactivate artsd\n");
    }
    
    audio_open = 1;  
    return 1;
  }
  else if (audio_open)
  {
    fprintf(stderr, "T_AUDIO_OUT - audio device already open!\n");
    return 1;
  }  
  else 
    fprintf(stderr, "T_AUDIO_OUT - no more memory\n");   

  return 0;
}

//-- setup SDL-audio device for (new) decoder/audio-format   --
//-- transfer audio sample-rate (allowed 48 or 44.1 KHz) and --
//-- frame rate of video stream (min 20 fps / max. 30 fps) ! --   
//-------------------------------------------------------------
void T_AUDIO_OUT::setup(T_MPEG_DEC *decoder)
{
  //-- switch to silence -- 
  cb_audio_on = 0;

  audio_ctx.a_decoder = decoder->getAudioDecoder(); 
  audio_ctx.align     = audio_spec->size / (MIN_OUT_SAMPLES * 4);

  //-- calc resample factor --
  cb_mult = audio_ctx.a_decoder->getSamplesPVF()/(double)MIN_OUT_SAMPLES;

  //-- now activate sound --
  cb_audio_on = 1;
}

//-- switch to silent-mode --
//---------------------------
void T_AUDIO_OUT::switchOff(void)
{
  cb_audio_on = 0;
}

//-- test wether audio-device is available --
//-------------------------------------------
int T_AUDIO_OUT::isOpen(void)
{
  return audio_open;
}

void T_AUDIO_OUT::waitBuffer(void)
{
  //SDL_mutexP(au_mutex);
  //au_count--;
  //SDL_CondWait(au_cond, au_mutex);
  SDL_SemWait(au_sem);
}

// eof
