/*
 *  decoder.c
 *
 *  Copyright (C) Gerhard Monzel - October 2003
 *
 *  This file is part of lve, a linux mpeg video editor
 *      
 *  lve tools are free software; you can redistribute it and/or modify
 *  it under the terms of the GNU General Public License as published by
 *  the Free Software Foundation; either version 2, or (at your option)
 *  any later version.
 *   
 *  lve is distributed in the hope that it will be useful,
 *  but WITHOUT ANY WARRANTY; without even the implied warranty of
 *  MERCHANTABILITY or FITNESS FOR A PARTICULAR PURPOSE.  See the
 *  GNU General Public License for more details.
 *   
 *  You should have received a copy of the GNU General Public License
 *  along with GNU Make; see the file COPYING.  If not, write to
 *  the Free Software Foundation, 675 Mass Ave, Cambridge, MA 02139, USA. 
 *
 */
#include "decoder.h"

//============================================================================
//==                             V I D E O                                  ==
//============================================================================

//#define AUTO_DETECT 1

#define MPA_MAX_FRAME_SIZE (MPA_MAX_CODED_FRAME_SIZE)

#define VO_MAX_WIDTH       720
#define VO_MAX_HEIGHT      576

#define VO_MAX_SIZE_LUMA   (VO_MAX_WIDTH*VO_MAX_HEIGHT)
#define VO_MAX_SIZE_CHROMA (VO_MAX_SIZE_LUMA>>2)

static uint8_t *fb0[3] = {NULL, NULL, NULL};
static uint8_t *fb1[3] = {NULL, NULL, NULL};
static uint8_t *fb2[3] = {NULL, NULL, NULL};
static int     ids[3] = {0, 1, 2};

#define SAMPLES_PVF    samples_cnt[0]
#define SAMPLES_NEEDED samples_cnt[1]
#define SAMPLES_USED   samples_cnt[2]

//======================
//== class T_MPEG_DEC ==
//======================
uint8_t *T_MPEG_DEC::vo_pages[3] = {NULL, NULL, NULL};
int T_MPEG_DEC::decode(uint8_t * current, uint8_t * end)
{
  const mpeg2_info_t *info;
  int                rc, state;

  mpeg2_buffer(decoder, current, end);
  info = mpeg2_info (decoder);
  rc   = 0;
  
  while (1) 
  {
    state = mpeg2_parse (decoder);
    switch (state) 
    {
	    
	case STATE_SEQUENCE:
	  if ( !fb0[0] )
	  {
	    fb0[0] = (uint8_t *)mpeg2_malloc(6*VO_MAX_SIZE_CHROMA, ALLOC_YUV);
	    fb0[1] = fb0[0] + VO_MAX_SIZE_LUMA;
	    fb0[2] = fb0[1] + VO_MAX_SIZE_CHROMA;
	    mpeg2_set_buf (decoder, fb0, &ids[0]);
	    
	    fb1[0] = (uint8_t *)mpeg2_malloc(6*VO_MAX_SIZE_CHROMA, ALLOC_YUV);
	    fb1[1] = fb1[0] + VO_MAX_SIZE_LUMA;
	    fb1[2] = fb1[1] + VO_MAX_SIZE_CHROMA;
	    mpeg2_set_buf (decoder, fb1, &ids[0]);
	    
	    fb2[0] = (uint8_t *)mpeg2_malloc(6*VO_MAX_SIZE_CHROMA, ALLOC_YUV);
	    fb2[1] = fb2[0] + VO_MAX_SIZE_LUMA;
	    fb2[2] = fb2[1] + VO_MAX_SIZE_CHROMA;
	    mpeg2_set_buf (decoder, fb2, &ids[0]);
	    
	    if ( !fb0[0] || !fb1[0] || !fb2[0] )
	    {
	      fprintf (stderr,"ERROR: T_MPEG_DEC::decode - Can not create fbufs\n");
	      return 1;
	    }
	  }
	  break;
	
	case STATE_PICTURE:
	  //fprintf(stderr, "P1\n");
	  break;
	
	case STATE_PICTURE_2ND:
	  //fprintf(stderr, "P2\n");
	  break;
	
	case STATE_SLICE:
	case STATE_END:
	case STATE_INVALID_END:
	  if (!vo_visible) break;
	  if (info->display_fbuf) 
	  {
	    if ( (fb0[0] != NULL) && (vo_pages[0] != NULL) )
	    {
	      memcpy(vo_pages[0], info->display_fbuf->buf[0], vo_size_luma);
	      memcpy(vo_pages[1], info->display_fbuf->buf[1], vo_size_chroma);
	      memcpy(vo_pages[2], info->display_fbuf->buf[2], vo_size_chroma);
	    }
	  }
	  rc = 1;
	  break;
	  
	case STATE_BUFFER:  
	case -1:
	  return rc;
	  break;

    }
  }
  return 0;
}    

int T_MPEG_DEC::ff_init_done = 0;

T_MPEG_DEC::T_MPEG_DEC()
{
  decoder        = NULL;
  act_frameSrv   = NULL;
  audio_dec      = NULL;
  mSeqHdrLen     = 0; 
  rebuild_idx    = 0;
  needADemux     = 0;
  
  vo_visible     = 0;
  vo_size_luma   = 0;
  vo_size_chroma = 0;

  N_BUF          = 2;  
  closed_gop     = 0;
  
  
  //-- place here initialisation of libavcodec subsystem --
  if (!ff_init_done)
  {
    avcodec_init();
    register_avcodec(&mp3_decoder);
    register_avcodec(&ac3_decoder);
    ff_init_done = 1;
  }
  
  init();
}

int T_MPEG_DEC::init(void)
{
  if (!decoder)
  { 
    //-- since mpeg2dec-0.3.1 there is no need of accel stuff --
    if ( (decoder = mpeg2_init()) == NULL )
      fprintf (stderr,"ERROR: T_MPEG_DEC::init - Can not create decoder\n");
  }
  return (decoder != NULL);
}

T_MPEG_DEC::~T_MPEG_DEC()
{
  cleanup();
}

void T_MPEG_DEC::cleanup(void)
{
  if (decoder) mpeg2_close(decoder);
  mSeqHdrLen   = 0;
  decoder      = NULL;
  act_frameSrv = NULL;
  audio_dec    = NULL;
  rebuild_idx  = 0;
  needADemux   = 0;
  
  vo_visible     = 0;
  vo_size_luma   = 0;
  vo_size_chroma = 0;
  
  closed_gop     = 0;
}

void T_MPEG_DEC::close(T_FRAME_SRV *frame_srv)
{
  if (frame_srv) delete frame_srv;
}

void T_MPEG_DEC::forceIndex(void)
{
  rebuild_idx = 1;
}

int T_MPEG_DEC::open(T_FRAME_SRV *fs, int *p)
{
  int          n, total_len;
  uint8_t      *end;
  
  if (fs)
  {
    if (rebuild_idx) fs->createIndex();
      
    if (!fs->loadIndex()) return 0;
    
    mSeqHdrLen = 0;
    closed_gop = fs->closedGOP();
    act_frame  = fs->firstSequence(1);  // must be a real sequence !
    total_len  = fs->readFrame(chunk_buf, act_frame);
    
    if (!fs->isGOP(act_frame))
    {
      if ( (n = fs->findStartCode(chunk_buf, total_len, GROUP_START_CODE)) > 0)
      {
        if (n < 128) 
        {
          mSeqHdrLen = n;
          memcpy(mSeqHdr, chunk_buf, n);
        }
      }      
    }
    
    *p  = act_frame;
    end = chunk_buf + total_len;
    n   = decode(chunk_buf, end);
  }
  else
    *p = 0;

  rebuild_idx = 0;
  
  return 1;
}

T_FRAME_SRV *T_MPEG_DEC::open(char *fname, int *p)
{
  T_FRAME_SRV  *FrameSrv = NULL;
  
  if ( (FrameSrv = T_FRAME_SRV::createFrameSrv(fname)) != NULL )
  {
    if (!open(FrameSrv, p)) 
    {
      delete FrameSrv;
      return NULL;
    }
  }  
     
  return FrameSrv;  
}

void T_MPEG_DEC::setAudioSync(int dp)
{
  T_AUDIO_DEC::avs_dp = dp;
  if (audio_dec) audio_dec->resetBuffer(act_frame);
}

int T_MPEG_DEC::getBitRate(void)
{
  const mpeg2_info_t *inf = mpeg2_info(decoder);
  if (inf && inf->sequence)
    return (inf->sequence->byte_rate * 8);
  else
    return 10000000;
}

static void showFrameInfo(uint8_t *buf, int len)
{
  T_PIC_HDR *ph = NULL;
    
  if ( (ph = T_FRAME_SRV::getPicHdr(buf, len)) != NULL )
  {
    fprintf(stderr, "t_ref (%d), type(%d=%c)\n", 
            ph->temporal_ref, ph->type, (ph->type==2)?'P':'B'); 
  } 
  else
    fprintf(stderr, "ts (0), t(1=I)\n");
}

int T_MPEG_DEC::readFrame(T_FRAME_SRV *frameSrv, int p, 
                          uint8_t *pchunk, int *len, int gop_bound)
{
  int  stream_changed = (act_frameSrv != frameSrv);
  
  //-- setup A/V-sync correction --
  //-------------------------------
  nFrames = frameSrv->nFrames();
  
  p_sav = p;
  
  //-- check for media change --
  //----------------------------
  if (stream_changed) 
  { 
    //-- force refilling of video- and audio- buffers --
    frameSrv->resetPESbuffer();
    if (needADemux) audio_dec->resetBuffer(p);
  }
  act_frameSrv = frameSrv;
    
  //-- continueous play ... --
  //--------------------------  
  if ( !gop_bound || (p == (act_frame+1)) )
  {
    //-- get chunk --
    act_frame = p;
    *len      = frameSrv->readFrame(pchunk, p);      
  }
  //-- ... or random seeking --
  //---------------------------
  else 
  {
    //-- begin at previous GOP boundaries --
    act_frame = frameSrv->prevSequence(p);
    
    if (needADemux)
    {
      //-- request for refilling of audio-buffer --
      audio_dec->resetBuffer(p);
    }
        
    //-- get chunk --
    *len = frameSrv->readFrame(pchunk, act_frame);
  }

  return act_frame;
}

int T_MPEG_DEC::decodeFrame(T_FRAME_SRV *frameSrv, int p, int *step,
                            uint8_t *py, uint8_t *pu, uint8_t *pv)
{
  int          rc, n, total_len;
  uint8_t      *end;
  
  vo_pages[0]    = py;
  vo_pages[1]    = pu;
  vo_pages[2]    = pv;
  
  p_n_buf = p + N_BUF;
  nFrames = frameSrv->nFrames();
  
  //-- on stream change --
  //----------------------
  if (act_frameSrv != frameSrv)
  { 
    //-- force refilling of video- and audio- buffers --
    frameSrv->resetPESbuffer();
    if (needADemux) audio_dec->resetBuffer(p);
    
    vo_size_luma   = frameSrv->videoWidth() * frameSrv->videoHeight();
    vo_size_chroma = vo_size_luma>>2;  
    
    act_frame      = 0;
    act_frameSrv   = frameSrv;
  }
  //-- "p" was decoded already in last call --
  //------------------------------------------ 
  else if (p == act_frame)
  {
    return 1;
  }  
  
  //-- check limit for max frame no. --
  //-----------------------------------
  if ( p_n_buf >= nFrames ) p_n_buf = nFrames-1;
  
  p_sav = p; // need for wav audio 
  
  //- on continueous play --
  //------------------------
  if ( p == (act_frame+1) )  
  {
    //-- read the frames in flow --
    total_len = frameSrv->readFrame(chunk_buf, p_n_buf);
    end       = chunk_buf + total_len;
  }
  //-- on random seek --
  //--------------------
  else 
  {
    int       d;
    T_PIC_HDR *ph;
    
    //-- begin GOP aligned --
    //-----------------------
    act_frame = frameSrv->prevSequence(p);
    d         = p-act_frame;
    
    //-- read frame ... --
    //--------------------
    total_len  = frameSrv->readFrame(chunk_buf, act_frame);
    end        = chunk_buf + total_len;
    ph         = T_FRAME_SRV::getPicHdr(chunk_buf, total_len);
    
    //-- on GOP start --
    //------------------     
    if (ph && ph->gop_valid) 
    {
      //-- ... temporal ref indicates open GOP --
      //-----------------------------------------
      if ( ph && ph->temporal_ref && ((ph->closed_gop && ph->broken_link) == 0) )
      {  
        //-- "p" is I-Frame --
        //--------------------
        if (!d && *step != -1 )
        {
          //-- the same data must be decoded for 3 times at all, do it   --
          //-- here twice, because last run have to be visible. Reset of --
          //-- "p" to -2 will force seeking again also for next picture  --
          //-- in flow, p_n_buf = 0 prevents reading of other frames.    --
          decode(chunk_buf, end);
          decode(chunk_buf, end);
          if (ph->temporal_ref != N_BUF) p = -2;
          p_n_buf = 0;
          
          //-- this special handling of I-Frames in open GOPs allows --
          //-- high speed GOP wise seeking in stream, because there  --
          //-- is no need to decode an complete additional frame     --
          //-- block (up from previous GOP).                         --
          //-- Except from this rule is jumping to an I-Frame while  --
          //-- stepping one frame back (indicated by step = -1).     --                   --
          //-- In forward direction decoder outputs same I-Frame for --
          //-- each single step as long as there are "open" B-Frames --
        }
        //-- "p" resides in "open" B-Frame-Zone --
        //----------------------------------------
        else if ( d <= ph->temporal_ref)  
        {
          //-- rewind to previous GOP --
          act_frame = frameSrv->prevSequence(act_frame-1);
          
          //-- update decoder buffer with new frame data -- 
          total_len = frameSrv->readFrame(chunk_buf, act_frame);
          end       = chunk_buf + total_len;
        }
      } 
    }
    //-- "p" may be decoded with dissortions --
    //-----------------------------------------
    else
    {
      //-- decode here twice --
      decode(chunk_buf, end);
      decode(chunk_buf, end);
      p_n_buf = 0;
    }
    
    //-- may decode some further frames except for last turn --
    //---------------------------------------------------------
    if (p_n_buf) do
    {
      n = decode(chunk_buf, end);
      act_frame++;
      
      total_len = frameSrv->readFrame(chunk_buf, act_frame);
      end       = chunk_buf + total_len;
      
    } while (act_frame < p_n_buf);

    //-- request for refilling of audio-buffer --
    //-------------------------------------------
    if (needADemux) audio_dec->resetBuffer(p);

    //-- buffer depth may be change --
    //--------------------------------  
    if (ph)
      N_BUF = (ph->pict_struct != 3)? 3:2;   // (3) for field pictures
    else                                     // (2) for frame pictures
      N_BUF = 3;  // TODO 
  }
   
  //-- decoding of last turn must be visible --
  //-------------------------------------------
  vo_visible = 1;
  act_frame  = p;
  n          = decode(chunk_buf, end);
  vo_visible = 0;

  *step = 0;
  return n;
}

//-- TODO: in decode mode (instead dump mode using readFrame)  --
//-- we have to return "act_frame - N_BUF" -> for this purpose --
//-- we could use a parameter to indicate the mode.            --
int T_MPEG_DEC::getActFrame()
{
  return p_sav; 
}

T_AUDIO_DEC *T_MPEG_DEC::getAudioDecoder(void)
{
  return audio_dec;
}

void T_MPEG_DEC::closeAudio(void)
{
  if (audio_dec) 
  {
    audio_dec->close();
    delete audio_dec;
    audio_dec = NULL;
  }
  needADemux = 0;
}

int T_MPEG_DEC::openAudio(T_FRAME_SRV *frame_srv, int track)
{
  closeAudio();
  if ( (audio_dec = T_AUDIO_DEC::createAudioDec(frame_srv, track)) != NULL ) 
  {
    int rc;
    
    rc         = audio_dec->open(frame_srv, this);
    needADemux = audio_dec->needDemux;
   
    return rc;
  }
  else
  {
    needADemux = 0;
    return 0;
  }
}

//============================================================================
//==                             A U D I O                                  ==
//============================================================================

//=======================
//== class T_AUDIO_DEC ==
//=======================
int T_AUDIO_DEC::avs_dp = 0;

T_AUDIO_DEC* T_AUDIO_DEC::createAudioDec(T_FRAME_SRV *frame_srv, int track)
{
  char        fname[1024];
  int         l;
  struct stat stat_buf;

  //-- TODO: better detection handling of wav-File --
  strcpy(fname, frame_srv->getFileName());
  l = strlen(fname);
  fname[l-4] = '\0';
  strcat(fname, ".wav");
  
  if (stat(fname, &stat_buf) != -1) 
  {
    // fprintf(stderr, "*** using WAV audio decoder\n");
    return new T_AUDIO_DEC_WAV;
  }
 
  //-- sets the desired track in frame-server --
  //-- and get the type of audio stream       --
  uint8_t t = frame_srv->getAudioType(track);
  switch( t )
  {  
    case A_TYPE_A52_SUB:
    case A_TYPE_PCM_SUB:
      return new T_AUDIO_DEC_FFMPEG(CODEC_ID_AC3, track);
      break;
 
    case A_TYPE_MPEG1_L2:
      return new T_AUDIO_DEC_FFMPEG(CODEC_ID_MP3, track);
      break;
          
    default:
      fprintf(stderr, "INFO: T_AUDIO_DEC - no decoder available for (0x%02x) streams\n", t);  
      return new T_AUDIO_DEC_NONE();
      break;
  }
  
  return NULL;
}

T_AUDIO_DEC::T_AUDIO_DEC(int track)
{
  needDemux    = 0;
  track_no     = track;
  video_rate   = 25.0;
  sample_rate  = 48000;
  samples_pvf  = 2048.0;
  bit_rate     = 0;
  v_decoder    = NULL;
  act_frameSrv = NULL;
  af           = NULL;
  
  ffmpeg_id    = CODEC_ID_NONE;
  
  v_decoder    = NULL;
  act_frameSrv = NULL;
}

T_AUDIO_DEC::~T_AUDIO_DEC()
{
  if (af) fclose(af);
  af           = NULL;
  v_decoder    = NULL;
  act_frameSrv = NULL;
}

CodecID T_AUDIO_DEC::getCodecId(void)
{
  return ffmpeg_id;
}

int T_AUDIO_DEC::open(T_FRAME_SRV *frame_srv, T_MPEG_DEC *video_dec)
{
  v_decoder    = video_dec;
  act_frameSrv = frame_srv; 
  
  video_rate   = frame_srv->videoRate();
  sample_rate  = frame_srv->sampleRate(track_no);
  bit_rate     = frame_srv->bitRate(track_no);

  samples_pvf  = sample_rate/video_rate;
  samples_paf  = (ffmpeg_id == CODEC_ID_AC3)? 1536 : 1152;
  size_paf     = (int)( (samples_paf * bit_rate) / (8.0 * sample_rate)); // + 0.5);

  needDemux    = 0;
  
  return 0;
}

void T_AUDIO_DEC::close(void)
{
  needDemux = 0;
  return;
}

void T_AUDIO_DEC::resetBuffer(int p)
{
  if (act_frameSrv) act_frameSrv->resetAudioBuffer(p-avs_dp);
  bytes_used   = 0;
  bytes_unused = 0;
  refilled     = 1;
}

int T_AUDIO_DEC::readSamples(uint32_t *sample_buf, double *samples_cnt)
{
  return 0;
}

int T_AUDIO_DEC::readChunkData(uint8_t *chunk_buf, double *samples_cnt)
{
  return 0;
}

int T_AUDIO_DEC::getSampleRate(void)
{
  return sample_rate;
}

double T_AUDIO_DEC::getSamplesPVF(void)
{
  return samples_pvf;
}

int T_AUDIO_DEC::getEncodedFs(void)
{
  return size_paf;
}

T_FRAME_SRV *T_AUDIO_DEC::getFrameSrv(void)
{
  return act_frameSrv;
}  

//======================
//== T_AUDIO_DEC_NONE ==
//======================
int T_AUDIO_DEC_NONE::open(T_FRAME_SRV *frame_srv, T_MPEG_DEC *video_dec)
{
  act_frameSrv = frame_srv; 
  needDemux    = 0;
  return 0;
}

//=====================
//== T_AUDIO_DEC_WAV ==
//=====================
int T_AUDIO_DEC_WAV::open(T_FRAME_SRV *frame_srv, T_MPEG_DEC *video_dec)
{
  char         fname[1024];
  char         *ps;
  T_WAV_HEADER wav_hdr;
  int          n  = 0;

  strcpy(fname, frame_srv->getFileName());
  if ( (ps = strrchr(fname, '.')) != 0 ) *ps = '\0';
  strcat(fname, ".wav");
  
  v_decoder    = video_dec;
  act_frameSrv = frame_srv; 
  act_frame    = -2;
  sample_rate  = 48000;
  video_rate   = act_frameSrv->videoRate();

  if ( (af = fopen(fname, "rb")) != NULL )
  { 
    n = fread(&wav_hdr, 1, sizeof(T_WAV_HEADER), af);
  
    if ( n == sizeof(T_WAV_HEADER) )
    {  
      sample_rate  = ((unsigned char)(wav_hdr.sample_rate[1]) << 8);
      sample_rate |= (unsigned char)(wav_hdr.sample_rate[0]);
      samples_pvf  = sample_rate / video_rate;
      
      fprintf(stderr, "INFO: using external audio (%s) - (%d Hz), header size (%d):\n", 
             fname, sample_rate, n);
      return 1;
    }
    close();
  }
  return 0; 
}

void T_AUDIO_DEC_WAV::close(void)
{
  if (af != NULL) fclose(af);
  af = NULL;
}

int T_AUDIO_DEC_WAV::readSamples(uint32_t *sample_buf, double *samples_cnt)
{
  int n = 0;
  int p = v_decoder->getActFrame() - avs_dp;

  if (p<0) return 0;

  if (af)
  {
    SAMPLES_PVF = samples_pvf; 
    if ( (p-act_frame) != 1 ) 
    {
      int64_t sk;
      
      if (p==0) p++;
      sk = (int64_t)(sizeof(T_WAV_HEADER) + ( (p-1)*4*SAMPLES_PVF ) + 0.5); 
      
      fseek(af, sk, SEEK_SET ); 
    }
    n = fread((void *)sample_buf, 1, (uint32_t)(4 * SAMPLES_PVF + 0.5), af);
    act_frame = p;
  } 
  return n;
}

//========================
//== T_AUDIO_DEC_FFMPEG ==
//========================


T_AUDIO_DEC_FFMPEG::T_AUDIO_DEC_FFMPEG(CodecID id, int track):T_AUDIO_DEC(track)
{
  ffmpeg_id    = id;
  ffmpeg_codec = NULL;
  ffmpeg_ctx   = NULL;
  isOpen       = 0;
}

int T_AUDIO_DEC_FFMPEG::open(T_FRAME_SRV *frame_srv, T_MPEG_DEC *video_dec)
{
  T_AUDIO_DEC::open(frame_srv, video_dec);

  bytes_used   = 0;
  bytes_unused = 0;
  refilled     = 1;
  
  if (!ffmpeg_codec)
  {   
    ffmpeg_codec = avcodec_find_decoder(ffmpeg_id);
    if (!ffmpeg_codec)
      fprintf (stderr,"ERROR: T_AUDIO_DEC_FFMPEG::open - cannot create decoder (%d)\n", ffmpeg_id);
    else
    {
      ffmpeg_ctx = avcodec_alloc_context();
      
      if (avcodec_open(ffmpeg_ctx, ffmpeg_codec) < 0)
      {
        fprintf(stderr, "ERROR: T_AUDIO_DEC_FFMPEG - could not open decoder (%d)\n", ffmpeg_id);
        return 0;
      }  
      
      ffmpeg_ctx->channels = 2;  // force to stereo
      
      needDemux = frame_srv->getTotalAudioStreams();
      isOpen    = 1;
    } 
  }
  return (ffmpeg_codec != NULL);
}

void T_AUDIO_DEC_FFMPEG::close(void)
{
  if (ffmpeg_codec && isOpen) avcodec_close(ffmpeg_ctx);
  if (ffmpeg_ctx) free(ffmpeg_ctx);
  
  ffmpeg_codec = NULL;
  ffmpeg_ctx   = NULL;
  isOpen       = 0;  
  needDemux    = 0;
}

//-- decoder amount of audio fitting to 1 video frame -- 
int T_AUDIO_DEC_FFMPEG::decode(int8_t *sample_buf, double *samples_cnt)
{
  uint8_t *stream_buf;
  int     len, total_len, buf_len;
  int     bytes_needed, out_size;
  
  total_len = 0;
  out_size  = 0;
  
  //-- preset this --
  SAMPLES_PVF = samples_pvf; 
  bytes_needed = (int)(SAMPLES_PVF * 4 + 0.5);
  
  //-- get buffer of encoded audio-stream --
  stream_buf = act_frameSrv->readAStream(&buf_len);

  //-- buffer fill level should be enough --
  if (buf_len >= MPA_MAX_FRAME_SIZE) 
  {
    //-- move unused samples to the beginning of the sample --
    //-- buffer and ajust pointer behind this data.         --
    if (bytes_unused)
    {
      memmove(sample_buf, sample_buf+bytes_needed, bytes_unused);
      sample_buf  += bytes_unused;
    }
    bytes_used = bytes_unused;

    while (bytes_used < bytes_needed)
    {
      //-- refilled buffer must be synced first --
      if (refilled)
      {
        int      sr, br;
        int      id = (ffmpeg_id == CODEC_ID_MP3)? A_TYPE_MPEG1_L2:A_TYPE_A52_SUB;
        
        refilled = 0;
      
        dosync:
          len      = act_frameSrv->syncAStream
                        (id, stream_buf, buf_len, &sr, &br);
      
        if (len != -1)
        {
          //-- dismiss these bytes --
          buf_len    -= len;
          stream_buf += len;
        
          //-- MP2 audio needs extra checks ! --
          if (id == A_TYPE_MPEG1_L2)
          {
            if (memcmp(stream_buf, &stream_buf[size_paf], 4))
            {
              stream_buf += 2;
              buf_len    -= 2;
              goto dosync;
            }  
          }
        }
          
      }
      
      len = avcodec_decode_audio(ffmpeg_ctx, (int16_t *)sample_buf, &out_size,
                                 (uint8_t *)stream_buf, buf_len); 
      stream_buf += len;
      buf_len    -= len;

      if (len < 0 || buf_len <= 7)
      {
        fprintf(stderr, "WARNING: could not decode audio-frame\n");
        bytes_needed = 0;
        break;
      } 
      else 
      {
        total_len  += len;
        sample_buf += out_size;
        bytes_used += out_size;
      }  
      
    }  
    bytes_unused = bytes_used - bytes_needed;
    
    refilled = 0;
    
    if (total_len < 0) total_len = 0;
    act_frameSrv->adjustAStream(total_len);
  }
  else
  {
    bytes_needed = 0;
    fprintf(stderr, "WARNING: audio buffer underrun (%d)\n", buf_len);
  }
      
  return bytes_needed;
}


int T_AUDIO_DEC_FFMPEG::readChunkData(uint8_t *chunk_buf, double *samples_cnt)
{
  uint8_t *start, id;
  int     bytes_sync, bytes_chunk, buf_len;
  int     sr, br;

  int p = v_decoder->getActFrame() - avs_dp;
  if (p<0 || p>act_frameSrv->nFrames()) return 0;
  
  //-- set type of audio stream --
  id = (ffmpeg_id == CODEC_ID_MP3)? 0xC0 : 0x80;

  //-- preset amount of bytes needed for 1 video frame in decoded state  --
  SAMPLES_PVF = samples_pvf; 

  //-- count total needed samples (since 1st. read) --
  SAMPLES_NEEDED += samples_pvf;
  
  //-- fill source with audio-stream data --
  act_frameSrv->fillAudioBuffer();
  start = act_frameSrv->readAStream(&buf_len);

  //-- get at least amount of needed samples --
  bytes_sync  = 0;
  bytes_chunk = 0;
  while (SAMPLES_USED < SAMPLES_NEEDED) 
  {
    //-- refilled buffer must be synced first ... --
    if (refilled)
    {
      refilled = 0;
      
      dosync:  
        //-- ... to start of a new audio frame --
        bytes_sync = act_frameSrv->syncAStream(id, start, buf_len, &sr, &br);
      
      if (bytes_sync != -1)
      {
        //-- dismiss these bytes --
        buf_len -= bytes_sync;
        start   += bytes_sync;
    
        //-- MP2 needs extra checks ! --
        if (id == 0xC0)
        {
          if (memcmp(start, &start[size_paf], 4))
          {
            start   += 2;
            buf_len -= 2;
            goto dosync;
          }  
        }
      }
    }  
    
    // TODO check buffer underun (buf_len > size_paf) ...
    bytes_chunk    += size_paf;
    SAMPLES_USED   += samples_paf;
  }

  //-- transfer (payload) chunk data to destination --          
  memcpy(chunk_buf, start, bytes_chunk);
  
  //-- remove processed bytes from source buffer --
  act_frameSrv->adjustAStream(bytes_chunk + bytes_sync);

  refilled = 0;
  
  return bytes_chunk;
}

int T_AUDIO_DEC_FFMPEG::readSamples(uint32_t *sample_buf, double *samples_cnt)
{
  int p = v_decoder->getActFrame() - avs_dp;
  if (p<0 || p>act_frameSrv->nFrames()) return 0;
  
  act_frameSrv->fillAudioBuffer();
  return decode((int8_t *)sample_buf, samples_cnt);
}

// eof
