/*
 *  display.c
 *
 *  Copyright (C) Gerhard Monzel - October 2003
 *
 *  This file is part of lve, a linux mpeg video editor
 *      
 *  lve tools are free software; you can redistribute it and/or modify
 *  it under the terms of the GNU General Public License as published by
 *  the Free Software Foundation; either version 2, or (at your option)
 *  any later version.
 *   
 *  lve is distributed in the hope that it will be useful,
 *  but WITHOUT ANY WARRANTY; without even the implied warranty of
 *  MERCHANTABILITY or FITNESS FOR A PARTICULAR PURPOSE.  See the
 *  GNU General Public License for more details.
 *   
 *  You should have received a copy of the GNU General Public License
 *  along with GNU Make; see the file COPYING.  If not, write to
 *  the Free Software Foundation, 675 Mass Ave, Cambridge, MA 02139, USA. 
 *
*/
#include <stdio.h>
#include <stdlib.h>
#include <string.h>
#include <SDL/SDL.h>
#include "lve.h"
#include "display.h"
#include "lve_gui.h"

#define CURSOR1_WH  24
#define CURSOR1_SIZE (CURSOR1_WH*(CURSOR1_WH/8))

static SDL_Surface *screen  = NULL;
static SDL_Surface *bmp     = NULL;
static SDL_Cursor  *cursor0 = NULL;
static SDL_Cursor  *cursor1 = NULL;

//== Drag-Drop Cursor Image ==
//============================
static Uint8 c1_data[CURSOR1_SIZE] =
{
       0, 0x00,    0,
       0, 0x18,    0, 
       0, 0x3C,    0,
       0, 0x7E,    0,
    0x00, 0x18, 0x00,
       0, 0x18,    0,
       0, 0x18,    0,
    0x00, 0x18, 0x00, 
    
    0x00, 0x18, 0x00,
    0x10, 0x18, 0x08,
    0x30, 0x3C, 0x0C,
    0x7F, 0xFF, 0xFE,
    0x7F, 0xFF, 0xFE,
    0x30, 0x3C, 0x0C, 
    0x10, 0x18, 0x08,
    0x00, 0x18, 0x00, 

    0x00, 0x18, 0x00,
       0, 0x18,    0,
       0, 0x18,    0,
    0x00, 0x18, 0x00,
       0, 0x7E,    0,
       0, 0x3C,    0, 
       0, 0x18,    0,
       0, 0x00,    0,
};

static Uint8 c1_mask[CURSOR1_SIZE] =
{
       0, 0x18,    0,
       0, 0x3C,    0, 
       0, 0x7E,    0,
       0, 0xFF,    0,
    0x01, 0xFF, 0x80,
       0, 0x3C,    0,
       0, 0x3C,    0,
    0x08, 0x3C, 0x10, 
    
    0x18, 0x3C, 0x18,
    0x38, 0x3C, 0x1C,
    0x7F, 0xFF, 0xFE,
    0xFF, 0xFF, 0xFF,
    0xFF, 0xFF, 0xFF,
    0x7F, 0xFF, 0xFE, 
    0x38, 0x3C, 0x1C,
    0x18, 0x3C, 0x18, 

    0x08, 0x3C, 0x10,
       0, 0x3C,    0,
       0, 0x3C,    0,
    0x01, 0xFF, 0x80,
       0, 0xFF,    0,
       0, 0x7E,    0, 
       0, 0x3C,    0,
       0, 0x18,    0,
};

void display_set_cursor(char t)
{
  static char _t = 0;
  
  if (_t == t) return;
  _t = t; 
   
  if (t==1)
    SDL_SetCursor(cursor1);
  else
    SDL_SetCursor(cursor0);
}

SDL_Surface* display_init(T_LVE_CTX *ctx, char *lib_path, char *title)
/*
 * Initialize an SDL screen and an SDL YUV overlay.
 *
 *    params : width  == width of video we'll be displaying.
 *             height == height of video we'll be displaying.
 *             fullscreen == want to be fullscreen?
 *             title == Title for window titlebar.
 *   returns : non-zero on success, zero on error.
 */
{
  char     fname[1024];
  Uint8    bpp;
  SDL_Rect dst_rect;

  int      rc 		= 0;
  int      i 		= 0;
  int      desiredWidth	= -1;
  int      desiredHeight= -1;
  SDL_Rect **modes	= NULL;
  Uint32   sdlflags	= SDL_HWSURFACE; // | SDL_ASYNCBLIT | SDL_HWACCEL;
  
  const SDL_VideoInfo *vidInfo = NULL;
    
    
  //== init SDL functions ==
  //========================  
  //rc = SDL_Init(SDL_INIT_VIDEO | SDL_INIT_TIMER | SDL_INIT_AUDIO | SDL_INIT_NOPARACHUTE);
  rc = SDL_Init(SDL_INIT_VIDEO | SDL_INIT_TIMER | SDL_INIT_NOPARACHUTE);
  if (rc != 0)
  {
    fprintf(stderr, "ERROR: display_init() - SDL_init() failed! rc == (%d).\n", rc);
    return(NULL);
  } 
  atexit(SDL_Quit);

  vidInfo = SDL_GetVideoInfo();

  modes = SDL_ListModes(vidInfo->vfmt, sdlflags);
  if (modes == NULL)
  {
    fprintf(stderr, "WARNING: SDL modes w/o HWSURFACE\n");
    sdlflags &= ~SDL_HWSURFACE;
    modes = SDL_ListModes(vidInfo->vfmt, sdlflags);   // give me ANYTHING.

    if (modes == NULL)
    {
      fprintf(stderr, "ERROR: SDL_ListModes() failed.\n");
      return(NULL);
    } 
    
    for (i=0; modes[i]; i++)
      printf("   %d x %d\n", modes[i]->w, modes[i]->h);
  }
  else if (modes == (SDL_Rect **)-1)   // anything is fine.
  {
    desiredWidth  = 0;
    desiredHeight = 0;
  } 
  else
  {
    // we want to get the lowest resolution that'll fit the video.
    //  ...so start at the far end of the array.
    i = 0;
    while (modes[i] != NULL)
    {
      if ( (modes[i]->w < ctx->w) || (modes[i]->h < ctx->h) ) break;
      desiredWidth  = modes[i]->w;
      desiredHeight = modes[i]->h;
      i++;
    }
  } 

  if ((desiredWidth < 0) || (desiredHeight < 0))
  {
    fprintf(stderr, "ERROR: Couldn't produce a mode with at least"
                    " a (%dx%d) resolution!\n", ctx->w, ctx->h);
      return(NULL);
  } 

  //-- Show Cursor --
  SDL_ShowCursor(1);
   
  //memset(c1_data,0xFF, CURSOR1_SIZE);
  cursor0 = SDL_GetCursor();
  cursor1 = SDL_CreateCursor(c1_data, c1_mask, 
                             CURSOR1_WH, CURSOR1_WH, 
                             CURSOR1_WH/2, CURSOR1_WH/2);

  // YUV overlays need at least 16-bit color depth, but the
  //  display might less. The SDL AAlib target says it can only do
  //  8-bits, for example. So, if the display is less than 16-bits,
  //  we'll force the BPP to 16, and pray that SDL can emulate for us.
  bpp = vidInfo->vfmt->BitsPerPixel;
  if (bpp < 16)
  {
    fprintf(stderr,
            "\n\n"
            " WARNING: Your SDL display target wants to be at a color\n"
            "  depth of (%d), but we need it to be at least 16 bits,\n"
            "  so we need to emulate 16-bit color. This is going to slow\n"
            "  things down; you might want to increase your display's\n"
            "  color depth, if possible.\n\n", bpp);
    bpp = 16;  // (*shrug*)
  } 
 
  //== loading main Bitmap in separat surface ==
  //============================================ 
  strcpy(fname, lib_path);
  strcat(fname, FN_SCREEN_BMP);
  if ((bmp = SDL_LoadBMP(fname)) == NULL)
  {
    fprintf(stderr, "ERROR: SDL could not load screen bitmap (%s)\n", fname);
  }
  else if (!desiredWidth && !desiredHeight)
  {
    desiredWidth  = bmp->w;
    desiredHeight = bmp->h;
  }
     
  //== create screen with probed width and height ==
  //================================================   
  screen = SDL_SetVideoMode(desiredWidth, desiredHeight, bpp, sdlflags);
  if (screen == NULL)
  {
    printf("ERROR: SDL could not set video mode!\n");
    return(NULL);
  } 
  ctx->screen = screen;

  //== now blit the bitmap to display screen  and free bitmap surface ==
  //====================================================================
  dst_rect.x = 0;
  dst_rect.y = 0;
  dst_rect.w = screen->w;
  dst_rect.h = screen->h;
  SDL_BlitSurface(bmp, NULL, screen, &dst_rect);
  SDL_FreeSurface(bmp);

  //== loading system font ==
  //=========================
  strcpy(fname, lib_path);
  strcat(fname, FN_FONT_SYS_BMP);
  DT_LoadFont(fname, TRANS_FONT);
  
  strcpy(fname, lib_path);
  strcat(fname, FN_FONT_SMALL_BMP);
  DT_LoadFont(fname, TRANS_FONT);

  strcpy(fname, lib_path);
  strcat(fname, FN_FONT_LARGE_BMP);
  DT_LoadFont(fname, TRANS_FONT);

      
  if (title == NULL) title = "Linux Video Editor (LVE) V1.1";
  SDL_WM_SetCaption(title, "LVE");

  //-- don't need mouse movement events --
  SDL_EventState(SDL_MOUSEMOTION, SDL_IGNORE);
  SDL_EventState(SDL_SYSWMEVENT, SDL_IGNORE);
  
  SDL_EnableKeyRepeat(500, 100);

  //SDL_WM_ToggleFullScreen(screen);

  return (screen);
}

static SDL_Overlay *overlay = NULL;
static SDL_Rect    ov_rect;

void display_cleanup_ov()
{
  if (overlay)
  {
    SDL_FreeYUVOverlay(overlay);
    overlay = NULL;
  }
}

void display_reset_ov(T_LVE_CTX *ctx)
{
  ctx->reset_ov = 1;
}

static int w_last = 0;
static int h_last = 0;

SDL_Overlay *display_init_ov(T_LVE_CTX *ctx)
{ 
  ov_rect.x = REGION(RG_ID_OV).GetRectX(0);
  ov_rect.y = REGION(RG_ID_OV).GetRectY(0);
  ov_rect.w = OV_WIDTH;
  ov_rect.h = OV_HEIGHT;
  
  //-- optimal adjusting for SDL resizer --
  if (ctx->w > OV_WIDTH) 
    ov_rect.w = OV_WIDTH;
  if (ctx->h > 460) 
    ov_rect.h = ctx->h>>1;
  else if (ctx->h > OV_HEIGHT)
    ov_rect.h = OV_HEIGHT;
    
  //-- now center overlay --
  ov_rect.x += ((OV_WIDTH - ov_rect.w)>>1);
  ov_rect.y += ((OV_HEIGHT - ov_rect.h)>>1);

  if (ctx->w == w_last && ctx->h == h_last)
    return overlay;
  
  w_last = ctx->w;
  h_last = ctx->h;
 
  display_cleanup_ov();

  overlay = SDL_CreateYUVOverlay(ctx->w, ctx->h, ctx->ov_fmt, screen);
  if (overlay == NULL)
  {
    fprintf(stderr, "ERROR: Couldn't create an SDL-based YUV overlay!\n");
    ctx->page = NULL;
  } 
  else
  {
    if (overlay->hw_overlay)
      fprintf(stderr, "Video overlay is hardware accelerated :-)\n");
    ctx->page = overlay->pixels;
  }
  
  ctx->ov = overlay; 

  ctx->refresh_bg = 2;
  display_refresh(ctx);

  return(overlay);  // non-zero == SUCCESS. Oooh yeah.
}


Uint32 display_draw_frame(T_LVE_CTX *ctx)
/*
 * Draw a frame to the SDL YUV overlay.
 *
 *   params : *src[] == the Y, U, and V planes that make up the frame.
 *  returns : non-zero on success, zero on error.
 */
{
  if (overlay) SDL_DisplayYUVOverlay(overlay, &ov_rect);
  return(1);
}

void display_lock()
{
  while (SDL_LockYUVOverlay(overlay) < 0);
}

void display_unlock()
{
  SDL_UnlockYUVOverlay(overlay);
}  

static int ov_h;
static int up_w, up_h;
static int up_h1, up_w1;
void display_init_update(void)
{
  up_w = REGION(RG_ID_OV).GetRectX(0);
  ov_h = REGION(RG_ID_OV).GetRectY(0) + OV_HEIGHT;
  up_h = screen->h - ov_h;

  up_h1 = REGION(RG_ID_OV).GetRectY(0);
  up_w1 = screen->w - (REGION(RG_ID_OV).GetRectX(0) + OV_WIDTH);
}

void display_refresh(T_LVE_CTX *ctx)
{
  if (ctx->refresh_bg > 1)
  {
    display_draw_frame(ctx);
    SDL_UpdateRect(screen, up_w, 0, OV_WIDTH, up_h1);
    SDL_UpdateRect(screen, up_w + OV_WIDTH, 0, up_w1, ov_h);
  }
  SDL_UpdateRect(screen, 0, 0, up_w, ov_h);
  SDL_UpdateRect(screen, 0, ov_h, screen->w, up_h);  
  
  ctx->refresh_bg = 0;
}

void display_OvText(T_LVE_CTX *ctx, char *txt, uint8_t flag)
{
  static SDL_Surface *sf  = NULL;
  static int         line = 0;

  if (!overlay) display_init_ov(ctx);

  if (!sf || (flag & 0x04) )
  {
    if ( (flag & 0x04) && sf ) SDL_FreeSurface(sf);
    sf = SDL_CreateRGBSurfaceFrom
           (ctx->page[0], ctx->w, ctx->h, 8, ctx->w, 0xFF, 0, 0, 0);
  }
    
  if (sf)
  {
      int h;
      int font = (ctx->w > 400)? 2 : 1; 
      int n    = ctx->w * ctx->h / 4;
 
      if (flag & 0x01)
      {
        line = 0;
        memset(ctx->page[0], 0xFF, n*4);
        memset(ctx->page[1], 0x7F, n); 
        memset(ctx->page[2], 0x7F, n); 
      }
    
      h = DT_FontHeight(font);
      h = (h + h/4) * line;
      DT_DrawText(txt, sf, font, 10, h+10, 1);
      line ++;
    
      if (flag & 0x02) display_draw_frame(ctx);
  }
}



// eof
