/*
 *  frame_srv.c
 *
 *  Copyright (C) Gerhard Monzel - October 2003
 *
 *  This file is part of lve, a linux mpeg video editor
 *      
 *  lve tools are free software; you can redistribute it and/or modify
 *  it under the terms of the GNU General Public License as published by
 *  the Free Software Foundation; either version 2, or (at your option)
 *  any later version.
 *   
 *  lve is distributed in the hope that it will be useful,
 *  but WITHOUT ANY WARRANTY; without even the implied warranty of
 *  MERCHANTABILITY or FITNESS FOR A PARTICULAR PURPOSE.  See the
 *  GNU General Public License for more details.
 *   
 *  You should have received a copy of the GNU General Public License
 *  along with GNU Make; see the file COPYING.  If not, write to
 *  the Free Software Foundation, 675 Mass Ave, Cambridge, MA 02139, USA. 
 *
 */
#include <ctype.h>
#include "media_buf.h"
#include "frame_srv.h"

#ifdef __cplusplus
extern "C"
{
#endif

#include "a52.h"

#ifdef __cplusplus
}
#endif

static uint8_t _pes_vs_buf[PES_VS_BUF_LEN];
static uint8_t _pes_as_buf[PES_AS_BUF_LEN];
static uint8_t *pes_vs_buf = _pes_vs_buf;
static uint8_t *pes_as_buf = _pes_as_buf;

static uint8_t audio_buf[MAX_ABUF_LEN];
static uint8_t *paudio_buf = audio_buf;
static int     ab_level    = 0;

//===========================
//== class T_USER_PROGRESS ==
//===========================
T_USER_PROGRESS::T_USER_PROGRESS()
{};

void T_USER_PROGRESS::show(int p)
{
  fprintf(stderr, "Picture #(%d)\r\n", p);
}

//=======================
//== class T_FRAME_SRV ==
//=======================

T_USER_PROGRESS *T_FRAME_SRV::userProgress = NULL;

void T_FRAME_SRV::setUserProgress(T_USER_PROGRESS *progress)
{
  userProgress = progress;
}

T_FRAME_SRV *T_FRAME_SRV::createFrameSrv(char *fname)
{
  char     *p;
  int      ch, cnt, done;
  T_CODE   code32 = 0xFFFFFF;
  uint16_t code16 = 0;  
  char     type   = MPEG_UNKNOWN;
  uint8_t  vid    = 0; 
  FILE     *tf;

  //-- test for index extension and cut it --
  if ( ((p=strstr(fname, EXT_INDEX)) != NULL) && (p[4] == '\0') ) 
  {
    ch = *p;
    
    //-- test for numeric index count "-#99" --
    //-- and cut it too, if available.       --
    if ( strlen(fname) > 8 )
    { 
      p -= 4;
      if ( !strncmp(p ,"-#", 2) && isdigit(p[2]) && isdigit(p[3]) ) 
        ch = *p;
      else
        p += 4;  
    }
    *p = '\0';
  }
  else
    p = NULL;
    
  //-- check here type of media file --
  if ( (tf = fopen(fname,"rb")) != NULL )
  {
    //-- restore fname --
    if (p) *p = ch;

    cnt  = 0;
    done = 0;
    
    if (sizeof(EDIT_BUF_MAGIC) == fread(pes_vs_buf, 1, sizeof(EDIT_BUF_MAGIC), tf) )
    {
      if (!memcmp(pes_vs_buf, EDIT_BUF_MAGIC, sizeof(EDIT_BUF_MAGIC)))
      {
        type = MPEG_LVE;
        done = 1;
      }
    }

    //-- test for transport stream (TS) --
    //------------------------------------
    cnt  = 0;
    //done = 0;
    while (!done && (ch = fgetc(tf)) != EOF)
    {
      //-- lock to sync byte --
      if (ch == TS_SYNC_BYTE)  
      {
        off_t sav  = ftello(tf);
        int ts_cnt = 0;
        int i;
        
        for (i=0; i<10; i++)
        {
          //-- look ahead at pack boundary --
          fseeko(tf, TS_PACK_LEN-1, SEEK_CUR);
          ch = fgetc(tf);
        
          //-- again sync byte ? --
          if (ch == TS_SYNC_BYTE) 
          {
            ts_cnt++;
            continue;
          }
        }
        
        if (ts_cnt == 10)
        {
          type = MPEG_TS;
          done = 1;
        }
        else
          fseeko(tf, sav, SEEK_SET);
      }
      
      if (cnt > 10*TS_PACK_LEN) break;
      cnt++; 
    }  
    
    //-- test for PVA stream --
    //-------------------------
    cnt = 0;
    while (!done && (ch = fgetc(tf)) != EOF)
    {
      code16 = (code16 << 8) | ch; 
      
      //-- lock to sync word --
      if (code16 == PVA_SYNC_WORD)  
      {
        off_t sav   = ftello(tf);
        int pva_cnt = 0;
        int i;
        
        for (i=0; i<3; i++)
        {
          fread(pes_vs_buf, 1, PVA_HDR_SIZE, tf);
          if ( pes_vs_buf[0] == PVA_VIDEO_ID || pes_vs_buf[0] == PVA_AUDIO_ID )
          {
            int len = (pes_vs_buf[4]<<8) | pes_vs_buf[5];
            fseeko(tf, len, SEEK_CUR);
            code16 = (fgetc(tf)<<8) | fgetc(tf);
          
            //-- again sync word ? --
            if (code16 == PVA_SYNC_WORD) 
            {
              pva_cnt++;
              continue;
            }
          }
        }
        
        if (pva_cnt == 3)
        {
          type = MPEG_PVA;
          done = 1;
        }
        else
          fseeko(tf, sav, SEEK_SET);
      }
      
      if (cnt > 1*1024*1024) break;
      cnt++; 
    }  
    
    //-- test for PS/PES/ES streams --
    //--------------------------------
    cnt = 0;
    fseeko(tf, 0, SEEK_SET);
    while (!done && (ch = fgetc(tf)) != EOF)
    {
      code32 = (code32 << 8) | ch;
      switch(code32)
      {
         case PACK_START_CODE:
           done = 1;
           type = MPEG_PS;
           break;
           
         case SEQ_START_CODE:
         case GROUP_START_CODE:
           done = 1;
           type = MPEG_ES;
           break;
           
         default:
           if ( (code32 & STREAM_ID_MASK) == VSTREAM_MASK ) 
           {  
             vid  = code32 & 0x0F; // take first id found
             type = MPEG_PES;
             done = 1;
           }  
           break;
      }
      
      //-- scan only max. first 64 MB of file --
      if (cnt > 64*1024*1024) break;
      cnt++;
    }
    
    fclose(tf);      
  }
  
  switch (type)
  {
    case MPEG_PS:
      return (T_FRAME_SRV *)(new T_FRAME_SRV_PS(type, fname, vid));
  
    case MPEG_ES:
      return (T_FRAME_SRV *)(new T_FRAME_SRV_ES(type, fname, vid));
  
    case MPEG_PES:
      return (T_FRAME_SRV *)(new T_FRAME_SRV_PES(type, fname, vid));

    case MPEG_TS:
      return (T_FRAME_SRV *)(new T_FRAME_SRV_TS(type, fname, vid));
  
    case MPEG_PVA:
      return (T_FRAME_SRV *)(new T_FRAME_SRV_PVA(type, fname, vid));
  
    case MPEG_LVE:
      return (T_FRAME_SRV *)(new T_FRAME_SRV(type, fname, vid));
  
    default:
      fprintf(stderr, "T_FRAME_SRV - unknown media-type for (%s)\n", fname);
      return NULL;
  }
}


T_FRAME_SRV::T_FRAME_SRV(int t, char* _fname, uint8_t vid)
{
  stream   = &_stream;
  type     = t;
  maxIndex = (uint32_t)-1;  // no restriction
  
  memcpy(stream->magic, LVE_MAGIC, MAX_MAGIC_LEN);
  stream->mpeg2_flag = -1;
  
  stream->atype    = 0;  // 0=substream; 1=audio stream start code
  stream->vs_id    = vid;
  stream->as_id    = 0;
  stream->pack_cnt = 0;
  stream->vs_cnt   = 0;
  stream->as_cnt   = 0;
  stream->sys_cnt  = 0;
  stream->seq_cnt  = 0;
  stream->gop_cnt  = 0;
  stream->pict_cnt = 0;
  stream->width    = -1;
  stream->height   = -1;
  stream->open_gop = 0;
  stream->n_as     = 0;
  stream->n_vs     = 0;

  ls_pict = NULL;
  vf      = NULL;
  pf      = NULL;
  af      = NULL;
  
  strcpy(fname, _fname);

  g_pes_pos_old = 0;
  g_pes_len_old = 0;
}

T_FRAME_SRV::~T_FRAME_SRV()
{
  cleanup();

  if (ls_pict) delete ls_pict;
  ls_pict = NULL;
}

void T_FRAME_SRV::cleanup(void)
{
  if (vf) fclose(vf);
  if (pf) fclose(pf);
  if (af) fclose(af);
  
  vf = NULL;
  pf = NULL;
  af = NULL;
}

uint8_t *T_FRAME_SRV::getPESbuffer(int t)
{
  if (!t) 
    return pes_vs_buf;
  else
    return pes_as_buf;
}

void T_FRAME_SRV::resetPESbuffer(void)
{
  // this flags pes-buffer as dirty (read from file is needed)
}

void T_FRAME_SRV::resetAudioBuffer(int p)
{
  if (p<0) p = 0;
  if (af && ls_pict) fseeko(af, ls_pict[p].pos+ls_pict[p].len, SEEK_SET);

  ab_level   = 0;
  paudio_buf = audio_buf;
}

void T_FRAME_SRV::processVCode(T_CODE code, 
                               off_t seek_pos, int32_t read_len, 
                               int32_t i, int32_t end)
{
  switch (code)
  {
    case SEQ_START_CODE:
        stream->seq_cnt++;
              
        seq_pos.pos = seek_pos;
        seq_pos.len = (uint16_t)read_len;
        seq_pos.num = -1;
        seqValid    = 1;
        gop_p       = 0; 
            
        if ( (i+4) < end )
        {
          int width, height;
            
          width   = pes_vs_buf[i+1];
          width <<= 4;
          width  |= (pes_vs_buf[i+2]>>4);
                
          height   = (pes_vs_buf[i+2] & 0x0F);
          height <<= 8;
          height  |= pes_vs_buf[i+3]; 
          
          //-- handle more than one sequence -- 
          if ( (width != stream->width) || (height != stream->height) )      
          {
            if (stream->width != -1) 
            {
              //fprintf(stderr, "WARNING: T_FRAME_SRV - new sequence detected after frame(%d):\n"\
              //                "         size changed from (%dx%d) to (%dx%d) !\n",
              //        stream->pict_cnt, stream->width, stream->height, width, height);
              changeIndexFile();
            }
            
            stream->width      = width;
            stream->height     = height;
            stream->aspect     = (pes_vs_buf[i+4]>>4);
            stream->frame_rate = (pes_vs_buf[i+4] & 0x0F);
                      
          }
        }
        break;

    case GROUP_START_CODE:
        stream->gop_cnt++;
              
        gop_pos.pos = seek_pos;
	gop_pos.len = (uint16_t)read_len;
        gop_pos.num = 0;
        gopValid    = 1;
        gop_p       = 0;
              
        //if ( ((i+4) < end) && stream->closed_gop )
        //  stream->closed_gop = ((pes_vs_buf[i+4] & 0x40) == 0x40);
        break;
            
    case PICT_START_CODE:
        stream->pict_cnt++;
              
        if (seqValid)
        {
          //fprintf(stderr, "#%d: seq -> num (%d)\n", stream->pict_cnt, seq_pos.num);
          fwrite(&seq_pos, sizeof(T_POS_REC), 1, pf);
        }  
        else if (gopValid)
        {
          //fprintf(stderr, "#%d: gop -> num (%d)\n", stream->pict_cnt, gop_pos.num);
          fwrite(&gop_pos, sizeof(T_POS_REC), 1, pf);
        }
        else
        {
          gop_p++;
          pict_pos.pos = seek_pos;
          pict_pos.len = (uint16_t)read_len; 
          pict_pos.num = gop_p;
          
          //-- check for real open GOPs --
          if ( ((i+2) < end) && (gop_p == 1) )    
          {
            uint16_t tr = (pes_vs_buf[i+1] << 2) | (pes_vs_buf[i+2] >> 6);
            if ( !tr && (stream->open_gop<127) ) stream->open_gop++; 
          }
              
          fwrite(&pict_pos, sizeof(T_POS_REC), 1, pf);
        }  
        seqValid = 0;
        gopValid = 0;
 
        if ( (stream->pict_cnt % 100) == 0 )
        {
          if (userProgress)
            userProgress->show(stream->pict_cnt);
          else 
            fprintf(stderr, "Picture #(%d)\r", stream->pict_cnt);
        }
        break;
    
    case SEQ_END_CODE:
        fprintf(stderr, "\nend of sequence detected\n");
    default:
        break;  
  }
}

int T_FRAME_SRV::loadIndex(void)
{
  struct stat stat_buf;
  char        *p;
  int         rc  = 1;
  int         l   = strlen(fname);
  char        *fn = new char[l+9];
  int         isIdx;
  
check_index:  
  isIdx = 0;
  
  //-- fname must be untouched     --
  //-- therefore working on a copy --
  if (fn) 
  {
    strcpy(fn, fname);
  }
  else
  {
    fprintf(stderr, "ERROR: cannot load index for (%s)\n", fname); 
    return 0;
  }
      
  if ( ((p=strstr(fn, EXT_INDEX)) != NULL) && (p[4] == '\0') )
  {
    //-- is real index file --
    isIdx = 1; 
  }
  else
  {
    //-- is media file -- 
    strcat(fn, EXT_INDEX);
    
    //-- build index, if missing --
    if (stat(fn, &stat_buf) == -1) rc = createIndex();
  }
  //-- at this point fn is everytime a filename with extension EXT_INDEX --

  if (rc)  
  { 
    //-- open index file --
    pf = fopen(fn, "rb");
    
    if (!pf)
    {
      fprintf(stderr, "ERROR: unable to open index (%s)\n", fn);
      delete fn;
      cleanup();
      return 0;
    }
    
    //-- On real index file cut complete extension(s) --
    if (isIdx)
    { 
      l     -= 4;
      fn[l]  = '\0';
    
      //-- test for numeric index count "-#99" --
      //-- and cut it too, if available.       --
      if ( l > 4 )
      { 
        p = &fn[l-4];
        if ( !strncmp(p,"-#", 2) && isdigit(p[2]) && isdigit(p[3]) ) *p = '\0';
      }
    }
    else
      strcpy(fn, fname);
    //-- at this point fn is everytime the media-filename --
    
    //-- check index version  --
    //--------------------------
    l = fread(stream, 1, MAX_MAGIC_LEN, pf); 
    if ( l < MAX_MAGIC_LEN || strncmp(stream->magic, LVE_MAGIC, MAX_MAGIC_LEN) )
    {
      fprintf(stderr, "WARNING: (%s.idx) is no valid index file -> recreating\n", fname);
      cleanup();
      
      memcpy(stream->magic, LVE_MAGIC, MAX_MAGIC_LEN);
      rc = createIndex();
      
      goto check_index;
    }
    
    //-- process index data from file --
    //----------------------------------
    fseeko(pf, 0, SEEK_SET);
    l = fread(stream, 1, sizeof(T_STREAM), pf);
    if ( l == sizeof(T_STREAM) )   
    {
      fprintf(stderr, "\nLoading index ...\n");
      ls_pict = new T_POS_REC[stream->pict_cnt]; 
      if (ls_pict) fread(ls_pict, sizeof(T_POS_REC), stream->pict_cnt, pf);
      fprintf(stderr, "... (%d) picture records\n", stream->pict_cnt); 

      cleanup();

      //-- open media file --
      //---------------------    
      vf = fopen(fn, "rb");
      if (!vf) 
      {
        fprintf(stderr, "ERROR: opening media (%s)\n", fn);
        delete fn;
        return 0;
      }
      
      //-- using 2nd. file descriptor to scan audio --
      //-- if not availbale -> disable audio.       --
      af = fopen(fn, "rb");
      if (!af) 
      {
        stream->as_cnt = 0;
        fprintf(stderr, "WARNING: creating 2nd. file descriptor for audio\n");
      }
      delete fn;
      return 1;
    }
    else
      fprintf(stderr, "ERROR: idx-file for (%s) to short (%d)\n", fname, l);
  }
  else
    fprintf(stderr, "ERROR: cannot load index (%s)\n", fn); 
  
  delete fn;
  cleanup();
  return 0;  
}

void T_FRAME_SRV::setMaxIndex(int n)
{
  maxIndex = (uint32_t)n;
}

int T_FRAME_SRV::firstSequence(int onlySeq)
{
  if (onlySeq)
  {
    int s = 0;
    while ( ls_pict[s++].num >= 0)
    {
      if (s > stream->pict_cnt)
      {
        fprintf(stderr, "WARNING: T_FRAME_SRV - there is no sequence header !\n");
        return nextSequence(0);
      }    
    }
    return (s-1);  
  }
  else
    return nextSequence(0);
}

int T_FRAME_SRV::prevSequence(int p)
{
  int s;
  
  if (p >= stream->pict_cnt) p = stream->pict_cnt-1;
  
  //-- sequence record hit directly --
  if (ls_pict[p].num <= 0) return p;
  
  //-- go back amount of "num" records --
  s = p - ls_pict[p].num;
  if (s < 0) s = 0;

  //-- sequence or alternativley GOP --  
  if ( (ls_pict[s].num != -1) && (ls_pict[s].num != 0) )
  {
    //fprintf(stderr, "WARNING: T_FRAME_SRV - there is no previous sequence/gop picture !\n");
    //-- force first available sequence instead --
    return firstSequence(1);
  }
    
  return s;  
}

int T_FRAME_SRV::nextSequence(int p)
{
  int s;
  
  if (p >= stream->pict_cnt) p = stream->pict_cnt - 1;
  
  //-- sequence record hit directly --
  if (ls_pict[p].num <= 0) return p;
  
  //-- sequence or alternativley GOP --  
  s = p;
  while ( ls_pict[s++].num > 0)
  {
    if (s > stream->pict_cnt)
    {
      //fprintf(stderr, "WARNING: T_FRAME_SRV - there is no next sequence/gop picture !\n");
      return stream->pict_cnt;
    }    
  }
    
  return (s-1);  
}

int T_FRAME_SRV::isSequence(int p)
{
  return (ls_pict[p].num <= 0);
}

int T_FRAME_SRV::isGOP(int p)
{
  return (ls_pict[p].num == 0);
}

off_t T_FRAME_SRV::filePos(int p)
{
  return (ls_pict[p].pos);
}

int T_FRAME_SRV::nFrames(void)
{
  return stream->pict_cnt;
}
   
double T_FRAME_SRV::videoRate(void)
{
  static const double ratetab[8] =
  { 
    30000.0/1001.0,  
    24.0,                
    25.0,
    30000.0/1001.0,
    30.0,
    50.0,
    60000.0/1001.0,
    60.0
  };
  int c = stream->frame_rate - 1;
   
  if ( (c < 0) || (c > 7) )
    return 25.0;             // default if unknown code;
  else
    return ratetab[c];   
}    
    
int T_FRAME_SRV::videoWidth(void)
{
  return stream->width;
}

int T_FRAME_SRV::videoHeight(void)
{
  return stream->height;
}

int T_FRAME_SRV::videoAspect(void)
{
  return stream->aspect;
}

char T_FRAME_SRV::closedGOP(void)
{
  return (stream->open_gop==0);
}

char *T_FRAME_SRV::getFileName(void)
{
  return fname;
}    

int T_FRAME_SRV::getTotalAudioStreams(void)
{
  return stream->n_as;
}

uint8_t T_FRAME_SRV::setAudioTrack(int track)
{
  if ( (track < 0) || (track >= MAX_ATRACKS) ) return A_TYPE_NONE;

  //-- set stream id if track available --
  if (stream->audio_tracks[track].id != 0) 
  {  
    stream->as_id  = stream->audio_tracks[track].id; 
    stream->as_pid = stream->audio_tracks[track].pid;
  }
  return stream->as_id;
}

uint8_t  T_FRAME_SRV::getAudioTrack(int track)
{
  if ( stream->n_as == 0 || track < 0 || track >= MAX_ATRACKS ) 
    return A_TYPE_NONE;
  else
    return stream->audio_tracks[track].id;
}

uint8_t  T_FRAME_SRV::getAudioType(int track)
{
  uint8_t t;
  
  if (stream->n_as == 0) return A_TYPE_NONE;
  
  t = (setAudioTrack(track) & 0xF0);
  return t;  
}

int T_FRAME_SRV::sampleRate(int track)
{
  if ( (track < MAX_ATRACKS) && (stream->audio_tracks[track].id != 0x00) ) 
    return stream->audio_tracks[track].sample_rate;
  else
  {
    fprintf(stderr, "WARN: unknown sample rate, using 48kHz\n");
    return 48000;
  }  
}

int T_FRAME_SRV::bitRate(int track)
{
  if ( (track < MAX_ATRACKS) && (stream->audio_tracks[track].id != 0x00) ) 
    return stream->audio_tracks[track].bit_rate;
  else
    return 0;
}

void T_FRAME_SRV::updSampleRate(int track, int sr)
{
  if (track<MAX_ATRACKS) stream->audio_tracks[track].sample_rate = sr;
}

void T_FRAME_SRV::patchSEQ(uint8_t *chunk_buf, T_VIDEO_PARMS *vp)
{
  // apsect => (pes_vs_buf[3]>>4);
  if (vp->aspect)
  {
    uint8_t h = chunk_buf[3] & 0x0F;
    chunk_buf[3] = (h | vp->aspect);
  }
  
  //-- patch frame rate code if available --
  if (vp->rate_code)
  {
    chunk_buf[3] &= 0xF0;
    chunk_buf[3] |= vp->rate_code;

/*    
    chunk_buf[4]  = 0xFF;
    chunk_buf[5]  = 0xFF;
    chunk_buf[6] |= 0xC0;
*/    
  }
}

int T_FRAME_SRV::findStartCode(uint8_t *chunk_buf, int len, uint32_t sc)
{
  uint32_t code = 0xFFFFFFFF; 
  
  for (int i=0; i<len; i++)
  {
    code = (code << 8) | chunk_buf[i];
    if ( (code & CODE_PREFIX_MASK) == CODE_PREFIX )
    {
      if (code == sc) return (i-3);
    }
  }  
  return -1;
}


int T_FRAME_SRV::nextStartCode(uint8_t *chunk_buf, int len)
{
  uint32_t code = 0xFFFFFFFF; 
  
  for (int i=4; i<len; i++)
  {
    code = (code << 8) | chunk_buf[i];
    if ( (code & CODE_PREFIX_MASK) == CODE_PREFIX )return (i-3);
  }  
  return -1;
}

void T_FRAME_SRV::patchGOP(uint8_t *chunk_buf, T_VIDEO_PARMS *vp)
{
  int      h, m, s;
  uint32_t n = vp->nFrames;
  double   f = vp->frame_rate * 3600;

  //-- calculate hours ... --  
  h  = (int)(n/f);
  n -= (uint32_t)(h*f);
  f /= 60;
  
  //-- ... minutes ... --
  m  = (int)(n/f);
  n -= (uint32_t)(m*f);
  f /= 60;
  
  //-- ... seconds and pictures rests --
  s  = (int)(n/f);
  n -= (uint32_t)(s*f);

  //-- patch in buffer now -- 
  chunk_buf[0] &= 0x80;
  chunk_buf[0] |= (uint8_t)(h<<2);
  chunk_buf[0] |= (uint8_t)(m>>4);
  
  chunk_buf[1] &= 0x08;
  chunk_buf[1] |= (uint8_t)(m<<4);
  chunk_buf[1] |= (uint8_t)(s>>3);
  
  chunk_buf[2]  = 0;
  chunk_buf[2] |= (uint8_t)(s<<5);
  chunk_buf[2] |= (uint8_t)(n>>1);
  
  chunk_buf[3] &= 0x7F;
  chunk_buf[3] |= (uint8_t)(n<<7);
  
  if (vp->set_closed_gop) 
  {
    chunk_buf[3] |= 0x40;
    vp->set_closed_gop = 0;
  }  
  if (vp->set_broken_link) 
  {
    chunk_buf[3] |= 0x20;
    vp->set_broken_link = 0;
  }  
}

void T_FRAME_SRV::patchPIC(uint8_t *chunk_buf, T_VIDEO_PARMS *vp)
{
  chunk_buf[0] &= 0xFC;
  chunk_buf[1] &= 0x3F;
  
  //chunk_buf[0] |= (vp->temporal_ref>>2);
  //chunk_buf[1] |= (vp->temporal_ref<<6);
}

T_PIC_HDR *T_FRAME_SRV::getPicHdr(uint8_t *chunk_buf, int len = 0)
{
  static  T_PIC_HDR ph;
  int32_t ofs  = 0;
  uint8_t *buf = chunk_buf;

  ph.gop_valid   = 0;
  ph.pict_struct = 3;  // default frame 
  
  //-- only picture data available ... --
  //-------------------------------------
  if (buf[3] == 0x00) goto lb_picture;
  /*
  {
    ph.temporal_ref = (buf[4] << 2) | (buf[5] >> 6);
    ph.type         = (buf[5] >> 3) & 0x07; 
 
    return &ph;
  }
  */
  
  //-- ... or check first for additional GOP data ... --
  //----------------------------------------------------
  if ( (len >0) && (ofs = findStartCode(buf, len, GROUP_START_CODE)) != -1 )
  {
    buf += ofs;
    len -= ofs;
    
    ph.gop_valid   = 1;
    ph.closed_gop  = buf[7] & 0x40;
    ph.broken_link = buf[7] & 0x20;
  }
   
  //-- ... last search for picture (and extension) headers --
  //---------------------------------------------------------  
  if ( (len>0) && (ofs = findStartCode(buf, len, PICT_START_CODE)) != -1)
  {
    buf += ofs;
    len -= ofs;
   
lb_picture:    
    ph.temporal_ref = (buf[4] << 2) | (buf[5] >> 6);
    ph.type         = (buf[5] >> 3) & 0x07; 
 
    //fprintf(stderr, "+++ p-type (%d) +++\n", ph.type);
    // if (stream->mpeg2_flag == 0) return &ph; TODO
 
    if ( (len>0) && (ofs = findStartCode(buf, len, EXT_START_CODE)) != -1)
    {
      buf += ofs;
    
      //-- if it's really a "picture coding extension" ... --
      if ( (buf[4] & 0xF0) == 0x80 )
      {
        //-- ... we can check the bits -- 
        ph.pict_struct = (buf[6] & 0x03);
      } 
    } 
    return &ph;
  }
  
  return NULL;
}

//==========================
//== class T_FRAME_SRV_PS ==
//==========================
static char *id_str_ps = "Program stream (PS)";
T_FRAME_SRV_PS::T_FRAME_SRV_PS(int t, char *fname, uint8_t vid) 
  : T_FRAME_SRV(t, fname, vid)
{
  id_str     = id_str_ps;
  paudio_buf = audio_buf;
  ab_level   = 0;
  for (int i=0; i<MAX_ATRACKS; i++)
  {
    _stream.audio_tracks[i].id          = 0x00;
    _stream.audio_tracks[i].sample_rate = 0; //44100;
    _stream.audio_tracks[i].bit_rate    = 0;
    _stream.audio_tracks[i].pid         = 0;
  }
  memset(atracks, 0, sizeof(atracks) );
  memset(detected_sr, 0, sizeof(detected_sr));
  memset(detected_br, 0, sizeof(detected_br));
}

void T_FRAME_SRV_PS::changeIndexFile(void)
{
    static int idx_cnt = 1;
    
    int l = strlen(fname);
    int j = 0;
    int t = 0;
    
    fprintf(stderr,
      "  (%d) video stream (0xE%x) records, type (%s)\n", stream->vs_cnt, 
      stream->vs_id, stream->mpeg2_flag? "mpeg2":"mpeg1");
    fprintf(stderr,
      "    (%d) sequence records with\n", 
      stream->seq_cnt);
    fprintf(stderr,
      "      frame size (%dx%d)\n", 
      stream->width, stream->height);
    fprintf(stderr,
      "      aspect code (%d)\n", 
      stream->aspect);
    fprintf(stderr,
      "      frame rate code (%d)\n", 
      stream->frame_rate);
    fprintf(stderr,
      "      (%d) GOP records [# of open>=%d]\n", 
      stream->gop_cnt, stream->open_gop);
    fprintf(stderr,
      "      (%d) picture records\n", 
      stream->pict_cnt);

    for (j=0; j<15; j++)
    {
      if (atracks[j] > 0)
      {
        stream->audio_tracks[t].id = j+A_TYPE_A52_SUB;
        stream->audio_tracks[t].sample_rate = detected_sr[j];
        stream->audio_tracks[t++].bit_rate = detected_br[j]; 
        fprintf(stderr,
          "  (%d) audio stream (0x%02x) records, (%d) Hz, (%d) bps\n", 
          atracks[j], j+0x80, detected_sr[j], detected_br[j]);
      }
    }  
      
    for (j=15; j<30; j++)
    {
      if (atracks[j] > 0) 
      {
        stream->audio_tracks[t].id = (j-15)+A_TYPE_PCM_SUB;
        stream->audio_tracks[t].sample_rate = detected_sr[j];  
        stream->audio_tracks[t++].bit_rate = detected_br[j];  
        fprintf(stderr,
          "  (%d) audio stream (0x%02x) records, (%d) Hz, (%d) bps\n", 
          atracks[j], (j-15)+0xA0, detected_sr[j], detected_br[j]);
      }
    }
      
    for (j=30; j<45; j++)
    {
      if (atracks[j] > 0) 
      {
        stream->audio_tracks[t].id = (j-30) + 0xC0; 
        stream->audio_tracks[t].sample_rate = detected_sr[j]; 
        stream->audio_tracks[t++].bit_rate = detected_br[j]; 
        fprintf(stderr,
          "  (%d) audio stream (0x%02x) records, (%d) Hz, (%d) bps\n", 
          atracks[j], (j-30)+0xC0, detected_sr[j], detected_br[j]);
      }
    }  
 
    stream->n_vs = 1;
    stream->n_as = t;

    fprintf(stderr,
      "  total audio/video streams: (%d/%d) \n", 
      stream->n_as, stream->n_vs);
    
    //-- update common stream record --
    //---------------------------------
    fseeko(pf, 0, SEEK_SET);
    fwrite(stream, sizeof(T_STREAM), 1, pf);

    if (pf) fclose(pf);
    
    //-- get next index file --
    //-------------------------
    sprintf(fname, "%s-#%02d.idx", fname, idx_cnt++); 
    pf       = fopen(fname, "w+b");
    fname[l] = '\0';
    
    if (pf)
    {
      fprintf(stderr, 
        "\n--- using next index file (#%d) ---\n", 
        idx_cnt-1);
    
      //-- start again --
      stream->pict_cnt   = 0;
      stream->open_gop   = 0;
      stream->n_as       = 0;
      stream->n_vs       = 0;
      
      //-- TODO --
      // reset atracks etc. 
      
      //-- write initial record with common stream data first --
      fwrite(stream, sizeof(T_STREAM), 1, pf);
      
    }
    else
    {
      fprintf(stderr, "ERROR: unable to create PS index (#%d) for %s\n", 
              fname, idx_cnt);
    } 
}

int T_FRAME_SRV_PS::createIndexInit(void)
{
  //-- setup audio stream scan parameters --
  stream->atype = 0; // not used 
  stream->as_id = 0; // not used
  g_pes_pos     = 0;
  g_pes_len     = 0;
  g_pes_pos_old = 0;
  g_pes_len_old = 0;
  
  seqValid = 0;
  gopValid = 0;
  gop_p    = 0;

  if ( (vf = fopen(fname,"rb")) != NULL)
  {
    int l = strlen(fname);
    
    strcat(fname, EXT_INDEX); 
    pf = fopen(fname, "w+b");
    fname[l] = '\0';           // restore fname
    
    if (!pf)
    {
      fprintf(stderr, "ERROR: unable to create PS index (#0) for %s\n", fname);
      cleanup();
      return 0;
    } 
    
    fprintf(stderr,
      "\n--- using first index file (#0) ---\n");
    
    //-- write initial record with common stream data first --
    //--------------------------------------------------------
    fwrite(stream, sizeof(T_STREAM), 1, pf);
  } 
  else
  {
    fprintf(stderr, "ERROR: unable to open (PS) media (%s)\n", fname);
    cleanup();
    return 0;
  }
  return 1;    
}  

void T_FRAME_SRV_PS::createIndexExit(void)
{
  //-- summary --
  fprintf(stderr, "\n%s scan done:\n", id_str);
  fprintf(stderr, "total Packs: %d\n", stream->pack_cnt);
  fprintf(stderr, "  (%d) system records\n", stream->sys_cnt);
  
  if (stream->vs_pid > 0)
  {
    fprintf(stderr,
      "  (%d) video pid (0x%03x), id (0xE%x) packs, type (%s)\n", stream->vs_cnt, 
      stream->vs_pid, stream->vs_id, stream->mpeg2_flag? "mpeg2":"mpeg1");
  }
  else
  {
    fprintf(stderr, 
      "  (%d) video stream (0xE%x) records, type (%s)\n", stream->vs_cnt, 
      stream->vs_id, stream->mpeg2_flag? "mpeg2":"mpeg1");
  
  }
  
  fprintf(stderr,
    "    (%d) sequence records with\n", 
    stream->seq_cnt);
  fprintf(stderr,
    "      frame size (%dx%d)\n", 
    stream->width, stream->height);
  fprintf(stderr,
    "      aspect code (%d)\n", 
    stream->aspect);
  fprintf(stderr,
    "      frame rate code (%d)\n", 
    stream->frame_rate);
  fprintf(stderr,
    "      (%d) GOP records [# of open=%d]\n", 
    stream->gop_cnt, stream->open_gop);
  fprintf(stderr,
    "      (%d) picture records\n", 
    stream->pict_cnt);

  int j = 0;
  int t = 0;
    
  for (j=0; j<15; j++)
  {
      if (atracks[j]>10 && detected_sr[j]>0 && detected_br[j]>0)
      {
        stream->audio_tracks[t].id = j+A_TYPE_A52_SUB;
        stream->audio_tracks[t].sample_rate = detected_sr[j]; 
        stream->audio_tracks[t++].bit_rate = detected_br[j]; 
        fprintf(stderr,
          "  (%d) audio stream (0x%02x) records, (%d) Hz, (%d) bps\n", 
          atracks[j], j+0x80, detected_sr[j], detected_br[j]);
      }
  }  
      
  for (j=15; j<30; j++)
  {
      if (atracks[j]>10 && detected_sr[j]>0 && detected_br[j]>0) 
      {
        stream->audio_tracks[t].id = (j-15)+A_TYPE_PCM_SUB;
        stream->audio_tracks[t].sample_rate = detected_sr[j];
        stream->audio_tracks[t++].bit_rate = detected_br[j];
        fprintf(stderr,
          "  (%d) audio stream (0x%02x) records, (%d) Hz, (%d) bps\n", 
          atracks[j], (j-15)+0xA0, detected_sr[j], detected_br[j]);
      }
  }
      
  for (j=30; j<45; j++)
  {
      if (atracks[j]>10 && detected_sr[j]>0 && detected_br[j]>0) 
      {
        stream->audio_tracks[t].id = (j-30) + 0xC0; 
        stream->audio_tracks[t].sample_rate = detected_sr[j]; 
        stream->audio_tracks[t++].bit_rate = detected_br[j]; 
        fprintf(stderr,
          "  (%d) audio stream (0x%02x) records, (%d) Hz, (%d) bps\n", 
          atracks[j], (j-30)+0xC0, detected_sr[j], detected_br[j]);
      }
  }  
 
  stream->n_vs = 1;
  stream->n_as = t;

  fprintf(stderr,
    "  total audio/video streams: (%d/%d) \n", 
    stream->n_as, stream->n_vs);
    
  //-- update common stream record --
  //---------------------------------
  fseeko(pf, 0, SEEK_SET);
  fwrite(stream, sizeof(T_STREAM), 1, pf);

  cleanup();
}    

int T_FRAME_SRV_PS::createIndex(void)
{
  T_CODE code = (CODE_PREFIX_MASK >> 8);
  int    ch;

  if (createIndexInit())
  {
    //-- full scan loop --
    //--------------------
    while ( (ch = fgetc(vf)) != EOF && (stream->pict_cnt < maxIndex) )
    {
      code = (code << 8) | ch;
     
      if ( (code & CODE_PREFIX_MASK) == CODE_PREFIX ) switch (code)
      {
          case PACK_START_CODE:
            ch   = fgetc(vf);
            code = (code << 8) | ch;
              
            if (stream->mpeg2_flag == -1)
              stream->mpeg2_flag = ((ch >> 6) == 1);
            
            stream->pack_cnt++;
            break;
       
          default:
            checkPESdata(code);
            break;   
      }
    }
    
    createIndexExit();
    
    return 1;
  }
  return 0;  
}

int T_FRAME_SRV_PS::syncAStream(uint8_t id, uint8_t *pbuf, int32_t len, 
                                int *detected_sr, int *detected_br)
{
  static int sr[] = {44100, 48000, 32000, 0};
  static int br[] = {0, 32, 48, 56, 64, 80, 96, 112, 128, 160, 192, 224, 256, 320, 384, 0 };

  if (id == 0xC0)
  {
    uint16_t a_sync = 0;
    int      i;
    
    for (i=0; i<len-2; i++)
    {
      a_sync = (a_sync<<8) | pbuf[i];
      if ( a_sync == 0x0FFF || ((a_sync & 0xFFF0) == 0xFFF0) ) 
      {
        uint32_t parms = ( (pbuf[i]<<16) | (pbuf[i+1]<<8) | pbuf[i+2] );
        if (pbuf[i-1] == 0xFF)
          parms <<= 4;
        else
          parms <<= 8;
    
        *detected_sr = sr[(parms & 0xC000) >> 14];
        *detected_br = br[(parms & 0xF0000) >> 16] * 1000;
            
        return (i-1);
      }
    }
  }
  else
  {
    int fl, i;
      
    for (i=0; i<len-7; i++)
    {
      if (a52_syncinfo(pbuf+i, &fl, detected_sr, detected_br))
        return (i);
    }
  }
    
  return -1;
}

void T_FRAME_SRV_PS::checkAStream(int32_t ofs, int32_t len, uint8_t chan)
{
  if (chan >= 45) return;

  //-- mpeg1 layer 2 audio --
  //-------------------------
  if (chan >= 30)
  {
    syncAStream(A_TYPE_MPEG1_L2, &pes_as_buf[ofs], len, 
                &detected_sr[chan], &detected_br[chan] );
  }
  //-- AC3 -> using a52_sync() --
  //-----------------------------
  else
  { 
    syncAStream(A_TYPE_A52_SUB, &pes_as_buf[ofs], len, 
                &detected_sr[chan], &detected_br[chan] );
  }
}

void T_FRAME_SRV_PS::checkVStream(int32_t ofs, int32_t len)
{
  static T_CODE pes_code = 0xFFFFFF;    // ! must be static
  
  int32_t  end      = ofs + len;
  int32_t  read_len = len;
  int32_t  i;
  off_t    seek_pos;
  
  for (i=ofs; i<end; i++)
  {
    pes_code = (pes_code << 8) | pes_vs_buf[i];
    
    if ( (pes_code & CODE_PREFIX_MASK) == CODE_PREFIX)
    {
      int d = ofs-i+3;
      
      if ( d > 0 ) 
      {
        seek_pos = g_pes_pos_old + g_pes_len_old - d; 
        read_len = d; 
      }  
      else
      {
        seek_pos = g_pes_pos + (i-3);
        read_len = len + d;
      }  

      processVCode(pes_code, seek_pos, read_len, i, end);
    }
  }
}          

int T_FRAME_SRV_PS::checkSYSheader(void)          
{
  if (!stream->n_vs)
  {
    //-- how much audio/video streams ? --
    stream->n_as = (pes_vs_buf[3] & 0xFC) >> 2;
    stream->n_vs = (pes_vs_buf[4] & 0x1F);
  }
  return 1;
}

int T_FRAME_SRV_PS::checkPESheader(uint8_t *pes_buf)          
{
  static int mpeg1_skip_table[16] = 
  {
    0, 0, 4, 9, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0
  };
  int hdr_len;
  
  //-- MPEG2 ... --
  if ( (pes_buf[0] & 0xC0) == 0x80 )
  {
    hdr_len = 9 + pes_buf[2];
  }
  //-- ... or MPEG1 --
  else
  {
    int i;
            
    hdr_len = 7;
    i       = 0;
    while (pes_buf[i++] == 0xFF)
    { 
      hdr_len++;
      if (hdr_len == 23) 
      {
        fprintf(stderr, "too much stuffing\n");
        break;
      }  
    }
    i--;
              
    if ( (pes_buf[i] & 0xC0) == 0x40 ) 
    {
      hdr_len += 2;
      i       += 2;
    }  
    hdr_len += mpeg1_skip_table[pes_buf[i] >> 4];  
  }
  return hdr_len;
}  

void T_FRAME_SRV_PS::checkPESdata(T_CODE code)
{
  int32_t pes_len, hdr_len, data_len;
  int32_t data_ofs;
  
  //-- match video stream ... --
  //----------------------------
  if ( code == (VSTREAM_MASK | stream->vs_id) ) 
  {
    stream->vs_cnt++;
    
    g_pes_pos_old = g_pes_pos;
    g_pes_len_old = g_pes_len;
    
    g_pes_len = fgetc(vf);
    g_pes_len = (g_pes_len << 8) | fgetc(vf);
    g_pes_pos = ftello(vf);

    //-- read pes pack now --    
    g_pes_len = fread(pes_vs_buf, 1, g_pes_len, vf);
    hdr_len   = checkPESheader(pes_vs_buf);
          
    data_ofs  = hdr_len - 6;
    data_len  = g_pes_len - data_ofs;
    
    //-- scan for startcodes in video stream -- 
    checkVStream(data_ofs, data_len);
  }
  //-- ... or plain audio stream ... --
  //-----------------------------------
  else if ( (code & 0xF0) == A_TYPE_MPEG1_L2 )        
  {
    uint8_t n;
   
    n = 30 + (code & 0x0F);
    if (n<45) atracks[n]++;
    
    pes_len = fgetc(vf);
    pes_len = (pes_len << 8) | fgetc(vf);
    
    if (!detected_sr[n])
    {
      pes_len = fread(pes_as_buf, 1, pes_len, vf);
      hdr_len = checkPESheader(pes_as_buf);
          
      data_ofs = hdr_len - 6;
      data_len = pes_len - data_ofs;

      checkAStream(data_ofs, data_len, n);
    }
    else
      fseeko(vf, pes_len, SEEK_CUR);
  }
  //-- ... or other PES packet including sub audio --
  //-------------------------------------------------
  else 
  {
    pes_len = fgetc(vf);
    pes_len = (pes_len << 8) | fgetc(vf);
      
    switch (code) 
    {
      case SYS_START_CODE:
        stream->sys_cnt++;
        pes_len = fread(pes_vs_buf, 1, pes_len, vf);
        checkSYSheader();     
        break;
      
      //-- check audio subs --
      case PRIV1_START_CODE:
        pes_len = fread(pes_as_buf, 1, pes_len, vf); 
        hdr_len = checkPESheader(pes_as_buf);
            
        data_ofs = hdr_len - 6;
        data_len = pes_len - data_ofs;

        if (data_ofs > 0)
        { 
          uint8_t n;
          uint8_t sub = pes_as_buf[data_ofs];
          
          //-- DD5.1 or DTS
          if ( (sub & 0xF0) == A_TYPE_A52_SUB )
          {
            n = (sub & 0x0F);
            if (n<45) atracks[n]++;
          
            if (!detected_sr[n])
              checkAStream(data_ofs+4, data_len-4, n);
          }
          //-- PCM --
          else if ( (sub & 0xF0) == A_TYPE_PCM_SUB )
          {
            
            n = 15 + (sub & 0x0F);
            if (n<45) atracks[n]++;
            
            if (!detected_sr[n])
              checkAStream(data_ofs+4, data_len-4, n);
          }
        }
        break;
    
      default:
        //-- skip pes data -- 
        fseeko(vf, pes_len, SEEK_CUR);
        break;
    }
  } 
}

#define RD_PES_ALL     -1
#define RD_PES_PARTIAL  0

int T_FRAME_SRV_PS::readPESvideo(T_CODE code, uint8_t *pbuf,  int32_t this_len)
{
  int32_t  pes_len, hdr_len, data_len;
  int32_t  ofs, data_ofs;

  //-- 1st. part can be read directly --
  //------------------------------------
  if (this_len != RD_PES_ALL)
  {
    //int len = fread(pes_vs_buf, 1, this_len, vf);
    //memcpy(pbuf, pes_vs_buf, this_len);
    //return len;
    return fread(pbuf, 1, this_len, vf);
  }    
  //-- match video stream ... -
  //----------------------------
  else if ( code == (VSTREAM_MASK | (T_CODE)stream->vs_id) ) 
  {
    pes_len = fgetc(vf);
    pes_len = (pes_len << 8) | fgetc(vf);

    pes_len = fread(pes_vs_buf, 1, pes_len, vf);
    hdr_len = checkPESheader(pes_vs_buf);
          
    data_ofs = hdr_len - 6;
    data_len = pes_len - data_ofs;

    memcpy(pbuf, &pes_vs_buf[data_ofs], data_len);
  
    return data_len;
  }
  else
  { 
    pes_len = fgetc(vf);
    pes_len = (pes_len << 8) | fgetc(vf);
    fseeko(vf, pes_len, SEEK_CUR);
    
    return 0;  // is no video !
  }   
}

int T_FRAME_SRV_PS::readPESaudio(T_CODE code, uint8_t *pbuf)
{
  int32_t pes_len, hdr_len, data_len;
  int32_t ofs, data_ofs;

  pes_len = fgetc(af);
  pes_len = (pes_len << 8) | fgetc(af);

  //-- need no audio at this time ? --
  if ( ab_level >= AB_MIN_LEVEL ) 
  {
    fseeko(af, pes_len, SEEK_CUR);
    return 0;
  }   

  //-- plain audio stream ... --
  //----------------------------
  if ( code == (ASTREAM_MASK | stream->as_id) ) 
  {
    int tmp_len = fread(pes_as_buf, 1, pes_len, af);
    if (tmp_len == pes_len)
    {
      //pes_len  = tmp_len;
      hdr_len  = checkPESheader(pes_as_buf);
      data_ofs = hdr_len - 6;
      data_len = pes_len - data_ofs;

      //-- check buffer range and        --
      //-- transfer data to audio buffer --
      //if ( (data_len+data_ofs) > MAX_PES_AS_LEN ) 
      //  data_len = MAX_PES_AS_LEN - data_ofs;
      memcpy(paudio_buf, &pes_as_buf[data_ofs], data_len);
      paudio_buf += data_len;
          
      ab_level = (paudio_buf - audio_buf);
    }
  }
  //-- ... or other PES packet including sub audio --
  //-------------------------------------------------
  else if (code == PRIV1_START_CODE)
  {
    //-- read PES packet ready to extract data --
    pes_len  = fread(pes_as_buf, 1, pes_len, af);
    hdr_len  = checkPESheader(pes_as_buf);
    data_ofs = hdr_len - 6;
    data_len = pes_len - data_ofs;
        
    //-- if track-id is desired track ... --
    if ( (data_ofs > 0) && (pes_as_buf[data_ofs] == (stream->as_id & 0xFF)) )
    {
      /*
      uint16_t sync_o = ((uint16_t)pes_as_buf[data_ofs+2])<<8 | pes_as_buf[data_ofs+3];
      fprintf(stderr, "*** (%d) %02x %02x ***\n",
                sync_o,
                pes_as_buf[data_ofs+sync_o+3], 
                pes_as_buf[data_ofs+sync_o+4]); 
      */
              
      data_ofs += 4;
      data_len -= 4;
          
      //-- transfer data to audio buffer --
      memcpy(paudio_buf, &pes_as_buf[data_ofs], data_len);
      paudio_buf += data_len;
          
      ab_level = (paudio_buf - audio_buf);
    }
  }
  else
    fseeko(af, pes_len, SEEK_CUR);

  return 0;
}

uint8_t *T_FRAME_SRV_PS::readAStream(int *buf_len)
{
  *buf_len = ab_level;
  return audio_buf; 
}

void T_FRAME_SRV_PS::adjustAStream(int bytes)
{
  if (bytes<=ab_level)
  {
    ab_level = ab_level - bytes;

    memcpy(audio_buf, audio_buf+bytes, ab_level);
    paudio_buf = audio_buf + ab_level;
  }
  else
    ab_level = 0;
}


int T_FRAME_SRV_PS::readFrame(uint8_t *buf, int p)
{
  uint8_t *pbuf = buf;

  T_CODE  code = CODE_PREFIX_MASK;
  int32_t ch, done, len;
  int32_t total_vlen; 
  int32_t this_len, next_len;
  off_t   this_pos, next_pos;

  this_pos = ls_pict[p].pos;
  this_len = ls_pict[p].len;
  
  next_pos = ls_pict[p+1].pos;
  next_len = ls_pict[p+1].len;
  
  total_vlen = 0;

  //fprintf(stderr, "picture(%d) - this (%d), next (%d)\n", 
  //        p, this_len, next_len);  
  
  fseeko(vf, this_pos, SEEK_SET);
    
  //-- read pes data (starting at frame pos) into decoder buffer --
  //code = VSTREAM_MASK | stream->vs_id;
  len = readPESvideo(code, pbuf, this_len);
    
  //-- PES packet contains also next frame ? -- 
  if (next_pos-this_pos < this_len)
  {
    total_vlen = this_len - next_len;
    return total_vlen;
  }
    
  pbuf       += len;
  total_vlen += len;  
  
  //-- full scan loop --
  //--------------------
  done = 0;
  while ( !done && (ch = fgetc(vf)) != EOF)
  {
    code = (code << 8) | ch;
     
    //-- at next PACK or PES start code -- 
    //------------------------------------
    if ( (code & CODE_PREFIX_MASK) == CODE_PREFIX )
    {
      off_t akt_pos;
      
      switch (code)
      {
        //-- nothing special on pack start --
        //-----------------------------------
        case PACK_START_CODE:
          break;
          
        //-- any pes packet --
        //--------------------  
        default:

          //-- read complete video pes-data packet --
          len = readPESvideo(code, pbuf, RD_PES_ALL);
          
          akt_pos = ftello(vf);
          if (akt_pos >= next_pos)
          {
            len -= (int32_t)(akt_pos-next_pos);
            done = 1;
          }
          
          pbuf       += len;
          total_vlen += len;
          break;     
      } 
    }
  }  

  return total_vlen;
}


void T_FRAME_SRV_PS::fillAudioBuffer(void)
{
  T_CODE code = CODE_PREFIX_MASK; 
  
  int    ch, pes_len;
 
  if ( ab_level >=  AB_MIN_LEVEL ) return;

  //-- full scan loop --
  //--------------------
  while ( (ab_level < AB_MIN_LEVEL) && (ch = fgetc(af)) != EOF )
  {
    code = (code << 8) | ch;
   
    //-- at next PACK or PES start code -- 
    //------------------------------------
    if ( (code & CODE_PREFIX_MASK) == CODE_PREFIX )
    {
      switch (code)
      {
        //-- nothing special on pack start --
        //-----------------------------------
        case PACK_START_CODE:
          break;
        
        //-- check for audio data --
        //--------------------------  
        default:
          readPESaudio(code, paudio_buf);
          break;
      }  
    }
  }
}

//===========================
//== class T_FRAME_SRV_PES ==
//===========================
static char *id_str_pes = "Program stream PES";
T_FRAME_SRV_PES::T_FRAME_SRV_PES(int t, char *fname, uint8_t vid) 
  : T_FRAME_SRV_PS(t, fname, vid)
{
  id_str = id_str_pes;
};

//==========================
//== class T_FRAME_SRV_ES ==
//==========================

int T_FRAME_SRV_ES::createIndex(void)
{
  T_CODE code = (CODE_PREFIX_MASK >> 8); 
  int    n, i;
  off_t  pos;
   
  seqValid = 0;
  gopValid = 0;
  gop_p    = 0;
  
  if ( (vf = fopen(fname,"rb")) != NULL)
  {
    int l = strlen(fname);
    
    strcat(fname, EXT_INDEX); 
    pf = fopen(fname, "w+b");
    fname[l] = '\0';
    
    if (!pf)
    {
      fprintf(stderr, "ERROR: unable to create ES index for %s\n", fname);
      cleanup();
      return 0;
    } 
    
    //-- write initial record with common stream data first --
    //--------------------------------------------------------
    fwrite(stream, sizeof(T_STREAM), 1, pf);
  
    pos = ftello(vf);
    while ( (n = fread(pes_vs_buf, 1, MAX_PES_VS_LEN, vf)) == MAX_PES_VS_LEN)
    {
      for (i=0; i<MAX_PES_VS_LEN; i++)
      {
        code = (code << 8) | pes_vs_buf[i];
        if ( (code & CODE_PREFIX_MASK) == CODE_PREFIX )
          processVCode(code, pos+i-3, 0, i, MAX_PES_VS_LEN);
      }
      pos = ftello(vf);
      
      if (stream->pict_cnt >= maxIndex) break;
    } 
      
    //-- update common stream record --
    //---------------------------------
    fseeko(pf, 0, SEEK_SET);
    fwrite(stream, sizeof(T_STREAM), 1, pf);
    
    cleanup();
  }
  else
  {
    fprintf(stderr, "ERROR: unable to open (ES) media (%s)\n", fname);
    cleanup();
    return 0;
  }    
    
  //-- summary --
  fprintf(stderr,
    "\nElementary video stream (ES) scan done:\n");
  fprintf(stderr,
    "  (%d) sequence records with\n", 
    stream->seq_cnt);
  fprintf(stderr,
    "    frame size (%dx%d)\n", 
    stream->width, stream->height);
  fprintf(stderr,
    "    aspect code (%d)\n", 
    stream->aspect);
  fprintf(stderr,
    "    frame rate code (%d)\n", 
    stream->frame_rate);
  fprintf(stderr,
    "    (%d) GOP records [# of open=%d]\n", 
    stream->gop_cnt, stream->open_gop);
  fprintf(stderr,
    "    (%d) picture records\n", 
    stream->pict_cnt);
  
  return 1;
}
  
int T_FRAME_SRV_ES::readFrame(uint8_t *buf, int p)
{
  int    total_len; 
  off_t  this_pos, next_pos;

  this_pos = ls_pict[p].pos;
  next_pos = ls_pict[p+1].pos;

  total_len = next_pos - this_pos;

  //-- seek begin of startcode --
  fseeko(vf, this_pos, SEEK_SET);
  total_len = fread(buf, 1, total_len, vf);  

  return total_len;
}

//==========================
//== class T_FRAME_SRV_TS ==
//==========================

static char *id_str_ts = "Transport stream (TS)";
T_FRAME_SRV_TS::T_FRAME_SRV_TS(int t, char *fname, uint8_t vid) 
  : T_FRAME_SRV_PS(t, fname, vid) 
{
  id_str = id_str_ts;
  _stream.vs_pid = -1;
  _stream.as_pid = -1;
};

int T_FRAME_SRV_TS::checkAdaptationField(uint8_t ch, FILE* fp)
{
  int n = 0;
  
  switch (ch & 0x30)
  {
    case TS_AFC_ALL:
      return 0;
     
    case TS_AFC_NONE:
      return -1;
      
    case TS_AFC_PARTIAL:
      n = fgetc(fp);
      fseeko(fp, n, SEEK_CUR);
      n++;
      //fprintf(stderr, "*** adp (%d) ***\n", n);    
      break;
  }  
  
  return n;
}


//== index stuff ==
//=================

void T_FRAME_SRV_TS::checkTESdata(int pid, int len)
{
  uint8_t *ch = pes_vs_buf;
  int32_t pes_len, data_len;
  int32_t data_ofs;
  off_t   pos1, pos2;
  
  int     n;    
  uint8_t baseId = 0x00;
  
  //-- read TS payload --
  //---------------------
  pos1 = ftello(vf);
  fread(ch, 1, len, vf);
  pos2 = ftello(vf);

  if (!memcmp(ch, "\0\0\1", 3)) baseId = ch[3]&0xF0;
       
  //-- known video --
  //-----------------
pid_known: 
  if (pid == stream->vs_pid)
  {
    stream->vs_cnt++;

    g_pes_pos_old = g_pes_pos;
    g_pes_len_old = g_pes_len;
    g_pes_pos     = pos1;
    g_pes_len     = pos2 - pos1;

    if (baseId != 0x00)
    {
      stream->vs_id = (ch[3]&0x0F);
      
      data_ofs = checkPESheader(&ch[6]); // - 6;          
      data_len = g_pes_len - data_ofs;
    }
    else
    {
      data_ofs = 0;
      data_len = g_pes_len;
    }
      
    //-- scan for startcodes in video stream -- 
    checkVStream(data_ofs, data_len);
  }
  //-- known audio --
  //-----------------
  else if (pid == stream->as_pid)
  {
    if (baseId == 0x00) return;

    //-- MP2 --
    //---------
    if (baseId == A_TYPE_MPEG1_L2) 
    {
      n = 30 + (ch[3]&0x0F);
      if (!detected_sr[n])
      {
        pes_len  = pos2 - pos1;
        data_ofs = checkPESheader(&ch[6]); // - 6;          
        data_len = pes_len - data_ofs;
    
        //-- get parameter of audio stream -- 
        checkAStream(data_ofs, data_len, n);
      }
      
      if (detected_sr[n]>0 && detected_br[n]>0) atracks[n]++;  
    } 
    //-- sub audio (only AC3 supported) --
    //------------------------------------
    else if (ch[3] == PR1_CODE) 
    {
      pes_len  = pos2 - pos1;
      data_ofs = checkPESheader(&ch[6]); 
      data_len = pes_len - data_ofs;
    
      if (data_ofs > 0)
      { 
        n = 0; // only 1 AC3 stream per TS file supported !!!
        if (!detected_sr[0])
          checkAStream(data_ofs, data_len, n);
        if (detected_sr[n]>0 && detected_br[n]>0) atracks[0]++;  
      }
    } 
  }
  //-- unknown pid -> search --
  //---------------------------
  else if (baseId == VID_CODE)
  {
    //-- ... only 1 video stream --
    if (stream->vs_pid < 0) 
    {
      stream->vs_pid = pid;
      goto pid_known;
    } 
  }
  else if ( (baseId == A_TYPE_MPEG1_L2) || (ch[3] == PR1_CODE) )
  {      
    //-- ... all MP2/AC3/... audio streams --
    for (n=0; n<stream->n_as; n++)
    {
      if (pid == stream->audio_tracks[n].pid)
      {
        stream->as_pid = pid;
        goto pid_known;
      }
    }
        
    n = stream->n_as;
    stream->n_as++;
        
    stream->audio_tracks[n].pid = pid;
    stream->as_pid              = pid;
    goto pid_known;
  }
     
}

int T_FRAME_SRV_TS::createIndex(void)
{
  uint8_t  code[4];
  uint8_t  pes_valid;
  uint16_t pid;
  
  int      ch, n = 0;
  off_t    pos;

  if (createIndexInit())
  { 
    //-- set audio / video pes buffers equal --     
    //-----------------------------------------
    pes_as_buf = pes_vs_buf;
    
    //-- full scan loop --
    //--------------------
    while ( (ch = fgetc(vf)) != EOF && (stream->pict_cnt < maxIndex) )
    {
      if ( ch == TS_SYNC_BYTE ) 
      { 
        stream->pack_cnt++;
        
        //-- get PID --
        ch   = fgetc(vf);
        pid  = (ch & 0x1F) << 8;
        ch   = fgetc(vf);
        pid |= ch;
         
        //-- check adaptation field --
        ch = fgetc(vf);  
        n  = checkAdaptationField(ch, vf);
        
        if ( n>=0 && pid>=TS_PID_AV_MIN && pid<=TS_PID_AV_MAX )  
        {
          checkTESdata(pid, TS_MAX_PAYLOAD-n);
        }
        else 
        {
          if (n<0) n=0;
          fseeko(vf, TS_MAX_PAYLOAD-n, SEEK_CUR);
        }  
      }
    }
    
    //-- reset pes buffer pointer --
    //------------------------------
    pes_as_buf = _pes_as_buf;
  
    createIndexExit();
  
    return 1;
  }  
  
  return 0;
}
 
//== video stuff ==
//=================  
int T_FRAME_SRV_TS::readTESvideo(uint8_t *pbuf, int this_len, char mode)
{
  uint8_t *ch      = pes_vs_buf;
  int32_t data_len = 0;
  int32_t data_ofs;
  off_t   pos1, pos2;

  if (mode != RD_PES_ALL)
  {
    return fread(pbuf, 1, this_len, vf);
  }
  else if (this_len > 6)
  {
    pos1 = ftello(vf);
    fread(ch, 1, this_len, vf);
    pos2 = ftello(vf);
  
    if (!memcmp(ch, "\0\0\1",3) && ((ch[3]&0xF0) == VID_CODE) )
    {
      data_ofs = checkPESheader(&ch[6]);          
      data_len = (pos2-pos1) - data_ofs;
    
      memcpy(pbuf, &pes_vs_buf[data_ofs], data_len);
    }
    else 
    {
      data_len = pos2 - pos1;
      if ( (data_len>0) && (data_len <= TS_MAX_PAYLOAD) ) 
        memcpy(pbuf, pes_vs_buf, data_len);
      else
        data_len = 0;
    }
    return data_len;
  }
  else
    return 0;
}


int T_FRAME_SRV_TS::readFrame(uint8_t *buf, int p)
{
  uint8_t *pbuf = buf;

  int      n, ch, done, len;
  int32_t  total_vlen; 
  int32_t  this_len, next_len;
  off_t    this_pos, next_pos, akt_pos;
  uint16_t pid;
      
  this_pos = ls_pict[p].pos;
  this_len = ls_pict[p].len;
  
  next_pos = ls_pict[p+1].pos;
  next_len = ls_pict[p+1].len;
  
  total_vlen = 0;
  
  //-- get part of 1st. PES packet of this frame --
  //-----------------------------------------------
  fseeko(vf, this_pos, SEEK_SET);
    
  //-- read pes data (starting at frame pos) into decoder buffer --
  len = readTESvideo(pbuf, this_len, RD_PES_PARTIAL);
  
  pbuf       += len;
  total_vlen += len;  
  
  //-- full scan loop --
  //--------------------
  done = 0;
  n    = 0;
  while ( !done && (ch = fgetc(vf)) != EOF )
  {
    if (ch == TS_SYNC_BYTE)
    {
      ch   = fgetc(vf);
      pid  = (ch & 0x1F) << 8;
      ch   = fgetc(vf);
      pid |= ch;
         
      //-- check adaptation field --
      ch = fgetc(vf);  
      n  = checkAdaptationField(ch, vf);
         
      if ( n>=0 && pid==stream->vs_pid )
      {
        off_t akt_pos;
        
        len = readTESvideo(pbuf, TS_MAX_PAYLOAD-n, RD_PES_ALL);
        
        akt_pos = ftello(vf);
        if (akt_pos >= next_pos)
        {
          len -= (int32_t)(akt_pos-next_pos);
          done = 1;
        }
          
        pbuf       += len;
        total_vlen += len; 
      }
      else
      {
        if (n<0) n=0;
        fseeko(vf, TS_MAX_PAYLOAD-n, SEEK_CUR);
      }
    }
  }      

  return total_vlen;
}

//== AUDIO stuff ==
//=================  
int T_FRAME_SRV_TS::readTESaudio(uint8_t *pbuf, int len)
{
  uint8_t *src     = pes_as_buf;
  int32_t data_len = 0;
  int32_t data_ofs = 0;

  len  = fread(src, 1, len, af);
  
  if (!memcmp(src, "\0\0\1",3))
  {
    int swOfs = 0;
    
    //-- check for valid IDs --
    if ( src[3] == 0xBD )
    {
      //-- skip syncword parameters --
      swOfs = 3;
    }
    else if ( (src[3]&0xF0) != A_TYPE_MPEG1_L2 ) 
      return 0;
      
    data_ofs = checkPESheader(&src[6]) + swOfs;          
    data_len = len - data_ofs;
    
    memcpy(pbuf, &src[data_ofs], data_len);
  }
  else 
  {
    data_len = len;
    if ( (data_len > 0) && (data_len <= TS_MAX_PAYLOAD) ) 
      memcpy(pbuf, src, data_len);
    else
      data_len = 0;
  }
  
  return data_len;  
}

void T_FRAME_SRV_TS::fillAudioBuffer(void)
{
  int      n, len, ch;
  uint16_t pid;
  
  
  if ( ab_level >=  AB_MIN_LEVEL ) return;

  //-- full scan loop --
  //--------------------
  while ( (ab_level < AB_MIN_LEVEL) && (ch = fgetc(af)) != EOF )
  {
    if (ch == TS_SYNC_BYTE)
    {
      ch   = fgetc(af);
      pid  = (ch & 0x1F) << 8;
      ch   = fgetc(af);
      pid |= ch;

      if (pid==stream->as_pid)
      {         
        //-- check adaptation field --
        ch = fgetc(af);  
        n  = checkAdaptationField(ch, af);
       
        if ( (n>=0) && (n<TS_MAX_PAYLOAD) )
        {
          len = readTESaudio(paudio_buf, TS_MAX_PAYLOAD-n);
          paudio_buf += len;
          ab_level   += len; 
          continue;
        }
        
        if (n<0) n = 0;
      }
      else
        n = 3;
        
      if (n<TS_MAX_PAYLOAD) fseeko(af, TS_MAX_PAYLOAD-n, SEEK_CUR);
    }
  }      
}

//===========================
//== class T_FRAME_SRV_PVA ==
//===========================

static char *id_str_pva = "PVA stream ";
T_FRAME_SRV_PVA::T_FRAME_SRV_PVA(int t, char *fname, uint8_t vid) 
  : T_FRAME_SRV_PS(t, fname, vid) 
{
  id_str        = id_str_pva;
  _stream.vs_id = 0;
};

//== index stuff ==
//=================
void T_FRAME_SRV_PVA::checkPVAdata(int sid, int len)
{
  uint8_t *ch = pes_vs_buf;
  off_t   pos1, pos2;
  
    
  //-- read PVA payload --
  //----------------------
  pos1 = ftello(vf);
  fread(ch, 1, len, vf);
  pos2 = ftello(vf);
  
  //-- video --
  //-----------
  if (sid == PVA_VIDEO_ID)
  {
    stream->vs_cnt++;

    g_pes_pos_old = g_pes_pos;
    g_pes_len_old = g_pes_len;

    g_pes_pos = pos1;
    g_pes_len = pos2-pos1;

    //-- scan for startcodes in video stream -- 
    checkVStream(0, g_pes_len);
  }
  //-- audio --
  //-----------
  else 
  { 
    int32_t pes_len, data_len;
    int32_t data_ofs;
    uint8_t id = 0x00;
    int     n;
    
    if (!memcmp(ch, "\0\0\1", 3)) id = ch[3] & 0xF0;

    //-- MP2 --
    //---------
    if (id == A_TYPE_MPEG1_L2) 
    {
      n = 30 + (id & 0x0F);
      atracks[n]++;
    
      if (!detected_sr[n])
      {
        pes_len  = pos2 - pos1 - 6;
        data_ofs = checkPESheader(&ch[6]) - 6;          
        data_len = pes_len - data_ofs;
    
        //-- get parameter of audio stream -- 
        checkAStream(data_ofs, data_len, n);
      }  
    } 
    //-- sub audio --
    //---------------
    else if (id == PR1_CODE) 
    {
      pes_len  = pos2 - pos1 - 6;
      data_ofs = checkPESheader(&ch[6]) - 6;          
      data_len = pes_len - data_ofs;
    
      if (data_ofs > 0)
      { 
        uint8_t sub = pes_vs_buf[data_ofs];
          
        //-- DD5.1 or DTS
        if ( (sub & 0xF0) == A_TYPE_A52_SUB )
        {
          n = (sub & 0x0F);
          atracks[n]++;
          
          if (!detected_sr[n])
            checkAStream(data_ofs+4, data_len-4, n);
        }
        //-- PCM --
        else if ( (sub & 0xF0) == A_TYPE_PCM_SUB )
        {
          n = 15 + (sub & 0x0F);
          atracks[n]++;
            
          if (!detected_sr[n])
            checkAStream(data_ofs+4, data_len-4, n);
        }
      }
    }
    else
      fprintf(stderr, "*** no PES packet header available ***\n");  
  }
}


int T_FRAME_SRV_PVA::createIndex(void)
{
  uint16_t  code = 0;
  int       sid, pts, pre, len;
  int       ch, n = 0;
  uint8_t   pva_hdr[PVA_HDR_SIZE];

  if (createIndexInit())
  { 
    //-- set audio / video pes buffers equal --     
    //-----------------------------------------
    pes_as_buf = pes_vs_buf;
    
    //-- full scan loop --
    //--------------------
    while ( (ch = fgetc(vf)) != EOF && (stream->pict_cnt < maxIndex) )
    {
      code = (code << 8) | ch;
     
      //-- sync to PVA pack start --
      //---------------------------- 
      if ( code == PVA_SYNC_WORD ) 
      { 
        stream->pack_cnt++;
        
        //-- get pva pack header data --
        //------------------------------
        if ( fread(pva_hdr, 1, PVA_HDR_SIZE, vf) == PVA_HDR_SIZE ) 
        {
          sid  = pva_hdr[0];
          pts  = pva_hdr[3] & 0x10;
          //post = pva_hdr[3] & 0x0C;
          pre  = pva_hdr[3] & 0x03;
          len  = (pva_hdr[4] << 8) | pva_hdr[5];
          
          if (sid == PVA_VIDEO_ID && pts)
          {
            //-- skip PTS + padding --
            fseek(vf, (4+pre), SEEK_CUR);
            len -= (4+pre);
          }
          checkPVAdata(sid, len);
        }  
      }
    }
    
    //-- reset pes buffer pointer --
    //------------------------------
    pes_as_buf = _pes_as_buf;
  
    createIndexExit();
  
    return 1;
  }  
  
  return 0;
}

//== video stuff ==
//=================  
int T_FRAME_SRV_PVA::readPVAvideo(uint8_t *pbuf, int this_len)
{
  return fread(pbuf, 1, this_len, vf);
}

int T_FRAME_SRV_PVA::readFrame(uint8_t *buf, int p)
{
  uint8_t *pbuf = buf;
  uint16_t code = 0;

  int      sid, pts, pre;  
  int      ch, done, len;
  int32_t  total_vlen; 
  int32_t  this_len, next_len;
  off_t    this_pos, next_pos, akt_pos;
      
  this_pos = ls_pict[p].pos;
  this_len = ls_pict[p].len;
  
  next_pos = ls_pict[p+1].pos;
  next_len = ls_pict[p+1].len;
  
  total_vlen = 0;

  //-- get part of 1st. PES packet of this frame --
  //-----------------------------------------------
  fseeko(vf, this_pos, SEEK_SET);
    
  //-- read pes data (starting at frame pos) into decoder buffer --
  len = readPVAvideo(pbuf, this_len);
  
  pbuf       += len;
  total_vlen += len;  
  
  //-- full scan loop --
  //--------------------
  done = 0;
  while ( !done && (ch = fgetc(vf)) != EOF )
  {
    code = (code << 8) | ch;
    
    if (code == PVA_SYNC_WORD)
    {
      //-- get pva pack header data --
      //------------------------------
      if ( fread(pes_vs_buf, 1, PVA_HDR_SIZE, vf) == PVA_HDR_SIZE ) 
      {
        sid  = pes_vs_buf[0];
        pts  = pes_vs_buf[3] & 0x10;
        //post = pes_vs_buf[3] & 0x0C;
        pre  = pes_vs_buf[3] & 0x03;
        len  = (pes_vs_buf[4] << 8) | pes_vs_buf[5];
  
        //-- process only video packs --
        //------------------------------        
        if (sid == PVA_VIDEO_ID)
        {
          if (pts)
          {
            //-- skip PTS + padding --
            fseek(vf, (4+pre), SEEK_CUR);
            len -= (4+pre);
          }
         
          len = readPVAvideo(pbuf, len);
        
          akt_pos = ftello(vf);
          if (akt_pos >= next_pos)
          {
            len -= (int32_t)(akt_pos-next_pos);
            done = 1;
          }
          
          pbuf       += len;
          total_vlen += len; 
        }
        else
          fseeko(vf, len, SEEK_CUR);
      }
    }  
  }

  return total_vlen;
}

// eof
