/*
 *  frame_srv.h
 *
 *  Copyright (C) Gerhard Monzel - October 2003
 *
 *  This file is part of lve, a linux mpeg video editor
 *      
 *  lve tools are free software; you can redistribute it and/or modify
 *  it under the terms of the GNU General Public License as published by
 *  the Free Software Foundation; either version 2, or (at your option)
 *  any later version.
 *   
 *  lve is distributed in the hope that it will be useful,
 *  but WITHOUT ANY WARRANTY; without even the implied warranty of
 *  MERCHANTABILITY or FITNESS FOR A PARTICULAR PURPOSE.  See the
 *  GNU General Public License for more details.
 *   
 *  You should have received a copy of the GNU General Public License
 *  along with GNU Make; see the file COPYING.  If not, write to
 *  the Free Software Foundation, 675 Mass Ave, Cambridge, MA 02139, USA. 
 *
 */#ifndef _FRAME_SRV_H
#define _FRAME_SRV_H 1

#include <stdio.h>
#include <string.h>
#include <sys/types.h>
#include <sys/stat.h>
#include <unistd.h>
#include <inttypes.h>

#define EXT_INDEX ".idx"

#define MAX_MAGIC_LEN    8
#define LVE_MAGIC        "LVE-V1.8"   // exactly 8 chars !!!

#define PVA_HDR_SIZE    6       // size without 2 sync bytes !
#define PVA_VIDEO_ID    0x01
#define PVA_AUDIO_ID    0x02
#define PVA_SYNC_WORD	0x4156  // is "AV" in ASCII

#define TS_PID_AV_MIN   0x0020
#define TS_PID_AV_MAX   0x1FFE
#define TS_AFC_ALL      0x10
#define TS_AFC_NONE     0x20
#define TS_AFC_PARTIAL  0x30
#define TS_PACK_LEN     188
#define TS_MAX_PAYLOAD  184
#define TS_SYNC_BYTE    0x47

#define PCK_CODE         0xBA
#define SYS_CODE         0xBB
#define SEQ_CODE         0xB3
#define GOP_CODE         0xB8
#define PIC_CODE         0x00
#define EXT_CODE         0xB5
#define SEQ_END          0xB7
#define PR1_CODE         0xBD
#define PR2_CODE         0xBF
#define VID_CODE         0xE0

#define A_TYPE_NONE      0x00              
#define A_TYPE_A52_SUB   0x80              // 80-87 (AC3) 88-8F (DTS)
#define A_TYPE_PCM_SUB   0xA0              // A0 bis AF
#define A_TYPE_MPEG1_L2  0xC0              // TODO ???

#define CODE_PREFIX      0x00000100
#define CODE_PREFIX_MASK 0xFFFFFF00
#define STREAM_ID_MASK   0xFFFFFFF0

#define PACK_START_CODE  (CODE_PREFIX + PCK_CODE)
#define SYS_START_CODE   (CODE_PREFIX + SYS_CODE)
#define SEQ_START_CODE   (CODE_PREFIX + SEQ_CODE)
#define GROUP_START_CODE (CODE_PREFIX + GOP_CODE)
#define PICT_START_CODE  (CODE_PREFIX + PIC_CODE)
#define EXT_START_CODE   (CODE_PREFIX + EXT_CODE)
#define SEQ_END_CODE     (CODE_PREFIX + SEQ_END)
#define PRIV1_START_CODE (CODE_PREFIX + PR1_CODE)
#define PRIV2_START_CODE (CODE_PREFIX + PR2_CODE)

#define VSTREAM_MASK     (CODE_PREFIX + VID_CODE)
#define ASTREAM_MASK     (CODE_PREFIX + A_TYPE_MPEG1_L2)

#define I_FRAME 1
#define P_FRAME 2
#define B_FRAME 3

#define VOB_PACK_LEN     2048              // VOB PS
#define VCD_PACK_LEN	 2324              // VCD/SVCD

#define MAX_PES_VS_LEN  ((uint16_t)0xFFFF)
#define MAX_PES_AS_LEN  MAX_PES_VS_LEN

#define PES_VS_BUF_LEN  (MAX_PES_VS_LEN+8)
#define PES_AS_BUF_LEN  (MAX_PES_AS_LEN+8)
#define MAX_ABUF_LEN (24*1024)          // 24k buffer for encoded audio stream
#define AB_MIN_LEVEL (8*1024)           //  8k buffer min level in astream buffer

#define MAX_ATRACKS  16

#define MPEG_UNKNOWN 0
#define MPEG_PS      1
#define MPEG_ES      2
#define MPEG_TS      3
#define MPEG_PES     4
#define MPEG_PVA     5
#define MPEG_LVE     6
#define MPEG_WA	     8   // ES with separate wav audio 
#define MPEG_NA      9   // no audio data found

typedef uint32_t T_CODE;

typedef struct
{
  uint8_t id;
  int     sample_rate;
  int     bit_rate;
  int     pid;
} T_AUDIO_TRACK;

typedef struct
{
  char    magic[MAX_MAGIC_LEN];
  int8_t  type;
  int8_t	mpeg2_flag;
  int8_t  atype;
  uint8_t	vs_id;
  uint8_t	as_id;
  
  int		vs_pid;
  int	  as_pid;
  
  int		pack_cnt;
  int		sys_cnt;
  int	  vs_cnt;
  int   as_cnt;
  int		seq_cnt;
  int		gop_cnt;
  int		pict_cnt;

  int   width;
  int   height;
  int   aspect;	
  int   frame_rate;

  char 	open_gop;
  char  n_vs;
  char  n_as;
  
  T_AUDIO_TRACK audio_tracks[MAX_ATRACKS];
} T_STREAM;

typedef struct 
{
  off_t    pos;
  uint16_t len;
  int8_t   num;
} T_POS_REC;

typedef struct
{
  uint8_t  gop_valid;
  uint8_t  closed_gop;
  uint8_t  broken_link;
  uint16_t temporal_ref;
  uint8_t  type;           // (1/2/3) for I/P/B-Frame
  uint8_t  pict_struct;    // (1 and 2/3) for field/frame structure
} T_PIC_HDR;

typedef struct
{
  double   frame_rate;
  uint8_t  aspect;
  uint8_t  rate_code;
  uint32_t nFrames;
  uint8_t  set_closed_gop;   // one shot value !
  uint8_t  set_broken_link;  // one shot value !
  int      ns;
  int      separateFields;
  
  T_PIC_HDR *ph; 
} T_VIDEO_PARMS;


typedef struct 
{
  uint8_t sync[2];
  uint8_t sid;
  uint8_t cnt;
  uint8_t reserved;   // everytime 0x55
  uint8_t flags;      // reserved(3), PTS-flag(1), PostBytes(2), PreBytes(2) 
  uint8_t length[2]; 
} T_PVA_HDR;

class T_USER_PROGRESS
{
  public:
    T_USER_PROGRESS();
  
    virtual void show(int p);
};

class T_FRAME_SRV 
{
  private:
   static T_USER_PROGRESS *userProgress;
    
  protected:
    FILE  *vf;
    FILE  *pf;
    FILE  *af;
    char	fname[1024];
    int   type;
    
    T_STREAM    _stream;
    T_STREAM    *stream;
    T_POS_REC   seq_pos;
    T_POS_REC   gop_pos;
    T_POS_REC   pict_pos;

    T_POS_REC   *ls_pict;

    off_t       g_pes_pos;
    off_t       g_pes_pos_old;
    int32_t     g_pes_len;
    int32_t     g_pes_len_old;
    
    int         seqValid, gopValid;
    int         gop_p;
    uint32_t    maxIndex;
  
  protected:
    virtual void changeIndexFile(void) {};
  
    void    cleanup(void);
    void    processVCode(T_CODE code, 
                         off_t seek_pos, int32_t read_len, 
                         int32_t i, int32_t end);
    uint8_t setAudioTrack(int track);
  
  public:
    static T_FRAME_SRV *createFrameSrv(char *fname);
    static void setUserProgress(T_USER_PROGRESS *progress);

    T_FRAME_SRV(int t, char *fname, uint8_t vid);
    ~T_FRAME_SRV(void);
    
    void    resetPESbuffer(void);
    void    resetAudioBuffer(int p);   
    
    void    setMaxIndex(int n);
    int     firstSequence(int onlySeq = 0);
    int     prevSequence(int p);
    int     nextSequence(int p);
    int     isSequence(int p);
    int     isGOP(int p);
    off_t   filePos(int p);
    
    int     nFrames(void);
    double  videoRate(void);
    int     videoWidth(void);
    int     videoHeight(void);
    int     videoAspect(void);
    char    closedGOP(void);
  
    int     loadIndex(void);
    char    *getFileName(void);
    int     getType() {return type;};
    
    int     sampleRate(int track);
    int     bitRate(int track);
    int     getTotalAudioStreams(void);
    uint8_t getAudioTrack(int track);
    uint8_t getAudioType(int track);
    void    updSampleRate(int track, int sr);
    
    virtual int     createIndex(void) { return 0; };
    virtual int     readFrame(uint8_t *buf, int p) { return 0; };
    virtual uint8_t *readAStream(int *buf_len) {return NULL; };
    virtual int      syncAStream(uint8_t id, uint8_t *pbuf, int32_t len, 
                                int *detected_sr, int *detectd_br) {return 0;};
    
    virtual void    adjustAStream(int bytes) {return; };
    virtual void    fillAudioBuffer(void) {return; };

    static uint8_t   *getPESbuffer(int t);

    static int       nextStartCode(uint8_t *chunk_buf, int len);
    static int       findStartCode(uint8_t *chunk_buf, int len, uint32_t sc);
    static void      patchPIC(uint8_t *chunk_buf, T_VIDEO_PARMS *vp);
    static void      patchGOP(uint8_t *chunk_buf, T_VIDEO_PARMS *vp);
    static void      patchSEQ(uint8_t *chunk_buf, T_VIDEO_PARMS *vp);
    static T_PIC_HDR *getPicHdr(uint8_t *chunk_buf, int len);
};

class T_FRAME_SRV_PS : public T_FRAME_SRV
{
  protected:
    uint32_t atracks[45];               // 0x80-8F, 0xA0-0xAF, 0xC0-0xCF
    int      detected_sr[45];           // dito
    int      detected_br[45];		// dito
    char     *id_str;
    
  protected:
    void  checkVStream(int32_t ofs, int32_t len);
    void  checkAStream(int32_t ofs, int32_t len, uint8_t chan);
    
    int   checkSYSheader(void);
    int   checkPESheader(uint8_t *pes_buf);          
    void  checkPESdata(T_CODE code);
    
    int   readPESvideo(T_CODE code, uint8_t *pbuf, int32_t this_len);
    int   readPESaudio(T_CODE code, uint8_t *pbuf);
    
    int   createIndexInit(void);
    void  createIndexExit(void);
    
  protected:
    virtual void changeIndexFile(void);
      
  public:
    T_FRAME_SRV_PS(int t, char *fname, uint8_t vid);
    
    virtual int     createIndex(void);
    virtual int     readFrame(uint8_t *buf, int p);

    virtual uint8_t *readAStream(int *buf_len);
    virtual void    adjustAStream(int bytes);
    virtual void    fillAudioBuffer(void);

    virtual int syncAStream(uint8_t id, uint8_t *pbuf, int32_t len, 
                            int *detected_sr, int *detectd_br);
};

class T_FRAME_SRV_PES : public T_FRAME_SRV_PS
{
  public:
    T_FRAME_SRV_PES(int t, char *fname, uint8_t vid);
};

class T_FRAME_SRV_TS : public T_FRAME_SRV_PS
{
  private:
    int checkAdaptationField(uint8_t ch, FILE *fp);  
    
  protected:
    void checkTESdata(int pid, int len);
    int  readTESvideo(uint8_t *pbuf, int this_len, char mode);
    int  readTESaudio(uint8_t *pbuf, int len);
        
  public:
    T_FRAME_SRV_TS(int t, char *fname, uint8_t vid);
    
    virtual int  createIndex(void);
    virtual int  readFrame(uint8_t *buf, int p);
    virtual void fillAudioBuffer(void);

};

class T_FRAME_SRV_PVA : public T_FRAME_SRV_PS
{
  protected:
    void checkPVAdata(int sid, int len);
    
    int  readPVAvideo(uint8_t *pbuf, int this_len);
    //int  readPVAaudio(uint8_t *pbuf, int len);
        
  public:
    T_FRAME_SRV_PVA(int t, char *fname, uint8_t vid);
    
    virtual int  createIndex(void);
    virtual int  readFrame(uint8_t *buf, int p);
    //virtual void fillAudioBuffer(void);

};

class T_FRAME_SRV_ES : public T_FRAME_SRV
{
  public:
    T_FRAME_SRV_ES(int t, char *fname, uint8_t vid) : T_FRAME_SRV(t, fname, vid) {};
    
    virtual int createIndex(void);
    virtual int readFrame(uint8_t *buf, int p);
};


#endif // _FRAME_SRV_H
