/*
 *  lve.c
 *
 *  Copyright (C) Gerhard Monzel - October 2003
 *
 *  This file is part of lve, a linux mpeg video editor
 *      
 *  lve tools are free software; you can redistribute it and/or modify
 *  it under the terms of the GNU General Public License as published by
 *  the Free Software Foundation; either version 2, or (at your option)
 *  any later version.
 *   
 *  lve is distributed in the hope that it will be useful,
 *  but WITHOUT ANY WARRANTY; without even the implied warranty of
 *  MERCHANTABILITY or FITNESS FOR A PARTICULAR PURPOSE.  See the
 *  GNU General Public License for more details.
 *   
 *  You should have received a copy of the GNU General Public License
 *  along with GNU Make; see the file COPYING.  If not, write to
 *  the Free Software Foundation, 675 Mass Ave, Cambridge, MA 02139, USA. 
 *
 */
#include <stdlib.h>
#include <SDL/SDL.h>
#include <time.h>
#include <sched.h>

#define _LVE_ 1
#include "lve.h"
#include "lve_config.h"
#include "display.h"
#include "region.h"
#include "lve_ev.h"
#include "lve_gui.h"
#include "scenelst.h"
#include "decoder.h"
#include "audio.h"

#define LVE_OV_FMT SDL_YV12_OVERLAY

#define MAX_LOOP_DELAY  40000                      // 40ms (PAL)
#define MIN_LOOP_DELAY  10000
#define DEF_LOOP_DELAY  15000                      // 15ms for none-PLAY mode
#define LOOPS_PER_SEC   (1000000 / DEF_LOOP_DELAY)

static void lve_av_handler(void);

static T_LVE_CTX   *pctx      = NULL;  // ptr to main context
static T_MPEG_DEC  *decoder   = NULL;  // decoder engine
static T_AUDIO_OUT *audio_out = NULL;  // audio output engine  

int main(int argc, char *argv[])
{
  T_APP              *lve_app;
  T_LVE_CTX	     lve_ctx;  // main context block
  char               *lib_path = PATH_LIB;
  int 		     i;
  struct sched_param sched_p; 
  
  pctx = &lve_ctx;

  fprintf(stderr, "\n===========================================\n");
  fprintf(stderr, "= Linux Video Editor (LVE) V1.1 by gmo18t =\n");
  fprintf(stderr, "= Copyright (c) Gerhard Monzel * 18/01/03 =\n");
  fprintf(stderr, "===========================================\n\n");

  //-- basic setup for main context --
  //----------------------------------
  lve_ctx.file = NULL;
  lve_ctx.w    = 1024;
  lve_ctx.h    = 768;

  //-- free "multimedia" resources --
  //---------------------------------
  system("killall mplayer");
    
  //-- display setup --
  //-------------------
  if (display_init(&lve_ctx, lib_path, "Linux Video Editor (LVE) V1.1")) 
  {
    int use_audio = 1;
    
    //-- this is very important to init the rest of main context !!! --
    //-- create also the mpeg-decoder instance !!!                   -- 
    lve_load_mdf(&lve_ctx);

    //-- setup from command line parameters --
    for (i=1; i<argc; i++)
    {
      if ( !strcmp(argv[i],"-noaudio") || !strcmp(argv[i],"-nosound") )
      {
        fprintf(stderr, "INFO: lve.c - audio disabled !\n");
        use_audio = 0;
      }  
      else if (!strcmp(argv[i],"-ntsc"))
      {  
        pctx->ntsc = 1;
        fprintf(stderr, "INFO: lve.c - NTSC mode enabled !\n");
      }
    } 
  
    //-- after initialization of basic  --
    //-- resources -> lauch application --
    //-- (install region-objects ...)   -- 
    //------------------------------------     
    lve_app = new T_APP(&lve_ctx, lib_path);
    
    display_init_update();

    if (use_audio)
    {
      //-- open audio output system --/
      if ( (audio_out = new T_AUDIO_OUT()) != NULL ) audio_out->open();
    }
    
    //-- get last work out of autosave files --
    REGION(RG_ID_TRIM_SW).SwitchOn(0);
    REGION(RG_ID_CL_GOP).OnMouseLeft(0);
    
    RG_MEDIA.LoadMediaList(lc_getAutoMDList());  
    RG_EDIT.LoadEditList(lc_getAutoEDList());
    
    //-- frame timing without audio-device ... --
    if ( !audio_out || !(audio_out->isOpen()) )
    {
      struct timeval  tv0, tv1;
      long            s0, s1;
 
      for(;;)
      {
        gettimeofday(&tv0, NULL);
        s0 = tv0.tv_sec * 1000000L + tv0.tv_usec;
      
        if (lve_ev(&lve_ctx)) break;

        lve_av_handler();

        if (lve_ctx.action == A_PLAY)      
        {
          gettimeofday(&tv1, NULL);
          s1 = tv1.tv_sec * 1000000L + tv1.tv_usec - s0;
          s1 = MAX_LOOP_DELAY - s1;
          if (s1>MIN_LOOP_DELAY) usleep(s1);
        }
        else
          usleep(DEF_LOOP_DELAY);
      }
    }  
    //-- ... or frame timing controled by audio-device --
    else
    {
      for(;;) 
      {
        if (lve_ev(&lve_ctx)) 
          lve_ctx.action = A_QUIT;
                
        SDL_LockAudio();
        lve_av_handler();
        SDL_UnlockAudio();
        
        //-- wait for new arming of audio out buffer --
        //-- this will be a trigger to display next  --
        //-- frame, because event repeats every 40ms -- 
        if (lve_ctx.action == A_PLAY)      
          audio_out->waitBuffer();
        else if (lve_ctx.action == A_QUIT)
          break;
        else  
          usleep(DEF_LOOP_DELAY);  // or default delay, while not playing
      }
    }   
   
    //-- cleanup open media + mpeg-decoder --
    lve_cleanup_media(&lve_ctx);

    //-- close audio system --
    if (audio_out) 
    {
      audio_out->close();
      delete audio_out;
      audio_out = NULL;
    }
    
    //-- cleanup display overlay --
    display_cleanup_ov();
  }
  else
    fprintf(stderr, "ERROR: main() - display init\n"); 

  return 0;
}

void lve_set_avs(int dp)
{
  if (decoder) 
  {  
    char   avs_str[16];
    T_LIST *ls = T_REGION::GetListGlobal();
    
    sprintf(avs_str, "%+5d", dp);
    RG_TEXT(RG_ID_TEXT_AVS).Print(avs_str);

    decoder->setAudioSync(dp);
    if (ls) ls->UpdateMember(NULL);    
  }
}

static void rc_read_line(char *pbuf, FILE *fp)
{
  char *p;
  fgets(pbuf, 1024, fp);
  if ( (p = strrchr(pbuf, '\n')) != NULL) *p = '\0';
}

static void rc_write_line(char *pbuf, FILE *fp)
{
  fputs(pbuf, fp); 
  if (!strrchr(pbuf, '\n')) fputs("\n", fp);
}

char* lve_fselect_media(int mode, char *filter)
{
  static char fs_buf[1024];
  static char dcache[3][1024];
  static int  dcache_init = 1;
  
  char        *fname = fs_buf;
  char        fn_res[256];
  int         dcache_num = 0;
  FILE        *fp;
  char        *pch;
  
  if (dcache_init)
  {
    sprintf(fn_res, "%s/.lverc", getenv("HOME"));
    if (( fp=fopen(fn_res, "r")) != NULL )
    {
      rc_read_line(dcache[0], fp);
      rc_read_line(dcache[1], fp);
      rc_read_line(dcache[2], fp);
      fclose(fp);
    }
    else
    {
      strcpy(dcache[0], getenv("HOME") );
      strcpy(dcache[1], getenv("HOME") );
      strcpy(dcache[2], getenv("HOME") );
    }
    dcache_init = 0;
  }
 
  dcache_num = atoi(filter);
  
  strcpy(fs_buf, PATH_BIN "qdir ");
  switch (mode)
  {
    case QDIR_NEW:
      strcat(fs_buf, "new ");
      break;
    
    case QDIR_OPEN:
      strcat(fs_buf, "open ");
      break;
      
    case QDIR_SAVE:
      strcat(fs_buf, "save ");
      break;
  }    
 
  strcat(fs_buf, "\"");
  strcat(fs_buf, dcache[dcache_num]);
  strcat(fs_buf, "\""); 
  strcat(fs_buf, filter);
  
  //fprintf(stderr, "call %s\n", fs_buf);
  
  if ( (fp = popen(fs_buf, "r")) != NULL)
  {
    int l;
    
    fs_buf[0] = '\0';
    fgets(fs_buf, 1024, fp);
    
    pch = strchr(fs_buf,'\n');
    if (pch) *pch = '\0';
    
    if (*fs_buf == '\0')
    { 
      pclose(fp);
      return NULL;
    }
    pclose(fp);
    
    l = strlen(fs_buf);
    if (fs_buf[0] == '\"') fname = fs_buf+1;
    if (fs_buf[l-1] == '\"') fs_buf[l-1] = '\0';
  }
  else
  {
    fprintf(stderr,"ERROR: start of (%s) not possible\n", PATH_BIN "qdir");
    return NULL;
  }
  
  //-- clear event queue --
  lve_ev_discard();
  
  //-- show a proper main window --
  pctx->refresh_bg = 2;
  display_refresh(pctx);
  
  //-- remember filename --
  strcpy(dcache[dcache_num], fname);
  sprintf(fn_res, "%s/.lverc", getenv("HOME"));

  if (( fp=fopen(fn_res, "w")) != NULL )
  {
    rc_write_line(dcache[0], fp);
    rc_write_line(dcache[1], fp);  
    rc_write_line(dcache[2], fp);
    fclose(fp);
  }
    
  return (char *)fname;
} 

//============================================================================
//======================= INTERNAL VIDEO-FUNCTIONS ===========================
//============================================================================

static int gmo_video_width(T_FRAME_SRV *file)
{
  return file->videoWidth();
}

static int gmo_video_height(T_FRAME_SRV *file)
{
  return file->videoHeight();
}
//=============================================================================

static int load_media_init = 0;

int lve_load_mdf(T_LVE_CTX *ctx, TMediaFile *mdf)
{
  T_FRAME_SRV *tmp_file;
  
  int i, j;
  int found = 0;
  int rc    = 1;
  
  if (!load_media_init)
  {
    ctx->iv         = IV_START;
    ctx->file       = NULL;
    ctx->v_stream   = 0;
    ctx->a_stream   = 0;
    ctx->action     = A_NOMEDIA;
    ctx->run_once   = 0;
    ctx->adj_pos    = 0;
    ctx->act_frame  = 0;
    ctx->act_real   = 0;
    ctx->w	    = 352; // default
    ctx->h	    = 288; // default	
    ctx->reset_ov   = LVE_OV_FMT;
    ctx->need_thumb = 0;
    ctx->ntsc       = 0;   // i like PAL ! 
    ctx->step       = 0;
    ctx->avs_dp     = 0;
    ctx->ov         = NULL;
    ctx->ov_fmt     = LVE_OV_FMT; 
        
    load_media_init = 1;

    //-- initialization of the decoder-engine --
    decoder = new T_MPEG_DEC();
  }

  //-- media known ? --
  if (mdf != NULL)
  {
    //-- if media already open -> set values and return --
    if (mdf->isOpen)
    {
      ctx->w        = mdf->videoW;
      ctx->h        = mdf->videoH;
      ctx->ov_fmt   = mdf->ovFmt;
      ctx->file     = (T_FRAME_SRV *)mdf->frmSrv;
          
      return rc;
    }

    //-- brand new media -> taking into account --
    if ( (tmp_file = decoder->open(mdf->getName(), &ctx->act_real)) != NULL )
    {
      ctx->file = tmp_file;
      
      int total_astreams;
      int total_vstreams;
      
      mdf->frmSrv = ctx->file;
      mdf->isOpen = 1;              // now file is open :-)
      
      total_astreams = ctx->file->getTotalAudioStreams();
      total_vstreams = 1; 
      
      //-- get and display infos about audio / video --
      fprintf(stderr, "Media (%d): %s\n", mdf->getId(), mdf->getName());
    
      fprintf(stderr, "- (%d/1) audio/video streams\n", total_astreams);
      
      for (j=0; j<total_astreams; j++)
      {
        fprintf(stderr, "- audio stream (0x%02x) with sample rate (%d) Hz, (%d) bps\n", 
                ctx->file->getAudioTrack(j), ctx->file->sampleRate(j), ctx->file->bitRate(j) );  
      }
      fprintf(stderr, "- video stream (0x%02x) with (%d) frames of size (%dx%d), (%0.3f) fps\n", 
              0xE0, ctx->file->nFrames(), 
              ctx->file->videoWidth(), ctx->file->videoHeight(), ctx->file->videoRate());

      ctx->run_once     = 0;
      ctx->adj_pos      = 0;
      ctx->act_frame    = 0;
      ctx->w            = gmo_video_width(ctx->file);
      ctx->h            = gmo_video_height(ctx->file);
      ctx->ov_fmt       = LVE_OV_FMT;

      if (ctx->a_stream < 0 || ctx->a_stream >= total_astreams) 
        ctx->a_stream = 0;
      if (ctx->v_stream < 0 || ctx->v_stream >= total_vstreams) 
        ctx->v_stream = 0;

      fprintf(stderr, "- audio/video stream (%d/%d) selected\n", 
              ctx->a_stream, ctx->v_stream);
      
      mdf->videoW  = ctx->w;
      mdf->videoH  = ctx->h;
      mdf->ovFmt   = ctx->ov_fmt;
    }
    else
    {
      fprintf(stderr, "ERROR: lve_load_mdf - opening file (%s)\n", mdf->getName()); 
      rc = 0;
    }  
  }
  
  return rc;
}  

void lve_close_mdf(TMediaFile *mdf)
{
  if (mdf->isOpen && mdf->frmSrv) decoder->close((T_FRAME_SRV *)mdf->frmSrv);
  mdf->cleanup();
}

void lve_cleanup_media(T_LVE_CTX *ctx)
{
  ctx->iv     = IV_START;
  ctx->action = A_NOMEDIA;
  
  TMediaFile::resetList();
  
  //-- set decoder to a clean state --
  if (decoder) 
  {
    decoder->closeAudio(); 
    delete decoder;
  }
  decoder = NULL;
}

void lve_reset_lists(T_LVE_CTX *ctx)
{
  T_LIST     *lst;

  unlink(lc_getAutoEDList());
  unlink(lc_getAutoMDList());

  RG_EDIT.ClearAllMark();
  RG_MEDIA.ClearAllMark();

  if ( (lst = RG_EDIT.GetList()) != NULL) 
    lst->Reset();
  
  if ( (lst = RG_MEDIA.GetList()) != NULL) 
  {
    lst->Reset();
    ((T_LST_MEDIA *)lst)->resetLastMDFId();
  }  

  TMediaFile::resetList();
  
  ctx->file   = NULL;
  ctx->iv     = IV_START;
  ctx->action = A_NOMEDIA;
  
  RG_TEXT(RG_ID_TEXT_MD).Clear();
  RG_TEXT(RG_ID_TEXT_PRJ).Clear();

  RG_MEDIA.DisplayUpdate();
  RG_EDIT.DisplayUpdate();
}

static int decoder_lock = 0;
void lve_decoder_switch_lock(int sw)
{
  decoder_lock = sw; 
}

//== main decoding and control handler ==
//== of lve triggerd in special ways.  ==
//======================================= 
void lve_av_handler(void)
{
  static int loop_cnt = 1;
  
  //-- loop counter for 60 seconds --
  if ( loop_cnt++ > (60*LOOPS_PER_SEC)) loop_cnt = 1;

  //-- handle requests for display refresh in collective --
  if (pctx->refresh_bg) display_refresh(pctx);

  if (pctx->action == A_NOMEDIA) return;

  //-- select action --
  //-------------------
select_action:  
  switch (pctx->action)
  {
    case A_ATRACK:
      if (audio_out && decoder) 
      {
        decoder->closeAudio();             
        if (decoder->openAudio(pctx->file, pctx->a_stream))  
        {
          audio_out->setup(decoder);
          RG_MEDIA.GetList()->RequestSave();  
          RG_EDIT.GetList()->RequestSave();  
        } 
        else
          audio_out->switchOff();  
      }
      pctx->action = pctx->action_sav;
      goto select_action;
      break;
      
    case A_NOMEDIA:
      return;
      break;
      
    case A_SLIDE:
      //-- slider not moved => return --
      if (!RG_PLAYSL.Update()) return;
          
      pctx->run_once = 0;
      break;

    case A_QUIT:
    case A_STOP:
      if (SDL_GetAudioStatus() == SDL_AUDIO_PLAYING) SDL_PauseAudio(1);
            
      //-- update slider position --
      RG_PLAYSL.Update();

      if (!pctx->run_once) 
      {
        //-- autosave after 3 sec.)--
        if ( (loop_cnt % (3*LOOPS_PER_SEC)) == 0 )
        {
          RG_MEDIA.GetList()->SaveMembers(lc_getAutoMDList(), 1);  
          RG_EDIT.GetList()->SaveMembers(lc_getAutoEDList(), 1);  
        }  
        return;
      }  
      pctx->run_once--;  
      break;
  
    default:
      if (SDL_GetAudioStatus() == SDL_AUDIO_PAUSED) SDL_PauseAudio(0);
      
      //-- update slider pos --
      if ( (loop_cnt % 10) == 0 ) RG_PLAYSL.Update();
      break;        
  }    

  //-- normal run (without seeking) --  
  //----------------------------------
  if (!pctx->adj_pos)
  {
    //-- ready for next picture --
    pctx->act_frame++; 
    pctx->act_real++;
    
    //-- test sync of video and media/edit list --  
    if ( (pctx->act_frame > pctx->end) ) pctx->adj_pos = 1;
  }

  //-- frame-"jump" required ? -- 
  //----------------------------- 
  if (pctx->adj_pos)
  {
    //-- stop and/or adjust it, if needed --
    if (pctx->act_frame >= pctx->dist_sum)
      pctx->act_frame = pctx->dist_sum - 3;
    else if (pctx->act_frame < 0)
      pctx->act_frame = 0; 

    //-- now get real frame (change media, if needed)     --        
    //-- adj_pos may be reset to 0 on media-scene changes --  
    pctx->act_real = T_REGION::GetPosGlobal();
    
    //-- no further jump --
    pctx->adj_pos = 0;
  } 

  //-- handle requests for overlay reset   --
  //-- may be other media after pos change --
  //-----------------------------------------
  if (pctx->reset_ov) 
  {
    //-- at this point we have already the context of the  --
    //-- new scene, so we can install appropriate audio    --
    //-- decoder along with setting up audio output device --  
    if (audio_out) 
    {
      decoder->closeAudio();             
      if (decoder->openAudio(pctx->file, pctx->a_stream))  
        audio_out->setup(decoder);       
      else
        audio_out->switchOff();  
    }
         
    pctx->reset_ov = 0;
    display_init_ov(pctx);
  }
  
  //-- Decoding Part --
  //-------------------
  
  //-- lock display-buffer for writing --
  display_lock();
  
  //-- decode frame and write data to display-buffer,  --
  //-- result is the exact start position of the frame --
  if (!decoder_lock)
  { 
    /*
    while (!decoder->decodeFrame(pctx->file, pctx->act_real, 
                                 &pctx->step,
                                 pctx->page[0], 
                                 pctx->page[2], 
                                 pctx->page[1])) 
    */
    while (!decoder->decodeFrame(pctx->file, pctx->act_real, 
                                 &pctx->step,
                                 pctx->page[0], 
                                 pctx->page[2], 
                                 pctx->page[1])) 
    {
      pctx->act_real++;
      pctx->act_frame++;
    }
  }
    
  //-- ... unlock buffer --
  display_unlock();

  display_draw_frame(pctx); 
    
  //-- generate initial thumbnail after loading new --
  //-- media or after trimming on left scene-side.  --
  if (pctx->need_thumb)
  { 
    if (pctx->need_thumb > 1)
    { 
      RG_EDIT.GetList()->GetFocusMember()->GenThumb();
      RG_EDIT.DisplayUpdate();
    }
    else
    {
      RG_MEDIA.GetList()->GetFocusMember()->GenThumb();
      RG_MEDIA.DisplayUpdate();
    }
    pctx->need_thumb = 0;
  }
    
  //-- autostop on end of video --
  if (pctx->act_frame >= (pctx->dist_sum-1)) 
    REGION(RG_ID_PLAY).SwitchOff(0);

  return;
}

//eof

