/*
 *  lve_read.c
 *
 *  Copyright (C) Gerhard Monzel - October 2003
 *
 *  This file is part of lve, a linux mpeg video editor
 *      
 *  lve tools are free software; you can redistribute it and/or modify
 *  it under the terms of the GNU General Public License as published by
 *  the Free Software Foundation; either version 2, or (at your option)
 *  any later version.
 *   
 *  lve is distributed in the hope that it will be useful,
 *  but WITHOUT ANY WARRANTY; without even the implied warranty of
 *  MERCHANTABILITY or FITNESS FOR A PARTICULAR PURPOSE.  See the
 *  GNU General Public License for more details.
 *   
 *  You should have received a copy of the GNU General Public License
 *  along with GNU Make; see the file COPYING.  If not, write to
 *  the Free Software Foundation, 675 Mass Ave, Cambridge, MA 02139, USA. 
 *
 */
//== briefly API example 
//== -------------------
//==
//== lr_init(tr);
//==
//== do 
//== {
//==   lr_get_frame(ctx);        // read 1 yuv frame
//==   lr_get_samples(ctx,...);  // read sound samples for one frame
//==   
//==   //-- check for scene change --
//==   if (ctx->new_scene)
//==   {
//==     //do your stuff here ...
//==
//==     ctx->new_scene = 0; // !!!
//==   }
//== } while ( ctx->read_mode != SC_EOF )
//== 
//== lr_cleanup(); 
//==

#include <stdio.h>
#include <stdlib.h>
#include <malloc.h>

#include "audio.h"
#include "decoder.h"

#include "lve_read.h"

#define FPS_PAL 25.0
#define FPS_MIN 23.0

//== global context-pointer ==
static T_LVE_READ_CTX *_pctx = NULL;

//== global file pointer (scene-list file) ==
static FILE *scene_lst_file  = NULL;

//== decoder engine ==
static T_MPEG_DEC *decoder = NULL;


//============================================================================
//======================= INTERNAL AUDIO-FUNCTIONS ===========================
//============================================================================

static uint32_t *scale_buf = NULL;

//== read/build audio portion for one picture as interlaced ==
//== stereo pcm-buffer, signed 16-Bit - also adjusted in    ==
//== duration on non PAL video, if forced to PAL with       ==
//== 'force_pal' = 1.                                       == 
//============================================================
int lr_get_samples(T_LVE_READ_CTX *pctx, uint8_t *pbuf, int force_pal)
{
  static double scale        = 1.0;
  static int    bytes_needed = 0;
  
  int bytes_real = 0;

  if (pctx->new_scene || !bytes_needed)
  {
    //pctx->new_scene = 0;  TODO
    scale           = 1.0;
    bytes_needed    = (int)(pctx->sample_rate/(pctx->frame_rate + 0.5)) * 4;
    
    if (force_pal && (pctx->frame_rate != FPS_PAL) ) 
    {
      scale        = FPS_PAL / (double)pctx->frame_rate; 
      bytes_needed = (int)(pctx->sample_rate / FPS_PAL) * 4;
    }
  }
  
  //-- read real amount of samples for 1 picture to internal buffer --
  //------------------------------------------------------------------
  if (pctx->has_audio)
  {
    bytes_real = (decoder->getAudioDecoder())->readSamples(scale_buf, pctx->samples_cnt); 
  }
  
  //-- (scene) without audio -> produce silence --
  if (!bytes_real)
  {
    memset(pbuf, 0, bytes_needed);
    return bytes_needed;
  }  
 
  //-- without scaling or ... --
  //----------------------------  
  if ( !force_pal || scale == 1.0)
  {
    memcpy(pbuf, scale_buf, bytes_real);
    bytes_needed = bytes_real;    
  }
  //-- ... resample to PAL
  else
  {
    uint32_t *dst = (uint32_t *)pbuf;
    for(int i=0; i<bytes_needed>>2; i++) dst[i] = scale_buf[(int)(i*scale)];
  }
  
  return bytes_needed;
}

//============================================================================
//======================= INTERNAL MEDIA-FUNCTIONS ===========================
//============================================================================

#define MAX_MEDIA_CNT 128 // no more than 32 differnt media files allowed !!!

typedef struct
{
  int         m_id;
  T_FRAME_SRV *fs;
  
  int     video_w;
  int     video_h;
  double  frame_rate;
  int     aspect;
  
  int     has_audio;
  int     sample_rate;
  int     channels;
  
} T_MEDIA_TAB;


static int load_media_init = 0;
static int id_cnt          = -1;

static T_MEDIA_TAB media_tab[MAX_MEDIA_CNT];

//== get media-file ready to work (using a parameter-cache) ==
//============================================================ 
static int lr_media_load(T_LVE_READ_CTX *pctx, char *fname, int media_id)
{
  int i, dmy;
  
  if (!load_media_init)
  {
    for (i=0; i<MAX_MEDIA_CNT; i++)
    {
      media_tab[i].m_id    = -1;
      media_tab[i].fs      = NULL;

      media_tab[i].video_w = 0;
      media_tab[i].video_h = 0;
      
      media_tab[i].frame_rate  = FPS_PAL;  // default
      media_tab[i].aspect      = 2;        // 4:3 default
      
      media_tab[i].has_audio   = 0;
      media_tab[i].sample_rate = 48000;     // default 
      media_tab[i].channels    = 2;         // default 
    }   
    load_media_init = 1;
  }

  if (fname != NULL)
  {
    //-- check if media is already cached --
    for (i=0; i<MAX_MEDIA_CNT; i++)
    {
      if (media_tab[i].m_id == media_id)
      {  
        pctx->src_w = media_tab[i].video_w;
        pctx->src_h = media_tab[i].video_h;
        pctx->fs    = (void *)media_tab[i].fs;
        
        pctx->frame_rate = media_tab[i].frame_rate;
        pctx->aspect     = media_tab[i].aspect; 
        pctx->has_audio  = media_tab[i].has_audio;
        
        if (pctx->has_audio)
        {
          pctx->sample_rate = media_tab[i].sample_rate;
          pctx->channels    = media_tab[i].channels;
        }
        
        //-- any scene with audio ? --
        pctx->total_audio |= pctx->has_audio;
      
        return media_id;
      }
    }
    
    //-- create frame server for this media --
    pctx->fs = (void *)decoder->open(fname, &dmy);
    if (pctx->fs)
    {
      T_FRAME_SRV *fs = (T_FRAME_SRV *)(pctx->fs);
      
      id_cnt++;
      
      media_tab[id_cnt].m_id = media_id;
      media_tab[id_cnt].fs   = fs;

      fprintf(stderr, "INFO: lve_read.c - using media file (%s)\n", fname); 
      fprintf(stderr, "  Audio streams: %d\n", fs->getTotalAudioStreams());
    
      pctx->src_w      = fs->videoWidth();
      pctx->src_h      = fs->videoHeight();
      pctx->frame_rate = fs->videoRate();
      pctx->aspect     = fs->videoAspect();

      pctx->has_audio    = fs->getTotalAudioStreams();
      pctx->sample_rate  = fs->sampleRate(pctx->a_stream);
      pctx->channels     = 2;
      
      media_tab[id_cnt].video_w     = pctx->src_w;
      media_tab[id_cnt].video_h     = pctx->src_h;
      media_tab[id_cnt].frame_rate  = pctx->frame_rate;
      media_tab[id_cnt].aspect      = pctx->aspect;
      
      media_tab[id_cnt].has_audio   = pctx->has_audio;
      media_tab[id_cnt].sample_rate = pctx->sample_rate;
      media_tab[id_cnt].channels    = pctx->channels;
     
      //-- any scene with audio ? --
      pctx->total_audio |= pctx->has_audio;

      return media_id;      
    }
    else
      fprintf(stderr, "ERROR: lve_read.c - opening media-file (%s)\n", fname); 
  }
  
  return -1;
}  

//== get the right media-scene for the frame-reader of the encoder ==
//== on scene-changes extract audio portion of the previous scene  ==
//===================================================================
static int hdr_done_v    = 0;
static int last_media_id = -1;

int lr_media_chk_range(T_LVE_READ_CTX *pctx, int force_next)
{
  int    need_next = force_next;
  char   *fname    = pctx->lst_fname;
  
  //-- eventually scene list must be opened --
  //------------------------------------------ 
  if (!hdr_done_v)
  {
    scene_lst_file = fopen(fname, "rb");
    need_next      = 1;
  }  
  //-- check position in media-file --
  //----------------------------------
  else if (!need_next)
  {
    if (decoder->getActFrame() >= pctx->md_buf.end_real)
    {
      fprintf(stderr, "\n");
      need_next = 1;
    }
  } 
   
  //-- setup next scene --
  //---------------------- 
  if (scene_lst_file && need_next)
  {
    //-- only once read list header --
    //--------------------------------
    if (!hdr_done_v)
    {
      if (fread(&pctx->md_header, sizeof(T_MEDIA_HDR), 1, scene_lst_file))
      {
        fprintf(stderr, "INFO: lve_read.c - Total (%d) scenes with total (%d) frames\n",    
                pctx->md_header.mb_cnt, pctx->md_header.dist_sum ); 
       
        fprintf(stderr, "INFO: lve_read.c - Estimated output size total (%d MB)\n",    
                pctx->md_header.output_size_mb ); 
        
#if 0        
        //-- while using tcmplex to mux this patch is needed --
        if (pctx->a_shift != 0) pctx->md_header.avs_dp += pctx->a_shift;
        
        if ( pctx->md_header.avs_dp != 0)
        {         
          fprintf(stderr, "INFO: lve-read.c - A/V sync correction (%d) frames\n", 
                  pctx->md_header.avs_dp);
        } 
#endif        
        hdr_done_v = 1;
      }
      else
      {
        fprintf(stderr, "ERROR: lve_read.c - no data found in edit-list (%s)\n", fname);
        pctx->read_mode = SC_EOF;
        return 0;
      }   
    }

    //-- read next scene-description --
    //---------------------------------
    pctx->new_scene = 1;
    
    //-- eventually save actual media-buffer --
    if (pctx->md_bup) 
      memcpy(pctx->md_bup, &pctx->md_buf, sizeof(T_MEDIA_BUF));
      
    //-- fillup media-buffer from file --
    if (!fread(&pctx->md_buf, sizeof(T_MEDIA_BUF), 1, scene_lst_file))
    {
      fprintf(stderr, "INFO: lve_read.c - no (more) scene data available\n"); 
      pctx->read_mode = SC_EOF;
      return 0; 
    }
    else
    {
      fprintf(stderr, "INFO: lve_read.c - new scene with (%d) frames to decode\n",
              pctx->md_buf.dist ); 
#if 0              
      decoder->setAudioSync(pctx->md_header.avs_dp);
#endif
    }

    pctx->v_stream = pctx->md_buf.vstream;
    
    if (pctx->force_atrack >= 0)
      pctx->a_stream = pctx->force_atrack;
    else
      pctx->a_stream = pctx->md_buf.astream; 

    //-- get media-file into account -> setup pctx->file now --
    //---------------------------------------------------------
    if (lr_media_load(pctx, pctx->md_buf.fname, pctx->md_buf.media_id) < 0)
    {
      fprintf(stderr, "ERROR: lve_read.c - trouble while loading media\n");
      
      pctx->read_mode = SC_EOF; // end session
      return 0;
    }      

    //-- set start frame of the new scene for video decoder --
    //--------------------------------------------------------
    pctx->act_real = pctx->md_buf.start_real;

    //-- on media-id change, set new audio decoder --
    //-----------------------------------------------
    if (last_media_id != pctx->md_buf.media_id)
    {
      int step = 0; // !!!
      
      decoder->closeAudio();
      decoder->openAudio((T_FRAME_SRV *)(pctx->fs), pctx->a_stream);    
      last_media_id = pctx->md_buf.media_id;
      
      //-- this will help to fill audio stream buffer --
      //-- and allow transcode to start with audio    --
      //-- the next decodeFrame() call on the same    --
      //-- frame doesn't need decode again.           -- 
      decoder->decodeFrame((T_FRAME_SRV *)(pctx->fs), 
                           pctx->act_real, &step, 
                           pctx->py, pctx->pv, pctx->pu);
    }
    
    //-- adjust global A/V shift with scene  --
    //-- A/V shift and deliver it to decoder --
    if (pctx->a_shift != 0) pctx->md_buf.avs_dp += pctx->a_shift;
    decoder->setAudioSync(pctx->md_buf.avs_dp);  
  }
  else if (!scene_lst_file)
  {
    fprintf(stderr, "ERROR: lve_read.c - opening edit-list (%s)\n", fname); 

    pctx->read_mode = SC_EOF; // end session 
    return 0;
  }
  
  
  return 1;
}

//== cleanup media resources ==
//=============================
static void lr_media_cleanup(void)
{
  int i;

  //-- don't call any guru --
  //-------------------------  
  if (!_pctx) return;
  
  //-- run only once --
  if (!load_media_init) return;
  
  //-- close scene-list file --
  if (scene_lst_file) fclose(scene_lst_file);
  scene_lst_file = NULL;

  //-- close all used media files --
  for (i=0; i<MAX_MEDIA_CNT; i++)
  { 
    if (media_tab[i].fs) decoder->close(media_tab[i].fs);
    media_tab[i].fs = NULL;
  }
  
  //-- free main context --
  //-----------------------
  free(_pctx); 
  _pctx = NULL; // !!!!
  
  load_media_init = 0;
}

//=============================================================================
//============================  GLOBAL FUNCTIONS ==============================
//=============================================================================

//== this has to be called first to test for edit-list ==
//== and store fname to further processing !!!!!!!!!!! ==
//=======================================================
static char lst_fname[1024] = "";

int lr_file_chk(char *fname)
{
  FILE *file;
  char magic[MAGIC_LEN+1];
  
  strcpy(lst_fname, fname);
  
  if ( (file = fopen(fname, "rb")) != NULL)
  {
    fread(magic, MAGIC_LEN, 1, file);
    magic[MAGIC_LEN] = '\0';
    fclose(file);
    
    if ( !strcmp(magic, EDIT_BUF_MAGIC) ) 
    {
      fprintf(stderr, "INFO: lve_read.c - using lve edit-list (%s)\n", fname);
      return 1;
    }
  }
  return 0;  
}

//== initialize reader subsystem called as 2nd. routine ==
//========================================================
T_LVE_READ_CTX *lr_init(int a_track, int a_shift)  
{
  T_LVE_READ_CTX *pctx = NULL;
  
  //-- first allocate memory to hold main work-context --
  if ((pctx = (T_LVE_READ_CTX *)malloc(sizeof(T_LVE_READ_CTX))) != NULL)
  {
    if ( (scale_buf = (uint32_t *)memalign(32, (MAX_SAMPLE_BUF+1)*4)) == NULL )
    {
      fprintf(stderr, "lve_read.c - cannot create audio scale buffer\n"); 
      lr_cleanup();
      return NULL;
    }        
    
    _pctx = pctx;  

    pctx->a_shift     = a_shift;
    
    pctx->fs          = NULL;
    pctx->read_mode   = SC_NORM;
    
    pctx->py          = NULL;
    pctx->pu          = NULL;
    pctx->pv          = NULL; 
    
    pctx->v_stream     = 0;
    pctx->a_stream     = 0;
    pctx->force_atrack = a_track;

    pctx->total_audio = 0;
    pctx->has_audio   = 0;
    pctx->aframe_size = 0;
    pctx->samples_cnt[0] = 0.0;
    pctx->samples_cnt[1] = 0.0;
    pctx->samples_cnt[2] = 0.0;
    
    pctx->md_buf.media_id    = -1;
    pctx->md_buf.start_real  = 0;
    pctx->md_buf.end_real    = 0;
    pctx->md_bup             = NULL;

    pctx->new_scene          = 1;
    pctx->total_frames       = 0;
    pctx->total_scenes       = 0;
    
    pctx->lst_fname          = lst_fname;
    
    if ( (decoder = new T_MPEG_DEC()) == NULL )
    {
      fprintf(stderr, "lve_read.c - cannot create MPEG decoder\n"); 
      lr_cleanup();
      return NULL;
    }
     
    //-- 1st. call will initialize media stuff --
    lr_media_chk_range(pctx, 1);
  
    pctx->pic_size_l = pctx->src_w * pctx->src_h;
    pctx->pic_size_c = pctx->pic_size_l>>2; 
    pctx->v_bitrate  = decoder->getBitRate();
  }
  
  return pctx;
}

//== cleanup reader subsystem ==
//==============================
void lr_cleanup(void)
{
  lr_media_cleanup();
  if (decoder) delete decoder;
  if (scale_buf) free(scale_buf);
  
  decoder   = NULL;
  scale_buf = NULL;
}

//== read decoded video YUV-frame ==
//==================================  
int lr_get_frame(T_LVE_READ_CTX *pctx)
{
  int corr;
  
  //-- decode next frame --
  while (!decoder->decodeFrame((T_FRAME_SRV *)(pctx->fs), 
                               pctx->act_real, &corr, 
                               pctx->py, pctx->pv, pctx->pu))
  {
    pctx->act_real++;
    pctx->total_frames++;
  }    

  //-- test for scene change --
  lr_media_chk_range(pctx, 0);
  if (pctx->read_mode == SC_EOF) return 0;
            
  //-- next real frame --
  if (!pctx->new_scene) pctx->act_real++;
 
  pctx->total_frames++;

  return 1;
}

//========================================================
//== TODO: dump media content selected by lve edit-list ==
//========================================================
//static FILE         *vdmp_fp = NULL;
static FILE         *admp_fp = NULL;
static T_WAV_HEADER wav_hdr;

static void fill_wav_header(uint32_t sr)
{
  uint32_t fsize = (uint32_t)ftello(admp_fp);
  //uint32_t sr    = _pctx->sample_rate;
  uint32_t br    = sr / 8;

  //-- RIFF-Segment --
  //------------------
  strncpy(wav_hdr.tag1, "RIFF", 4);
  wav_hdr.seg_len1[0] = 0x04; //(unsigned char)apos;

  strncpy(wav_hdr.fourcc, "WAVE", 4);

  //-- FMT-Segment --
  //-----------------
  strncpy(wav_hdr.tag2, "fmt ", 4);
  wav_hdr.seg_len2[0] = 0x10; // segment len 16
  
  wav_hdr.stereo[0] = 0x01;   // stereo 
  wav_hdr.stereo[1] = 0;
  
  wav_hdr.channels[0] = 0x02; // stereo
  wav_hdr.channels[1] = 0;
  
  wav_hdr.sample_rate[0] = (unsigned char)(sr);
  wav_hdr.sample_rate[1] = (unsigned char)(sr >> 8);
  wav_hdr.sample_rate[2] = (unsigned char)(sr >> 16);
  
  wav_hdr.bit_rate[0] = (unsigned char)br;
  wav_hdr.bit_rate[1] = (unsigned char)(br>>8);
  wav_hdr.bit_rate[2] = (unsigned char)(br>>16);
  wav_hdr.bit_rate[3] = (unsigned char)(br>>24);

  wav_hdr.Bps[0] = 0x04; // 4 Bytes per sample over all channels ?
  wav_hdr.bps[0] = 0x10; // 16 Bit per channel
  
  //-- DATA-Segment --
  //------------------
  strncpy(wav_hdr.tag3, "data", 4);
  
  wav_hdr.seg_len3[0] = (unsigned char)fsize;
  wav_hdr.seg_len3[1] = (unsigned char)(fsize >> 8);
  wav_hdr.seg_len3[2] = (unsigned char)(fsize >> 16);
  wav_hdr.seg_len3[3] = (unsigned char)(fsize >> 24);
}
  
//== open and prepare wav dump files ==
//=====================================
int lr_init_wav(FILE *afp, int wav_sr)
{
  if (afp) admp_fp = afp;
  
  if (wav_sr) 
  {
    memset(&wav_hdr, 0, sizeof(T_WAV_HEADER));
    fill_wav_header((int32_t)wav_sr);
    fwrite(&wav_hdr, sizeof(T_WAV_HEADER), 1, afp);
  }
       
  return 1;
}

//== read encoded (raw) video data into decoder buffer(!) ==
//========================================================== 

int lr_get_vchunk(T_LVE_READ_CTX *pctx, int p, int gop_bound, void *videoParms)
{
  T_FRAME_SRV   *fs = (T_FRAME_SRV *)pctx->fs;
  T_VIDEO_PARMS *vP = (T_VIDEO_PARMS *)videoParms;
  int           _p;

  //-- read chunk data into buffer --
  //---------------------------------
  _p = decoder->readFrame(fs, p, pctx->chunk_buf, &pctx->buf_len, gop_bound);
  
  //-- if patch-data available --
  //-----------------------------
  if ((_p>=0) && vP)
  {
    int     ofs     = 0;
    int     len     = pctx->buf_len;
    uint8_t *buf    = pctx->chunk_buf;
    uint8_t *gopPtr = NULL;

    //-- chunk buffers containing sequence and/or GOP should be patched --
    //-------------------------------------------------------------------- 
    switch (buf[3])
    {
      //-- chunk starts with sequence --
      //--------------------------------
      case (SEQ_START_CODE & 0xFF):
        
        //-- patch frame rate code --
        T_FRAME_SRV::patchSEQ(&buf[4], vP);
        
        //-- search for GOP header ... --
        if ( (ofs = T_FRAME_SRV::findStartCode(buf, len, GROUP_START_CODE)) >= 0 )
        {
          buf += ofs;
          len -= ofs;
        
          gopPtr = buf;
        }
        break;
          
      //-- chunks starts with GOP (no sequence available) -- 
      //----------------------------------------------------   
      case GROUP_START_CODE & 0xFF:
        gopPtr = buf;
        break;
      
      //-- nothing to do --
      //-------------------  
      default:
        ;    
    }

    vP->ph             = NULL;
    vP->separateFields = 0;
    
    //-- get picture picture + extension header values --     
    //---------------------------------------------------
    if ( (ofs = T_FRAME_SRV::findStartCode(buf, len, PICT_START_CODE)) >= 0)
    {
      vP->ph = T_FRAME_SRV::getPicHdr(buf, len);
#if 0
      if (vP->ph->pict_struct != 3) 
      {
        vP->separateFields = 1; 
        vP->nFrames += 1;
        vP->nFrames /= 2;
      }
#endif      
    }
    
    //-- patch timcode if GOP available --
    //------------------------------------
    if (gopPtr)
    {
      vP->set_broken_link = 1;  
      vP->set_closed_gop  = 1;
      T_FRAME_SRV::patchGOP(&gopPtr[4], vP);
    }
    
  }
  
  return _p;
}

//== simple audio resampler: resamples <pctx->buf_len> bytes in ==
//== <pctx->chunk_buf> from <pctx->sample_rate to> <dst_sr> and ==
//== returns adjusted count of bytes in <pctx->achunk_buf>.     ==
//================================================================
int lr_resample(T_LVE_READ_CTX *pctx, int dst_sr, int init)
{
  static uint32_t *pbuf = NULL; 
  static double   scale = 0.0;
  static int      dir   = 0;
  static int      rc    = 0;
  static int      samples_pvf = 0;

  if (init) 
  {
    pbuf        = (uint32_t *)pctx->chunk_buf;
    scale       = dst_sr/(double)pctx->sample_rate;
    samples_pvf = pctx->buf_len>>2;
    
    if (pctx->sample_rate < dst_sr) 
      dir = 1;
    else
      dir = 0;  
  }
  
  if (!dir)
    for (int i=0; i<samples_pvf; i++) pbuf[(int)(i*scale)] = pbuf[i];
  else
    for (int i=(int)(samples_pvf*scale)-1; i>=0; i--) pbuf[i] = pbuf[(int)(i/scale)];
    
  return ((int)(samples_pvf * scale) * 4);
}

uint8_t *lr_mSeqHdr(int *len)
{
  
  if ( (*len = decoder->mSeqHdrLen) > 0 )
    return decoder->mSeqHdr;
  else
    return NULL;  
}

//=============================
//== E X P E R I M E N T A L ==
//=============================

//== get access to raw audio chunk data ==
//========================================  
int lr_get_achunk(T_LVE_READ_CTX *pctx)
{
  pctx->buf_len    = (decoder->getAudioDecoder())->readChunkData(pctx->chunk_buf, pctx->samples_cnt);
  pctx->encoded_fs = (decoder->getAudioDecoder())->getEncodedFs();
  return pctx->buf_len;
}

int lr_acodec_id(void)
{
  return (int)(decoder->getAudioDecoder())->getCodecId();
}

// eof
