/*
 *  lvedemux.c
 *
 *  Copyright (C) Gerhard Monzel - October 2003
 *
 *  This file is part of lve, a linux mpeg video editor
 *      
 *  lve tools are free software; you can redistribute it and/or modify
 *  it under the terms of the GNU General Public License as published by
 *  the Free Software Foundation; either version 2, or (at your option)
 *  any later version.
 *   
 *  lve is distributed in the hope that it will be useful,
 *  but WITHOUT ANY WARRANTY; without even the implied warranty of
 *  MERCHANTABILITY or FITNESS FOR A PARTICULAR PURPOSE.  See the
 *  GNU General Public License for more details.
 *   
 *  You should have received a copy of the GNU General Public License
 *  along with GNU Make; see the file COPYING.  If not, write to
 *  the Free Software Foundation, 675 Mass Ave, Cambridge, MA 02139, USA. 
 *
 */
#include <stdio.h>
#include <string.h>
#include <stdlib.h>
#include "frame_srv.h"

class T_DEMUXER : public T_FRAME_SRV_PS
{
  private:
    FILE    *dmx_vf;
    FILE    *dmx_af;
    uint8_t *pes_buf;
    
  protected:
    int demuxPESdata(T_CODE code, int32_t *d_ofs, int32_t *d_len);  
    
  public:
    T_DEMUXER(char *_fname);
  
    int demuxAV(char *dir, char *name, int v2stdout, uint8_t vs_id, uint8_t as_id);
};

T_DEMUXER::T_DEMUXER(char *_fname) : T_FRAME_SRV_PS(-1, _fname, 0)
{
  dmx_vf = NULL;
  dmx_af = NULL;
  
  //-- setup packet buffer --
  pes_buf = getPESbuffer(0);
}

int T_DEMUXER::demuxPESdata(T_CODE code, int32_t *d_ofs, int32_t *d_len)
{
  int32_t pes_len, hdr_len;

  //fprintf(stderr, "*** (0x%03x / 0x%02x) ***\n", code, stream->vs_id); 
  
  //-- match video stream ... --
  //----------------------------
  if ( (code & 0xFF) == stream->vs_id) 
  {
    stream->vs_cnt++;
    
    pes_len = fgetc(vf);
    pes_len = (pes_len << 8) | fgetc(vf);

    //-- read pes pack now --    
    pes_len = fread(pes_buf, 1, pes_len, vf);
    hdr_len = checkPESheader(pes_buf);
          
    *d_ofs = hdr_len - 6;
    *d_len = pes_len - *d_ofs;
   
    return(stream->vs_id);
  }
  //-- ... or plain audio stream ... --
  //-----------------------------------
  else if ( (code & 0xFF) == stream->as_id )        
  {
    stream->as_cnt++;
    
    pes_len = fgetc(vf);
    pes_len = (pes_len << 8) | fgetc(vf);
    
    pes_len = fread(pes_buf, 1, pes_len, vf);
    hdr_len = checkPESheader(pes_buf);
          
    *d_ofs = hdr_len - 6;
    *d_len = pes_len - *d_ofs;
         
    return(code & 0xFF);      
  }
  //-- ... or other PES packet including sub audio --
  //-------------------------------------------------
  else 
  {
    pes_len = fgetc(vf);
    pes_len = (pes_len << 8) | fgetc(vf);
    
    switch (code) 
    {
      case SYS_START_CODE:
        stream->sys_cnt++;
        pes_len = fread(pes_buf, 1, pes_len, vf);
        break;
      
      //-- check audio subs --
      //----------------------
      case PRIV1_START_CODE:
        pes_len = fread(pes_buf, 1, pes_len, vf); 
        hdr_len = checkPESheader(pes_buf);
            
        *d_ofs = hdr_len - 6;
        *d_len = pes_len - *d_ofs;

        if (*d_ofs > 0)
        { 
          //-- DD5.1 / DTS / PCM --
          uint8_t sub = pes_buf[*d_ofs];
          if ( sub == stream->as_id ) 
          {
            stream->as_cnt++;
            *d_ofs += 4;
            *d_len -= 4;
            return sub;
          }
        }
        break;
    
      default:
        //-- skip pes data -- 
        fseeko(vf, pes_len, SEEK_CUR);
        break;
    }
  }
  
  return 0; 
}


int T_DEMUXER::demuxAV(char *dir, char *name, int v2stdout, uint8_t vs_id, uint8_t as_id)
{
  int32_t n, d_ofs, d_len; 
  int     ch;
  T_CODE  code = (CODE_PREFIX_MASK >> 8);

  stream->vs_id = vs_id;
  stream->as_id = as_id;

  fprintf(stderr, "demuxing media file (%s) ...\n", fname);

  //-- read from stdin or file --
  //----------------------------- 
  if (!strcmp(fname, "-"))
  { 
    if (strlen(name))
      strcpy(fname, name); 
    else
      strcpy(fname,"out"); 
    vf = fdopen(0,"rb");
  }
  else
    vf = fopen(fname,"rb");
    
  if (vf != NULL)
  {
    char full_fn[2048];
    char *pstr;
    
    int  l        = strlen(dir);
    char delim[2] = "/";
    
    if ( (l && dir[l-1]=='/') || (l == 0) ) strcpy(delim,"");
    
    
    //-- is there user defined name --
    if (strlen(name)) strcpy(fname, name);
    
    pstr = strrchr(fname, '.');
    if (pstr) *pstr = '\0';
    strcat(fname, ".m2v");
    if (vs_id) 
    { 
      //-- video redirect to stdout ... --
      //----------------------------------
      if (v2stdout)
      {
        fprintf(stderr, "dump video stream (0x%02x) to stdout\n", vs_id); 
        dmx_vf = fdopen(1, "wb");
      }
      //-- ... or write to file --
      //--------------------------
      else
      {
        if ( (pstr = strrchr(fname, '/')) == NULL ) pstr = fname;
        else pstr++;
      
        sprintf(full_fn, "%s%s%s", dir, delim, pstr);
        fprintf(stderr, "dump video stream (0x%02x) to %s\n", vs_id, full_fn); 
        dmx_vf = fopen(full_fn, "w+b"); 
      }  
        
      if (!dmx_vf) 
      {
        perror("video out");
        return 0;
      }
        
    }
    
    pstr = strrchr(fname, '.');
    if (pstr) *pstr = '\0'; 
    strcat(fname, ".mpa");
    if (as_id) 
    {
      if ( (pstr = strrchr(fname, '/')) == NULL ) pstr = fname;
      else pstr++;

      sprintf(full_fn, "%s%s%s", dir, delim, pstr);
      fprintf(stderr, "dump audio stream (0x%02x) to (%s)\n", as_id, full_fn);
      dmx_af = fopen(full_fn, "w+b"); 
      if (!dmx_af) 
      {
        perror("audio out");
        return 0;
      }
    }
        
    //-- full scan loop --
    //--------------------
    n = 0;
    while ( (ch = fgetc(vf)) != EOF)
    {
      code = (code << 8) | ch;
     
      if ( (code & CODE_PREFIX_MASK) == CODE_PREFIX ) 
      {
        switch (code)
        {
          case PACK_START_CODE:
            ch   = fgetc(vf);
            code = (code << 8) | ch;
              
            if (stream->mpeg2_flag == -1)
              stream->mpeg2_flag = ((ch >> 6) == 1);
            
            stream->pack_cnt++;
            break;
       
          default:
            switch(demuxPESdata(code, &d_ofs, &d_len) & 0xF0)
            {
              case VID_CODE:
                if (vs_id) fwrite(&pes_buf[d_ofs], d_len, 1, dmx_vf);
                break;
              
              case A_TYPE_MPEG1_L2:
              case A_TYPE_A52_SUB:
              case A_TYPE_PCM_SUB:
                if (as_id) fwrite(&pes_buf[d_ofs], d_len, 1, dmx_af);
                break; 
                
            }
            break;   
        }
        
        if (n%100 == 0) 
          fprintf(stderr, "  A:(%d) / V:(%d)\r", stream->as_cnt, stream->vs_cnt);
        n++;
      }  
    }
    
    if (dmx_vf && !v2stdout) fclose (dmx_vf);
    if (dmx_af) fclose (dmx_af);
  }
  else
  {
    fprintf(stderr, "ERROR: unable to open (PS) media (%s)\n", fname);
    return 0;
  }    

  fprintf(stderr, "\n");    
  return 1;
}

#if 0
int T_DEMUXER::remuxAV(char *dir, char *name, int v2stdout, uint8_t vs_id, uint8_t as_id)
{
  int32_t n, d_ofs, d_len; 
  int     ch;
  T_CODE  code = (CODE_PREFIX_MASK >> 8);

  stream->vs_id = vs_id;
  stream->as_id = as_id;

  fprintf(stderr, "remuxing media file (%s) ...\n", fname);

  //-- read from stdin or file --
  //----------------------------- 
  if (!strcmp(fname, "-"))
  { 
    if (strlen(name))
      strcpy(fname, name); 
    else
      strcpy(fname,"out"); 
    vf = fdopen(0,"rb");
  }
  else
    vf = fopen(fname,"rb");
    
  if (vf != NULL)
  {
    //char full_fn[2048];
    char *pstr;
    
    int  l        = strlen(dir);
    char delim[2] = "/";
    
    if ( (l && dir[l-1]=='/') || (l == 0) ) strcpy(delim,"");
    
    
    //-- is there user defined name --
    if (strlen(name)) strcpy(fname, name);
    
    pstr = strrchr(fname, '.');
    if (pstr) *pstr = '\0';
    strcat(fname, ".mxx");
    
    if ( !ffmpeg_open(fname, 10000000, 40.0, 224000, 48000 ) )
    {
      return 0;
    }
        
    //-- full scan loop --
    //--------------------
    n = 0;
    while ( (ch = fgetc(vf)) != EOF)
    {
      code = (code << 8) | ch;
     
      if ( (code & CODE_PREFIX_MASK) == CODE_PREFIX ) 
      {
        switch (code)
        {
          case PACK_START_CODE:
            ch   = fgetc(vf);
            code = (code << 8) | ch;
              
            if (stream->mpeg2_flag == -1)
              stream->mpeg2_flag = ((ch >> 6) == 1);
            
            stream->pack_cnt++;
            break;
       
          default:
            switch(demuxPESdata(code, &d_ofs, &d_len) & 0xF0)
            {
              case 0xE0:
                if (vs_id) fwrite(&pes_buf[d_ofs], d_len, 1, dmx_vf);
                break;
              
              case 0xC0:
              case 0x80:
              case 0xA0:
                if (as_id) fwrite(&pes_buf[d_ofs], d_len, 1, dmx_af);
                break; 
                
            }
            break;   
        }
        
        if (n%100 == 0) 
          fprintf(stderr, "  A:(%d) / V:(%d)\r", stream->as_cnt, stream->vs_cnt);
        n++;
      }  
    }
    
    if (dmx_vf && !v2stdout) fclose (dmx_vf);
    if (dmx_af) fclose (dmx_af);
  }
  else
  {
    fprintf(stderr, "ERROR: unable to open (PS) media (%s)\n", fname);
    return 0;
  }    

  fprintf(stderr, "\n");    
  return 1;
}
#endif

int main(int argc, char **argv)
{
  int     i;
  int     v2stdout = 0;
  uint8_t v_id     = VID_CODE;
  uint8_t a_id     = A_TYPE_MPEG1_L2;
  char    *dir     = "";
  char    *of_name = "";
  char    *media   = NULL;
  
  if (argc < 2)
  {
    usage:
      fprintf(stderr, "usage: lvedemux [options] <media-file>\n");
      fprintf(stderr, "options: -o <out-dir>   = write output to <out-dir>\n");
      fprintf(stderr, "         -n <name>      = use <name> for out-files\n");
      fprintf(stderr, "         -s             = write video to stdout\n");
      fprintf(stderr, "         -v <stream-id> = select video stream to dump\n");
      fprintf(stderr, "         -a <stream-id> = select audio stream to dump\n");
      fprintf(stderr, "         default stream ids are (0xE0) for video and (0xC0)\n");
      fprintf(stderr, "         for audio, use id (0x00) for no video/audio\n");
  }
  else
  {
    for (i=1; i<argc; i++)
    {
      if (!strcmp(argv[i], "-s"))
        v2stdout = 1;
      else if (!strcmp(argv[i], "-o"))
      {
        i++;
        if (i<argc) dir = argv[i];
      }
      else if (!strcmp(argv[i], "-n"))
      {
        i++;
        if (i<argc) of_name = argv[i];
      }
      else if (!strcmp(argv[i], "-v"))
      {
        i++;
        if (i<argc) v_id = (uint8_t)strtol(argv[i], NULL, 16);
      }
      else if (!strcmp(argv[i], "-a"))
      {
        i++;
        if (i<argc) a_id = (uint8_t)strtol(argv[i], NULL, 16);
      }
      else
        media = argv[i];  
    }
    
    if (media)
    {
      T_DEMUXER fs(media);
      return fs.demuxAV(dir, of_name, v2stdout, v_id, a_id);
    }
    else
      goto usage;
  }
  
  return 0;
}   
