/*
 *  lvedump.c
 *
 *  Copyright (C) Gerhard Monzel - October 2003
 *
 *  This file is part of lve, a linux mpeg video editor
 *      
 *  lve tools are free software; you can redistribute it and/or modify
 *  it under the terms of the GNU General Public License as published by
 *  the Free Software Foundation; either version 2, or (at your option)
 *  any later version.
 *   
 *  lve is distributed in the hope that it will be useful,
 *  but WITHOUT ANY WARRANTY; without even the implied warranty of
 *  MERCHANTABILITY or FITNESS FOR A PARTICULAR PURPOSE.  See the
 *  GNU General Public License for more details.
 *   
 *  You should have received a copy of the GNU General Public License
 *  along with GNU Make; see the file COPYING.  If not, write to
 *  the Free Software Foundation, 675 Mass Ave, Cambridge, MA 02139, USA. 
 *
 */
#include <stdio.h>
#include <stdlib.h>
#include <sys/types.h>
#include <sys/stat.h>
#include <unistd.h>
#include <string.h>
#include "decoder.h"
#include "lve_read.h"
#include "ffmpeg_aenc.h"
#include "mux_out.h"

enum {FF_ADEC_ONLY=0, FF_AENC_ONLY, FF_AENC_MUX, FF_ACOPY_MUX, FF_ACOPY_ONLY};

#define FN_FIFO "/tmp/requant"

static T_LVE_READ_CTX *lve_ctx       = NULL;
static EncoderCtx     *enc_ctx       = NULL;
static PackStream     *mux_ctx       = NULL;
static int            force_pal      = 0;
static int	          ext_wav        = 0;
static int            ff_mode        = FF_ADEC_ONLY;
static int            muxType        = MX_TYPE_PS;   
static uint8_t        adjAspect      = 0; // 0=none, (2<<4)=4:3, (3<<4)=16:9
   
#define V_CHUNK_BUF_LEN (256*1024)
#define A_CHUNK_BUF_LEN (12*1920)
static uint8_t v_chunk_buf[V_CHUNK_BUF_LEN];
static uint8_t a_chunk_buf[A_CHUNK_BUF_LEN];   
static uint8_t seq_end_code[] = {0x00, 0x00, 0x01, MX_SEQ_END};
static uint8_t seq_start_code [] = {0x00, 0x00, 0x01, MX_SEQ_CODE};
static uint8_t gop_start_code [] = {0x00, 0x00, 0x01, MX_GOP_CODE};

int main(int argc, char **argv)
{
  int  rc; 
  char str[1024];
  
  uint8_t audio_id     = AUDIO_ID_MP2;  // default
  double  shrink_f     = 1.0;
  int     needAudio    = 0;
  int     a_shift      = 0;
  int     force_copy   = 0;
  int     force_atrack = -1;
  
  int  total_frames    = 0;
  int  start_frame     = 0;
  int  total_real      = 0;  
  int  start_real      = 0;
  int  dst_sr          = 0;
  int  fields          = 0;
  int  cb_ofs          = 0; 
  int  ns_temporal_ref = -1;   
    
  T_VIDEO_PARMS newParms;
  T_VIDEO_PARMS *vP = &newParms;
   
  T_MEDIA_BUF   act_scene;

  FILE *vfp = NULL;
  FILE *afp = NULL;
  FILE *lfp = NULL;
  char *vfn = NULL;
  char *afn = NULL;
  char *ifn = NULL;
  char *lfn = NULL;

  rc = 0;
  
  //-- parse options --
  //------------------- 
  for (int i=1; i<argc; i++)
  {
    if ( !strcmp("-c", argv[i]) )
      force_copy = 1;
    if ( !strcmp("-M", argv[i]) )
      ff_mode = FF_ACOPY_MUX;
    if ( !strcmp("-m", argv[i]) )
      ff_mode = FF_AENC_MUX;
    else if ( !strcmp("-pes", argv[i]) )
    {
      muxType = MX_TYPE_PES;
      fprintf(stderr, "INFO: \'PES\' multiplex activated\n");
    }
    else if ( !strcmp("-ts", argv[i]) )
    {
      muxType   = MX_TYPE_TS;
      fprintf(stderr, "INFO: \'TS\' multiplex activated\n");
    }
    else if ( !strcmp("-E", argv[i]) )
      ff_mode = FF_ACOPY_ONLY;  
    else if ( !strcmp("-e", argv[i]) )
      ff_mode = FF_AENC_ONLY;
    else if ( !strcmp("-w", argv[i]) )
      ext_wav = 1;  
    else if ( !strcmp("-p", argv[i]) )
    { 
      force_pal = 1;
      fprintf(stderr, "INFO: output forced to PAL (= 25 fps)\n");
    }  
    else if ( !strcmp("-t", argv[i]) )
    {
      if ( (i+1) < argc ) 
      {
        force_atrack = atoi(argv[i+1]);
        fprintf(stderr, "INFO: using audio track (%d)\n", force_atrack);
      }
    }
    else if ( !strcmp("-sh", argv[i]) )
    {
      if ( (i+1) < argc ) a_shift = atoi(argv[i+1]);
      if (!a_shift) a_shift = -10;  
    }
    else if ( !strcmp("-i", argv[i]) )
    {
      if ( (i+1) < argc ) ifn = argv[i+1];
    }
    else if ( !strcmp("-av", argv[i]) )
    {
      if ( (i+1) < argc ) vfn = argv[i+1];
    }
    else if ( !strcmp("-a", argv[i]) )
    {
      if ( (i+1) < argc ) afn = argv[i+1];
    }
    else if ( !strcmp("-l", argv[i]) )
    {
      if ( (i+1) < argc ) lfn = argv[i+1];
    }
    else if ( !strcmp("-r", argv[i]) )
    {
      if ( (i+1) < argc ) shrink_f = atof(argv[i+1]);
    }
    else if ( !strcmp("-as", argv[i]) )
    {
      if ( (i+1) < argc ) adjAspect = atoi(argv[i+1])<<4;
    }
    else if ( !strcmp("-s", argv[i]) )
    {
      if ( (i+1) < argc ) 
      {
        dst_sr = atoi(argv[i+1]);
        if (( dst_sr != 32000) && (dst_sr != 44100) && (dst_sr != 48000) )
        {
          fprintf(stderr, "WARNING: wrong sample rate (%d) - ignored !\n", dst_sr);  
          dst_sr = 0;
        }
      }  
    }
  }

  if (force_copy)
  {
    if (ff_mode == FF_AENC_MUX) ff_mode = FF_ACOPY_MUX; 
    else if (ff_mode == FF_AENC_ONLY) ff_mode = FF_ACOPY_ONLY;
  }
  
  //-- check min. requirements --
  //-----------------------------  
  if (!ifn || !vfn)
  {
    fprintf(stderr, 
      "\nusage: lvedump -i <src> -av <video/audio> [-a <audio>] [options]\n\n"
      "options are:\n" 
      "  -l <list>  to write lve edit-list for destination media\n" 
      "  -c forces copy of audio data instead reencoding\n"
      "     only useful together with option \"-m\" or \"-e\" \n" 
      "  -m to compress audio and mux with video\n"
      "  -M shortcut for \"-c -m\"\n"
      "  -pes to force PES packet muxing (new)\n"
      "  -e to compress audio in separate file\n"
      "  -E shortcut for \"-c -e\"\n"
      "  -s <sample-rate> valid rates: 32000, 44100, 48000\n"
      "  -t <track#> to select another audio track (use with caution)\n"
      "  -p to force PAL output (=25fps) \n"
      "  -as <i>, where 2 for 4:3 or 3 for 16:9\n"
      "  -w to force using of external WAV-file\n"
      "  -sh [<value>] to shift audio by <value> video frames\n\n"
      "  NOTE: for options \"-m\",\"-M\" output can be redirected to\n"
      "  stdout using a \"-\" char for output filename (\"-av -\").\n" 
    );     
    return 0;
  }
  
  // TODO: testing some option constraints !!!
  
  //-- do libavformat initaization here --
  //--------------------------------------
  //ffmpeg_init_avformat();
  
  //-- check for valid input-media --
  //---------------------------------
  if (lr_file_chk(ifn))
  {
    CodecID aid; 
    
    lve_ctx = lr_init(force_atrack, a_shift);
    lve_ctx->md_bup = &act_scene;

    if (dst_sr == 0) dst_sr = lve_ctx->sample_rate;
    
    //-- resampling implies rencoding ! --
    if (dst_sr != lve_ctx->sample_rate) 
    {
      switch(ff_mode)
      {
        case FF_ACOPY_ONLY:
          ff_mode = FF_AENC_ONLY;
          break;
          
        case FF_ACOPY_MUX:
          ff_mode = FF_AENC_MUX;
          break;
      }
    }
    
    aid      = (CodecID)lr_acodec_id(); // !!!
    audio_id = (aid == CODEC_ID_AC3)? AUDIO_ID_AC3:AUDIO_ID_MP2;
  }
  else
  {
    fprintf(stderr, "ERROR: (%s) is no lve edit-list\n", ifn);
    return 0;
  }
   
  //-- force external WAV-audio ? --
  if (ext_wav) lve_ctx->has_audio = 1; 
  
  //-- input has audio ... --
  //-------------------------     
  if ( lve_ctx->has_audio )
  {
    switch (ff_mode)
    {
     //-- video and (compressed) audio in separate files --
     //----------------------------------------------------
      case FF_AENC_ONLY: 
        fprintf(stderr, "INFO: reencoding audio (%d Hz)\n", dst_sr);
        
        enc_ctx  = ffmpeg_aencode_open(224000, dst_sr);
        audio_id = AUDIO_ID_MP2;
        
        //-- break only on error !!! --
        if (!enc_ctx) break;
      
      case FF_ACOPY_ONLY:
      case FF_ADEC_ONLY:
        if (!afn)
        {
          strcpy(str, vfn);
          strcat(str, (ff_mode == FF_ADEC_ONLY)?".wav":".mpa");
          afn = str;
          
          fprintf(stderr, "INFO: using audio file (%s) for output\n", afn);
          afp = fopen(afn, "w+b");
          
          strcpy(str, vfn);
          strcat(str, ".m2v");
          vfn = str;
        }
        else
        {
          fprintf(stderr, "INFO: using audio file (%s) for output\n", afn);
          afp = fopen(afn, "w+b");
        }
        
        fprintf(stderr, "INFO: using video file (%s) for output\n", vfn);
        
        //-- check for requantizer mode --
        //--------------------------------
        if (shrink_f != 1.0)
        {
          char        cmd[1024];
          struct stat stat_buf;
  
          //-- check (create) fifo --
          if ( stat(FN_FIFO, &stat_buf) != -1 )
          {
            if (!S_ISFIFO(stat_buf.st_mode))
            fprintf(stderr, "ERROR: %s: already exists but is no FIFO!\n", FN_FIFO); 
          }
          else
          {
            sprintf(cmd, "mkfifo %s", FN_FIFO);
            system(cmd);
          }
          
          //-- video stream floats through requantizer fifo into output file --
          sprintf(cmd, "/usr/bin/lverequant %2.1f %s %s&", 
                  shrink_f, FN_FIFO, vfn);
          fprintf(stderr, "[0] launch: %s\n", cmd);
          system(cmd);
         
          //-- therefore redirect video writing -- 
          strcpy(str, FN_FIFO);
          vfn = str;
        }
        
        //-- now open video output --
        vfp = fopen(vfn, "w+b");
          
        if (afp && vfp) 
        {
          lr_init_wav(afp, (ff_mode==FF_ADEC_ONLY)? dst_sr:0 ); 
          rc = 1;
        }  
        break;
        
      //-- video and compressed audio muxed in one file --
      //--------------------------------------------------
      case FF_AENC_MUX:
        fprintf(stderr, "INFO: reencoding audio (%d Hz)\n", dst_sr);
        
        enc_ctx  = ffmpeg_aencode_open(224000, dst_sr);
        //enc_ctx  = ffmpeg_aencode_open(384000, dst_sr);
        audio_id = AUDIO_ID_MP2; // encode is MP2 only
        
        //-- break only on error !!
        if (!enc_ctx) break;
        
      case FF_ACOPY_MUX: 
        strcpy(str, vfn);
        if ( strcmp(str, "-") && !strstr(str,"mux") && !strchr(str, '.') ) 
          strcat(str, ".mpg");
        vfn = str;
        
        //mux_ctx = mux_open(vfn, 
        //                   lve_ctx->v_bitrate, lve_ctx->frame_rate, 
        //                   lve_ctx->a_bitrate, dst_sr, audio_id);
        // force fixed DVD muxrate 
        mux_ctx = mux_open(vfn, 0, lve_ctx->frame_rate, 0, dst_sr, audio_id);
        mux_ctx->mux_type   = muxType;   // new 
        
        if (mux_ctx) rc = 1;
        fprintf(stderr, "INFO: muxing audio and video to file (%s)\n", vfn);
        break;
        
      default:
        rc = 0;  
    }
  }
  //-- input has video only --
  //--------------------------
  else
  {
    ff_mode = FF_ADEC_ONLY;
    vfp     = fopen(vfn, "w+b");
    
    if (vfp) rc = 1;
    
    fprintf(stderr, "INFO: dumping video only to file (%s)\n", vfn);
  }  
  
  if (rc)   
  {
    T_PIC_HDR *ph;
    
    int p, drop;
    int need_start_real = 1;

    vP->aspect = adjAspect;

    //-- force PAL video --
    //---------------------
    if (force_pal)
    {
      vP->frame_rate = 25.0; // -> GOP-timecode will be adjusted to PAL     
      vP->rate_code  = 3;    // -> != 0 will force patch of rate code
    }
    else
    {
      vP->frame_rate = lve_ctx->frame_rate; // -> GOP-timecode will adjusted normally 
      vP->rate_code  = 0;                   // -> no patch of frame rate code
    }

    //-- ... loop until EOF (all frames in edit list) --
    //--------------------------------------------------
    while (1)
    {
      //-- read frame chunk(s) synced at previous GOP start --
      //------------------------------------------------------
      p = lve_ctx->act_real;
      do
      {
        #define CB_LEN  lve_ctx->buf_len
        #define CB      lve_ctx->chunk_buf
        
        //-- set act. scene state and act. frame count -- 
        vP->ns      = lve_ctx->new_scene;
        vP->nFrames = total_real + 1;
        
        //-- VIDEO --
        //-----------
        
        //-- read chunk data into 'lve_ctx->chunk_buf'-buffer --
        //-- also processing some video parameters included.  --
        CB = v_chunk_buf+cb_ofs;
        p  = lr_get_vchunk(lve_ctx, p, 1, vP);
        
        fields++;
        
        //-- on scene start ... --
        if (vP->ns)
        {
          drop = 1; // ... activate drop mode.
          
          fprintf(stderr, "  -> %d: picture structure: (%s)\n",
                  total_real, (vP->separateFields)? "2 fields":"frame");
          
          if (vP->ph)
          {
            ns_temporal_ref = vP->ph->temporal_ref;
            if (!ns_temporal_ref)
            {
              fprintf(stderr, "  -> closed GOP -> no B drops needed\n");
              drop = 0;
            } 
          }
          else
            ns_temporal_ref = -1;
        }
        //-- acting in drop mode ... --
        else if (drop)
        {
          //-- ... drops until I- or P-Frame --
          if (vP->ph) 
          {
            if ( (ns_temporal_ref != -1) && 
                 (ns_temporal_ref == vP->ph->temporal_ref) )
              fprintf(stderr, "  -> 2nd. field not dropped\n");
            else if (vP->ph->type == B_FRAME)
              CB_LEN = 0;
            else
              drop = 0;
          }
          else
            drop = 0;    
        }
        
        if (CB_LEN)
        {
          //-- remove EOS pattern to prevent (DVD-)player stop --
          if (!memcmp(&CB[CB_LEN-4], seq_end_code, 4))
          {
            fprintf(stderr,"\nEOS-pattern deleted ...\n");
            CB_LEN -= 4;  
          }
          
          //-- check for 1st. sequence header --
          if (total_real==0)
          {
            int     slen = 0;
            uint8_t *shdr;
            
            if ( memcmp(CB, seq_start_code, 4) && (shdr = lr_mSeqHdr(&slen)) != NULL )
            {
              fprintf(stderr, "INFO: inserting missing sequence header to start of file\n");
              
              //-- patch frame rate code --
              T_FRAME_SRV::patchSEQ(&shdr[4], vP);
              
              //-- insert header --
              memmove(CB+slen, CB, CB_LEN);
              memcpy(CB, shdr, slen);
              CB_LEN += slen;
            }
              
          }
          else if (vP->aspect && !memcmp(CB, seq_start_code, 4))
            T_FRAME_SRV::patchSEQ(CB+4, vP);
          
          //-- count real frames of destination media --  
          total_real++;
        
          if ((total_real % 100) == 0)
            fprintf(stderr, "  (%d) frames processed ...\r", total_real);
        
          
          //-- mux video ... --
          if (ff_mode == FF_AENC_MUX || ff_mode == FF_ACOPY_MUX)
          {
            mux_ctx->audio_encoded_fs = 0;
            if (vP->separateFields)
            {
              //-- first field will stay in buffer --
              if (fields==1)
              {
                cb_ofs = CB_LEN;
              }
              //-- if 2nd. field starts withs SEQ or GOP code   --  
              //-- split mux-write to allow VOBU pack insertion --
              else if ( !memcmp(CB, seq_start_code, 4) 
                        || !memcmp(CB, gop_start_code, 4) )   
              {
                mux_write_packet(mux_ctx, VIDEO_ID, &v_chunk_buf[0], cb_ofs);
                mux_write_packet(mux_ctx, VIDEO_ID, &v_chunk_buf[cb_ofs], CB_LEN);
                cb_ofs = 0;
              }
              //-- or write both fields at once --
              else
              {
                mux_write_packet(mux_ctx, VIDEO_ID, &v_chunk_buf[0], cb_ofs+CB_LEN);
                cb_ofs = 0;
              }  
            }
            else
              mux_write_packet(mux_ctx, VIDEO_ID, CB, CB_LEN);
          }   
          //-- ... or dump video only --
          else
            fwrite(CB, CB_LEN, 1, vfp);
        
          if ( !vP->separateFields || (fields > 1) ) 
          {  
            needAudio = 1;
            fields    = 0;
          }
        }
        
        //-- AUDIO --
        //----------- 
        if (needAudio && lve_ctx->has_audio)
        {
          needAudio = 0;

          CB = a_chunk_buf;
          if (ff_mode == FF_ACOPY_MUX  || ff_mode == FF_ACOPY_ONLY)
          {
            CB_LEN = lr_get_achunk(lve_ctx);
            if (mux_ctx) mux_ctx->audio_encoded_fs = lve_ctx->encoded_fs; 
          }  
          else
          {
            //-- read decoded samples into 'lve_ctx->chunk_buf'-buffer --
            //-- everytime adjusted to PAL Video duration (forced) and --
            //-- set amount of bytes in buffer to 'lve_ctx->buf_len'.  --                
            CB_LEN = lr_get_samples(lve_ctx, CB, force_pal);
     
            //-- resample audio, if destination rate differs from source -- 
            if (lve_ctx->sample_rate != dst_sr)
              CB_LEN = lr_resample(lve_ctx, dst_sr, vP->ns);
          }
            
          if (CB_LEN) switch (ff_mode)
          {
            //-- encode audio and mux --
            case FF_AENC_MUX:
              CB_LEN = ffmpeg_aencode(enc_ctx, CB, CB_LEN, A_CHUNK_BUF_LEN);
              mux_ctx->audio_encoded_fs = enc_ctx->encoded_fs;
              // no break !!
            
            //-- copy audio and mux --  
            case FF_ACOPY_MUX:  
              mux_write_packet(mux_ctx, audio_id, CB, CB_LEN);
              break;
              
            //-- encode audio to file --  
            case FF_AENC_ONLY:
              CB_LEN = ffmpeg_aencode(enc_ctx, CB, CB_LEN, A_CHUNK_BUF_LEN); 
              // no break !!
                  
            //-- copy audio to file --  
            FF_ACOPY_ONLY:
            FF_ADEC_ONLY:
            default:
              fwrite(CB, CB_LEN, 1, afp);
              break;
          }
          else
          {
            switch (ff_mode)
            {
              //-- muxer need flush of buffers --
              case FF_AENC_MUX:
              case FF_ACOPY_MUX:
                mux_write_packet(mux_ctx, audio_id, CB, 0);
                break;
  
            }    
            fprintf(stderr, "\nINFO: using silence for audio\n");
          }  

        }
        
        //-- scene change is fully processed here --
        lve_ctx->new_scene = 0;
        
        //-- next source picture --
        p++;

      } while (p <= lve_ctx->act_real);
    
      if (need_start_real) 
      {
        start_real      = total_real-1;
        need_start_real = 0;
      }
      
      //-- count visible frames of destination media --
      total_frames++;
          
      //-- check for scene change / EOF --
      //----------------------------------
      lr_media_chk_range(lve_ctx, 0);
      if (ext_wav) lve_ctx->has_audio = 1; // force external WAV
      
      //-- scene change (or EOF) ... --
      //------------------------------- 
      if ( lve_ctx->new_scene || (lve_ctx->read_mode == SC_EOF) )
      {
        //-- have to create destination edit-list ? --
        //--------------------------------------------   
        if (lfn)
        {        
          if (!lfp)
          {
            //-- open edit-list file for destination    --
            //-- and write header from source edit-list --
            lfp = fopen(lfn,"w+b");
            if (!lfp) 
            {
              fprintf(stderr, "ERROR: unable to open (%s)\n", lfn);
              lve_ctx->read_mode = SC_EOF;
              break;
            }  
            
            fwrite(&lve_ctx->md_header, sizeof(T_MEDIA_HDR), 1, lfp);
          }
        
          act_scene.media_id   = 0;
          act_scene.start_real = start_real;
          act_scene.end_real   = total_real-1;
          act_scene.start      = start_frame;
          act_scene.end        = total_frames-1;
          act_scene.dist       = act_scene.end - act_scene.start + 1; 
          strcpy(act_scene.fname, vfn);
   
          //-- write scene record --
          fwrite(&act_scene, sizeof(T_MEDIA_BUF), 1, lfp);
        
          if (lve_ctx->read_mode == SC_EOF) fclose(lfp);
         
          start_frame     = total_frames;
          need_start_real = 1;
        }
      }  
      //-- ... or next continoues real frame of source media --
      //-------------------------------------------------------
      else 
        lve_ctx->act_real++;

      //-- count visible frames of source media --
      lve_ctx->total_frames++;
    
      //-- check for EOF --
      if (lve_ctx->read_mode == SC_EOF) break;
    }    

    if (mux_ctx) mux_close(mux_ctx);
    if (enc_ctx) ffmpeg_aencode_close(enc_ctx);
  }
  else
    fprintf(stderr, "ERROR: creating output file(s) !\n");

  lr_cleanup();
}    
  

