/*
 *  lvemux.c
 *
 *  Copyright (C) Gerhard Monzel - October 2003
 *
 *  This file is part of lve, a linux mpeg video editor
 *      
 *  lve tools are free software; you can redistribute it and/or modify
 *  it under the terms of the GNU General Public License as published by
 *  the Free Software Foundation; either version 2, or (at your option)
 *  any later version.
 *   
 *  lve is distributed in the hope that it will be useful,
 *  but WITHOUT ANY WARRANTY; without even the implied warranty of
 *  MERCHANTABILITY or FITNESS FOR A PARTICULAR PURPOSE.  See the
 *  GNU General Public License for more details.
 *   
 *  You should have received a copy of the GNU General Public License
 *  along with GNU Make; see the file COPYING.  If not, write to
 *  the Free Software Foundation, 675 Mass Ave, Cambridge, MA 02139, USA. 
 *
 */
#include <stdio.h>
#include <stdlib.h>
#include <unistd.h>
#include <string.h>
#include <inttypes.h>
#include <pthread.h>
#include "mux_out.h"

//#define FPRINTF(a,b) fprintf(a,b);
#define FPRINTF(a,b) ;
 
#define CHUNK_BLOCK_LEN 2048
#define VC_BUF_LEN      (512*CHUNK_BLOCK_LEN)  // min. 1024 K buf size
#define AC_BUF_LEN      (128*CHUNK_BLOCK_LEN)  // min.  256 K buf size

#define F_VIDEO 0
#define F_AUDIO 1
#define F_MAIN  2

#define IS_FILE_MODE 0
#define IS_PIPE_MODE 1

typedef struct
{
  int             type;
  int             size;
  uint8_t         *start;
  uint8_t         *end;
  uint8_t         *read;
  uint8_t         *write;
  pthread_t       th;
  //int             rmode;         
  FILE *          fp;
  int             eof;
  int             reset;
  int             undo;
} Fifo;

typedef struct
{
  Fifo vf;
  Fifo af;
  Fifo mf;

  pthread_mutex_t vf_mutex;
  pthread_cond_t  vf_cond;
  pthread_mutex_t af_mutex;
  pthread_cond_t  af_cond;

  int  mf_working;
  int  rmode;
} FifoList;

typedef struct
{
  uint8_t  chunk_buf[VC_BUF_LEN];
  uint8_t  remaining[4];
  uint8_t  id;
  uint32_t code;
  uint32_t sc;
  int      frame_len;
  int      synced;
  FifoList *fl;
  
  uint8_t  chunk_block_buf[CHUNK_BLOCK_LEN];
  uint8_t  *block_pbuf;
  uint8_t  *block_end;

  int      unused;
  int      used;
  int      eof;

  int      width, height;
  char     *as;
  int      br;
  double   fr;
  int      fields;

} VideoParser;

typedef struct
{
  uint8_t  chunk_buf[AC_BUF_LEN];
  uint8_t  remaining[4];
  uint8_t  id;
  int      hdr_len;
  int      size_pf;  // alias audio_encoded_fs
  int      synced;
  FifoList *fl;
  
  uint8_t  chunk_block_buf[CHUNK_BLOCK_LEN];
  uint8_t  *block_pbuf;
  uint8_t  *block_end;

  int      unused;
  int      used;
  int      eof;

  int      frame_len;
  int      sr;
  int      br;
  int      samples_pf;
  

} AudioParser;



//=============
//== F I F O ==
//=============

#define VF_BLOCK_LEN CHUNK_BLOCK_LEN  // block size per fifo-read
#define AF_BLOCK_LEN CHUNK_BLOCK_LEN
#define VS_BUF_LEN   (6*VC_BUF_LEN)     // fifo buffer size
#define AS_BUF_LEN   (6*AC_BUF_LEN)  
 
//-- buffers for (fifo) stream reader threads --
//---------------------------------------------- 
static uint8_t vs_buffer[VS_BUF_LEN];
static uint8_t as_buffer[AS_BUF_LEN];
static uint8_t pulldown=0;
//-- main context for thread/fifo-handling --
//-------------------------------------------
static FifoList fifos =
{
  {F_VIDEO, VS_BUF_LEN, vs_buffer},
  {F_AUDIO, AS_BUF_LEN, as_buffer},
  {F_MAIN, 0, NULL},
  PTHREAD_MUTEX_INITIALIZER, 
  PTHREAD_COND_INITIALIZER,
  PTHREAD_MUTEX_INITIALIZER, 
  PTHREAD_COND_INITIALIZER,
  0,
  IS_FILE_MODE
};

//-- jump to start of file/fifo --
void rewind_video(FifoList *fl)
{
  if (fl->rmode == IS_FILE_MODE) fseeko(fl->vf.fp, 0, SEEK_SET);
}

void rewind_audio(FifoList *fl)
{
  if (fl->rmode == IS_FILE_MODE) fseeko(fl->af.fp, 0, SEEK_SET);
}

int read_video(uint8_t *pdst, size_t size, int cnt, FifoList *fl)
{
 Fifo *vf = &fl->vf;
 Fifo *af = &fl->af;
 Fifo *mf = &fl->mf;
  
 int len = size * cnt;
 int n;
  
 if (fl->rmode == IS_FILE_MODE) 
  return fread(pdst, size, cnt, vf->fp);

 for(;;)
 {
  //-- let video stream reader reach the wait state first --
  //--------------------------------------------------------
  if (!fl->mf_working) pthread_mutex_lock(&fl->vf_mutex);
  fl->mf_working = 1;
  
  //-- read, (hold) lock and return, if enough available -- 
  //-------------------------------------------------------
  n = vf->write - vf->read;
  if (len < n) 
  {
    n = len;
    memcpy(pdst, vf->read, n);
    if (vf->undo)
      vf->undo = 0;
    else
      vf->read += n;
    return cnt;
  }
  
  if (vf->eof)
  {
    FPRINTF(stderr, "\nvreader: eof reached\n");
    memcpy(pdst, vf->read, n);
    vf->read = vf->write;
    return (n/size);
  }
  
  //-- wakeup video stream reader now --
  //------------------------------------
  FPRINTF(stderr, "\nvreader: buffer empty now\n");

  fl->mf_working = 0;
    
  //-- request filling of streams buffer(s) / send signal ... --
  //------------------------------------------------------------
  pthread_cond_signal(&fl->vf_cond);
  FPRINTF(stderr, "vreader: signal to video + wait\n");

  //-- ... and wait for completion --
  //---------------------------------
  pthread_cond_wait(&fl->vf_cond, &fl->vf_mutex);
  pthread_mutex_unlock(&fl->vf_mutex);
  FPRINTF(stderr, "vreader: got signal from video\n");

 }
}

int read_audio(uint8_t *pdst, size_t size, int cnt, FifoList *fl)
{
 Fifo *vf = &fl->vf;
 Fifo *af = &fl->af;
 Fifo *mf = &fl->mf;
  
 int len = size * cnt;
 int n;
  
 if (fl->rmode == IS_FILE_MODE) 
   return fread(pdst, size, cnt, af->fp);

 for(;;)
 {
  //-- let video stream reader reach the wait state first --
  //--------------------------------------------------------
  if (!fl->mf_working) pthread_mutex_lock(&fl->vf_mutex);
  fl->mf_working = 1;
  
  if (vf->eof)
  {
    FPRINTF(stderr, "areader: lock audio stream reader\n"); 
    pthread_mutex_lock(&fl->af_mutex);
  }
  
  //-- read, (hold) lock and return, if enough available -- 
  //-------------------------------------------------------
  n = af->write - af->read;
  if (len < n) 
  {
    n = len;
    memcpy(pdst, af->read, n);
    if (af->undo)
      af->undo = 0;
    else
      af->read += n;
    
    if (vf->eof) pthread_mutex_unlock(&fl->af_mutex);  
    return cnt;
  }
  
  if (af->eof || vf->eof)
  {
    FPRINTF(stderr, "\nareader: eof reached\n");
    memcpy(pdst, af->read, n);
    
    af->read=af->write;
    if (vf->eof) pthread_mutex_unlock(&fl->af_mutex);  
    return (n/size);
  }
  
  //-- wakeup video stream reader now --
  //------------------------------------
  FPRINTF(stderr, "\nareader: buffer empty now\n");

  fl->mf_working = 0;

  //-- request filling of streams buffer(s) / send signal ... --
  //------------------------------------------------------------
  pthread_cond_signal(&fl->vf_cond);
  FPRINTF(stderr, "areader: signal to video + wait\n");

  //-- ... and wait for completion --
  //---------------------------------
  pthread_cond_wait(&fl->vf_cond, &fl->vf_mutex);
  pthread_mutex_unlock(&fl->vf_mutex);
  FPRINTF(stderr, "areader: got signal from video\n");
 } 
}

static void *read_vstream_thread(void *_fl)
{
  FifoList *fl = (FifoList *)_fl;
  Fifo     *vf = &fl->vf;
  Fifo     *af = &fl->af;
  Fifo     *mf = &fl->mf;
  int      n;
  
  //pthread_setcancelstate(PTHREAD_CANCEL_ASYNCHRONOUS, NULL);
  //pthread_setcanceltype(PTHREAD_CANCEL_ENABLE, NULL);
  
  for(;;)
  {
    //-- start reading from video stream (again) --
    //---------------------------------------------
    FPRINTF(stderr, "video: starting ...\n");
    pthread_mutex_lock(&fl->vf_mutex);
    
    //-- fill buffer --
    //-----------------
    FPRINTF(stderr, "video: ... working ...\n");
    while (vf->write < (vf->end - VF_BLOCK_LEN) )
    {
      pthread_testcancel();
      n = fread(vf->write, 1, VF_BLOCK_LEN, vf->fp);
      pthread_testcancel();
      
      vf->write += n;
      if (n<VF_BLOCK_LEN) 
      {
        vf->eof = 1;
        FPRINTF(stderr, "video: ... eof reached ...\n");
        break;
      }  
    }
    FPRINTF(stderr, "video: ... done\n");
    
    //-- pause (lock) audio stream reader too --
    //------------------------------------------ 
    pthread_mutex_lock(&fl->af_mutex);
    FPRINTF(stderr, "video: stream readers deactivated\n");
     
    //-- signal "buffer ready" to main ... --
    //---------------------------------------
    pthread_cond_signal(&fl->vf_cond);
    FPRINTF(stderr, "video: signal to main");
    
    if (vf->eof)
    {
      //-- prepare thread exit --
      //-------------------------
      FPRINTF(stderr, " + end\n");
      pthread_mutex_unlock(&fl->vf_mutex);
    }
    else
    {
      FPRINTF(stderr, " + wait\n");
      
      //-- ... and wait for completion --
      //---------------------------------
      pthread_cond_wait(&fl->vf_cond, &fl->vf_mutex);
      pthread_mutex_unlock(&fl->vf_mutex);
    
      FPRINTF(stderr, "video: signal got from main -> update pointers\n");
      
      //-- move unused + update read/write pointers --
      //----------------------------------------------
      n = vf->write - vf->read;
      if (n) memcpy(vf->start, vf->read, n);

      vf->read  = vf->start;
      vf->write = vf->start + n; 
   
      af->reset = 1;  // request for audio pointer updates
    }
    
    //-- continue (unlock) audio stream reader again --
    //-------------------------------------------------    
    pthread_mutex_unlock(&fl->af_mutex);
    FPRINTF(stderr, "video: stream readers activated ...\n");
  
    if (vf->eof) break; //pthread_exit(NULL);
  }
  fprintf(stderr, "video stream reader ended !\n");
}

static void *read_astream_thread(void *_fl)
{
  FifoList *fl = (FifoList *)_fl;
  Fifo     *vf = &fl->vf;
  Fifo     *af = &fl->af;
  Fifo     *mf = &fl->mf;
  int      n;
  
  //pthread_setcancelstate(PTHREAD_CANCEL_ASYNCHRONOUS, NULL);
  //pthread_setcanceltype(PTHREAD_CANCEL_ENABLE, NULL);
  
  for (;;)
  {
    //-- while enough space available ... --
    //--------------------------------------
    while (af->write < (af->end - AF_BLOCK_LEN) )
    {
      //-- read audio data --
      pthread_testcancel();
      n = fread(af->write, 1, AF_BLOCK_LEN, af->fp);
      pthread_testcancel();
      
      //-- adjust (read-)/write pointer(s) --
      pthread_mutex_lock(&fl->af_mutex);
      if (n<AF_BLOCK_LEN) af->eof = 1;
      
      af->write += n;
      if (af->reset)
      {  
        //-- move unused --
        n = af->write - af->read;
        if (n) memcpy(af->start, af->read, n);
        
        af->read  = af->start;
        af->write = af->start + n;
        af->reset = 0;
      }
      //-- should never happen ! --
      else if (af->write > af->end)
      {  
        fprintf(stderr, "WARNING: audio stream buffer overflow !\n");
        af->read = af->write = af->start;        
      }
      pthread_mutex_unlock(&fl->af_mutex);
      
      if (af->eof) break; //pthread_exit(NULL);
    }
    
    if (af->eof) break;
  }
  fprintf(stderr, "audio stream reader ended !\n");
        
}


FifoList *open_video(char *fn, int rmode)
{
  FifoList *fl = &fifos;
  Fifo     *vf = &(fl->vf);
  
  if ( (vf->fp = fopen(fn, "rb")) != NULL )
  {
    vf->end    = vf->start+vf->size;
    vf->read   = vf->start;
    vf->write  = vf->start;
    vf->eof    = 0;
    vf->reset  = 0;
    vf->undo   = 0;

    fl->rmode = rmode;    
    if (rmode == IS_FILE_MODE) return &fifos;
    
    fprintf(stderr, "video stream reader is starting ...\n");
    if ( !pthread_create(&vf->th, NULL, read_vstream_thread, (void *)&fifos) )
    {
      return fl;
    }
    fclose(vf->fp);
  }
  return NULL;
}

FifoList *open_audio(char *fn, int rmode)
{
  FifoList *fl = &fifos;
  Fifo     *af = &(fl->af);
  
  if ( (af->fp = fopen(fn, "rb")) != NULL )
  {
    af->end    = af->start+af->size;
    af->read   = af->start;
    af->write  = af->start;
    af->eof    = 0;
    af->reset  = 0;
    af->undo   = 0;

    fl->rmode = rmode;
    if (rmode == IS_FILE_MODE) return fl;

    fprintf(stderr, "audio stream reader is starting ...\n");
    if ( !pthread_create(&af->th, NULL, read_astream_thread, (void *)&fifos) )
    {
      return fl;
    }  
    fclose(af->fp);
  }
  return NULL;
}

void close_video(FifoList *fl)
{
  if (fl->rmode == IS_PIPE_MODE)
  {
    if (!fl->vf.eof)
    {
      fl->vf.eof = 1; 
      pthread_cancel(fl->vf.th);
    }
    usleep(50000);
  }  
  fclose(fl->vf.fp);
  
}

void close_audio(FifoList *fl)
{
  if (fl->rmode == IS_PIPE_MODE) 
  {
    if (!fl->af.eof) 
    {
      fl->af.eof = 1;
      pthread_cancel(fl->af.th); 
    }
    usleep(50000);  
  }  
  fclose(fl->af.fp);
}


//===============
//== V I D E O ==
//===============


//== sync to picture relevant start codes ==
static uint8_t *sync_video(VideoParser *vp, uint8_t *sc)
{
  uint8_t  *pbuf = vp->block_pbuf;
  uint32_t code  = vp->code;

  while (pbuf < vp->block_end)
  {
    code = (code<<8) | *pbuf;
    switch (code)
    {
      case MX_SEQ_START_CODE:
      case MX_GOP_START_CODE:
      case MX_PIC_START_CODE:
        *sc      = *pbuf;
        vp->code = code;
        return pbuf;
    }
    pbuf++;
  }
  
  vp->code = code;
  return NULL;
}

static uint8_t *find_startcode(VideoParser *vp, uint32_t sc)
{
  uint8_t  *pbuf = vp->block_pbuf;
  uint32_t code  = vp->code;

  while (pbuf < vp->block_end)
  {
    code = (code<<8) | *pbuf;
    if (code == sc) return pbuf;
    pbuf++;
  }
  return NULL;
}

//== init video frame reader (parser) ==
//======================================
VideoParser *init_video_parser(FifoList *fl)
{
  static const double frtab[] = {25.0, 30000/1001.0, 24.0, 25.0, 30000.0/1001.0,
                                 30.0, 50.0, 60000.0/1001.0, 60.0};
  static const char  *astab[] = {"---", "1:1", "4:3", "16:9", "2.21:1", "---"};
  static const char  as_oor[] = "out of range";
  static VideoParser vp;
  int    n;
   
  vp.fl        = fl;
  vp.id        = VIDEO_ID; 
  vp.synced    = 0;
  vp.frame_len = 0;
  vp.code      = 0xFFFFFF;
  vp.sc        = 0;

  vp.block_pbuf = NULL;
  vp.block_end  = NULL;
  
  vp.used   = 0;
  vp.unused = 0;
  vp.eof    = 1;
  
  vp.width  = -1;
  vp.height = -1; 
  vp.as     = (char *)astab[0];
  vp.fr     = 25.0;
  vp.br     = 10000000;
  vp.fields = 1;
  
  fl->vf.undo = 1;   // fifo: data of following read should remain in buffer
  if ( (n = read_video(vp.chunk_buf, 1, CHUNK_BLOCK_LEN, fl)) == CHUNK_BLOCK_LEN )
  {
    uint8_t *start;
    uint8_t sc;
    
    vp.block_pbuf = vp.chunk_buf;
    vp.block_end  = vp.chunk_buf + n - 128;
    if ( (start = sync_video(&vp, &sc)) != NULL )
    { 
      switch(sc)
      {
        case MX_SEQ_CODE:
          vp.width  = (start[1]<<4)| (start[2]>>4);
          vp.height = ((start[2]&0x0F)<<8) | start[3];
          
          if ( (start[4]>>4) < 6)
             vp.as = (char *)astab[start[4]>>4];
          else
             vp.as = (char *)as_oor;
  
          if ( (start[4] & 0x0F) < 9 )           
            vp.fr = frtab[(start[4] & 0x0F)];
	  else
	    vp.fr = 25.0;
	    
	  if(pulldown)
	  	{
			vp.fr=vp.fr*4./5.;
		}
          vp.br = (start[5]<<10) | (start[6]<<2) | (start[7]>>6);
          vp.br *= (8*50);
             
        case MX_GOP_CODE:
          vp.block_pbuf = start;
          start = sync_video(&vp, &sc);
          if (!start) break;
          
        case MX_PIC_CODE:
          vp.block_pbuf = start;
          vp.block_end  = vp.chunk_buf + n - 8;
          if ( (start = find_startcode(&vp, MX_EXT_START_CODE)) != NULL )
          {
            //-- only on real ext code --
            if ( ((start[1] & 0xF0) == 0x80) && ((start[3] & 0x03) != 0x03) ) 
              vp.fields = 2;
          }
          break;
          
        default:
          break; 
      }
      vp.eof = 0;
    }    
  }
  rewind_video(fl);  // file: seek to start
  
  if (!vp.eof)
  {
    fprintf(stderr, "\nvideo stream type mpeg detected\n");
    
    if (vp.width == -1)
      fprintf(stderr, "WARNING: video stream doesnt start with sequence -> using defaults\n");
    fprintf(stderr, "- size (%dx%d)\n- aspect ratio (%s)\n- framerate (%1.3f Hz)\n- bitrate (%d bps)\n",
            vp.width, vp.height, vp.as, vp.fr, vp.br);
    fprintf(stderr, "- picture structure (%s)\n", 
            (vp.fields==2)? "interlaced fields":"frames");        
    return &vp;
  }
  
  fprintf(stderr, "ERROR: no valid video stream\n");
  return NULL;
}

//== video parser: read chunk data for exactly 1  ==
//== video frame or 2 fields, if interlaced video ==
//==================================================
uint8_t *parse_video(VideoParser *vp)
{
  static uint8_t seq_end_code[] = {0x00, 0x00, 0x01, MX_SEQ_END};
  static uint8_t seq_start_code[] = {0x00, 0x00, 0x01, MX_SEQ_CODE};

  uint8_t *next = NULL;
  uint8_t code1, code2;
  
  int field          = 1;
  int done           = 0;
  int available      = 0;
  int s, n           = 0;
  int unused         = vp->unused;
  uint8_t *pbuf      = vp->chunk_buf;
  uint8_t *pbuf_end  = pbuf;
  uint8_t *start     = NULL;
  uint8_t *end       = NULL;
  
  if (unused) 
  {  
    memmove(pbuf, pbuf+(vp->used), unused);
    vp->used   = 0;
    vp->unused = 0;
  }
  
  vp->frame_len = 0;
  do
  {
    if ( (CHUNK_BLOCK_LEN-unused) > 0)
    {
      s = read_video(pbuf+unused, 1, CHUNK_BLOCK_LEN-unused, vp->fl); 
      n = s + unused;
    }
    else
      n = unused;
    unused = 0;  
    
    available += n;
    pbuf_end  += n;
  
    if (!start)
    {
      vp->block_pbuf = pbuf;
      vp->block_end  = pbuf_end;
      if ( (next = sync_video(vp, &code1)) != NULL )
      {
        end    = next + 1; 
        start  = end - 4;
      }
    }
    
    if (end)
    {
      vp->block_end = pbuf_end;
      
      do
      {
        vp->block_pbuf = end;
        if ( (next = sync_video(vp, &code2)) != NULL ) 
        {
          end = next + 1;
        
          if (code1 == code2 || code1 == MX_PIC_CODE)
          { 
            if ( (vp->fields==2) && (field==1) && (code2==MX_PIC_CODE) )
            {
              field++;
            }
            else
            {
              vp->frame_len = end - 4 - start;
              done = 1;
              break;;
            }
          }
          code1 = code2;
        }  
      } while (next);
    
      end = pbuf_end;
    }

    pbuf = pbuf_end;
  
  } while (s && !done && (available >= CHUNK_BLOCK_LEN) ); 

  if (vp->frame_len)  
  {
    vp->used   = start - vp->chunk_buf + vp->frame_len;
    vp->unused = available - vp->used;
    
    //-- check for picture structure on sequence starts --
    if ( !memcmp(start, seq_start_code, 4) )
    {
      uint8_t *p;
      
      vp->block_pbuf = start + 4;
      vp->block_end  = start + vp->frame_len - 128;
      if ((p = sync_video(vp, &code1)) != NULL)
      {
        vp->block_pbuf = p;
        vp->block_end  = start + vp->frame_len - 8;
      
        if ( (p = find_startcode(vp, MX_EXT_START_CODE)) != NULL )
        {
          //-- only on real ext code --
          if ( (p[1] & 0xF0) == 0x80 ) 
          {
            if( (p[3] & 0x03) != 0x03 ) 
            {  
              if ((vp->fields == 1) && (vp->fl->rmode == IS_PIPE_MODE) )
              {
                fprintf(stderr, "ERROR: frames to fields change detected -> forcing eof\n");
                vp->eof = 1;
              }
              else
                vp->fields = 2; 
            }
            else
            {
              if ( (vp->fields == 2) && (vp->fl->rmode == IS_PIPE_MODE) )
              {
                fprintf(stderr, "ERROR: fields to frames change detected -> forcing eof\n");
                vp->eof = 1;
              }
              else
                vp->fields = 1;
            }
          }      
        }
      }  
    }
    
    //-- remove sequence end codes --
    if ( !memcmp(&start[vp->frame_len-4], seq_end_code, 4) ) 
    {
      fprintf(stderr, "sequence end code detected -> skipping!\n");
      vp->frame_len -= 4;
      if (vp->frame_len < 0) vp->frame_len = 0;
    }  
  }
  else
  { 
    vp->frame_len = available;
    vp->eof       = 1;    
  }
  return start;
}

//===============
//== A U D I O ==
//===============

//-- borrowed from liba52 --
#define A52_DOLBY 10
#define A52_LFE   16

static int a52_syncinfo (uint8_t * buf, int * flags,
                         int * sample_rate, int * bit_rate)
{
    static int rate[] = { 32,  40,  48,  56,  64,  80,  96, 112,
			 128, 160, 192, 224, 256, 320, 384, 448,
			 512, 576, 640};
    static uint8_t lfeon[8] = {0x10, 0x10, 0x04, 0x04, 0x04, 0x01, 0x04, 0x01};
    static uint8_t halfrate[12] = {0, 0, 0, 0, 0, 0, 0, 0, 0, 1, 2, 3};

    int frmsizecod;
    int bitrate;
    int half;
    int acmod;

    if ((buf[0] != 0x0b) || (buf[1] != 0x77))	/* syncword */
	return 0;

    if (buf[5] >= 0x60)		/* bsid >= 12 */
	return 0;
    half = halfrate[buf[5] >> 3];

    /* acmod, dsurmod and lfeon */
    acmod = buf[6] >> 5;
    *flags = ((((buf[6] & 0xf8) == 0x50) ? A52_DOLBY : acmod) |
	      ((buf[6] & lfeon[acmod]) ? A52_LFE : 0));

    frmsizecod = buf[4] & 63;
    if (frmsizecod >= 38)
	return 0;
    bitrate = rate [frmsizecod >> 1];
    *bit_rate = (bitrate * 1000) >> half;

    switch (buf[4] & 0xc0) {
    case 0:
	*sample_rate = 48000 >> half;
	return 4 * bitrate;
    case 0x40:
	*sample_rate = 44100 >> half;
	return 2 * (320 * bitrate / 147 + (frmsizecod & 1));
    case 0x80:
	*sample_rate = 32000 >> half;
	return 6 * bitrate;
    default:
	return 0;
    }
}

//== sync to audio frame relevant start codes ==
static uint8_t *sync_audio(AudioParser *ap)
{
  static int sr[] = {44100, 48000, 32000, 0};
  static int br[] = {0, 32, 48, 56, 64, 80, 96, 112, 128, 160, 192, 224, 256, 320, 384, 0 };
  
  int dmy;

  uint16_t code  = 0; 
  uint8_t  *rc   = NULL;
  uint8_t  *pbuf = ap->block_pbuf;

  while (pbuf < ap->block_end)
  {
    if (ap->id == AUDIO_ID_MP2)
    {
      code = (code << 8) | *pbuf;  
    
      if ( code == 0x0FFF || ((code & 0xFFF0) == 0xFFF0) ) 
      {
        uint32_t parms = ( (pbuf[0]<<16) | (pbuf[1]<<8) | pbuf[2] );
        
        if ( pbuf[-1] == 0xFF ) 
          parms <<= 4;
        else
          parms <<= 8;  
        
        ap->sr = sr[(parms & 0xC000) >> 14];
        ap->br = br[(parms & 0xF0000) >> 16] * 1000;
      
        rc   = pbuf-1;
        break;
      }  
    }
    else
    {
      if ( a52_syncinfo(pbuf, &dmy, &(ap->sr), &(ap->br)) )
      {
        rc   = pbuf;
        break;
      }
    }
    pbuf++;
  }
  
  return rc;
}

//== init audio frame reader (parser) ==
//======================================
AudioParser *init_audio_parser(FifoList *fl)
{
  static AudioParser ap;
  int n;
 
  ap.fl         = fl;
  ap.synced     = 0;
  ap.frame_len  = 0; 
  
  ap.block_pbuf = NULL;
  ap.block_end  = NULL;
  
  ap.used   = 0;
  ap.unused = 0;
  ap.eof    = 0;
  
  ap.id         = AUDIO_ID_AC3;  // default type
  ap.hdr_len    = 7;             // default for AUDIO_ID_AC3
  ap.samples_pf = AC3_FRAME_SIZE;
  ap.sr        = 0;
  ap.br        = 0;

  fl->af.undo = 1;
  if ( (n = read_audio(ap.chunk_buf, 1, CHUNK_BLOCK_LEN, fl)) > ap.hdr_len)
  {
    ap.block_pbuf = ap.chunk_buf;
    ap.block_end  = ap.chunk_buf+n-ap.hdr_len;
    if ( sync_audio(&ap) == NULL )
    { 
      ap.hdr_len    = 2;
      ap.id         = AUDIO_ID_MP2;
      ap.samples_pf = MP2_FRAME_SIZE;
      
      ap.block_pbuf = ap.chunk_buf;
      ap.block_end  = ap.chunk_buf+n-ap.hdr_len;
      if ( sync_audio(&ap) == NULL ) ap.eof = 1; 
    }  
  }
  else
    ap.eof = 1;
    
  rewind_audio(fl);

  if (!ap.eof)
  {
    ap.size_pf = (int)( (ap.samples_pf * ap.br) / (8.0 * ap.sr) + 0.5);
    
    
    fprintf(stderr, "\naudio stream type (%s) detected \n", (ap.id==AUDIO_ID_MP2)? "MP2":"AC3");
    fprintf(stderr, "- sample rate (%d Hz)\n- bitrate (%d bps)\n- frame size (%d bytes)\n",
            ap.sr, ap.br, ap.size_pf);
  
    return &ap;
  }
  
  fprintf(stderr, "\nERROR: no valid audio stream\n");
  return NULL;
}

//== audio parser (read chunk data for exactly 1 audio frame) ==
//==============================================================
uint8_t *parse_audio(AudioParser *ap)
{
  uint8_t *next = NULL;

  int available     = 0;
  int s, n          = 0;
  int done          = 0;
  int unused        = ap->unused; 
  uint8_t *pbuf     = ap->chunk_buf;
  uint8_t *pbuf_end = pbuf;
  uint8_t *start    = NULL;
  uint8_t *end      = NULL;

  if (unused) 
  {  
    memmove(pbuf, pbuf+ap->used, unused);
    ap->used   = 0;
    ap->unused = 0;
  }
  
  ap->frame_len = 0;
  do
  {
    if ( (CHUNK_BLOCK_LEN - unused) > 0)
    {
      s = read_audio(pbuf+unused, 1, CHUNK_BLOCK_LEN - unused, ap->fl); 
      n = s + unused;
    }
    else
      n = unused;
    unused = 0;  
    
    available += n;
    pbuf_end  += n;
  
    if (!start)
    {
      ap->block_pbuf = pbuf;
      ap->block_end  = pbuf_end - ap->hdr_len;
      if ( (next = sync_audio(ap)) != NULL )
      {
        end   = next + ap->size_pf; //ap->hdr_len; TODO maybe bad for corrupted audio ?
        start = next;
      }
      else
        unused = ap->hdr_len;
    }
  
    if (end)
    {
      ap->block_end = pbuf_end - ap->hdr_len;
      do
      {
        ap->block_pbuf = end;
        if ( (next = sync_audio(ap)) != NULL ) 
        {
          ap->frame_len = next - start; 
          if (ap->frame_len < ap->size_pf)
          {
            end = next + ap->hdr_len;
          }
          else
          {
            done = 1;
            break;
          }
        }  
      } while (next);
    
      //if (done) break;
      
      end    = pbuf_end;
      unused = ap->hdr_len; 
    }

    pbuf = pbuf_end;  
  
  } while (s && !done && (available >= CHUNK_BLOCK_LEN) );  

  if (ap->frame_len)  
  {
    ap->used   = start - ap->chunk_buf + ap->frame_len;
    ap->unused = available - ap->used;
  }
  else
  { 
    //-- prevent segfaults -- 
    if (available > ap->size_pf) available = ap->size_pf;
    
    ap->frame_len = available;
    ap->eof       = 1;    
  }
  
  return start;
}

//=============
//== M A I N ==
//=============

int main(int argc, char *argv[])
{
  static uint8_t  a_buf[8*1024];     // buffer to collect audio frames
  
  int  rmode  = IS_FILE_MODE; // start in file mode (default) 
  int  cnt    = 0;
  int  i, rc  = 0;
  int  error  = 0;
   
  int  verbose = 0;
  char eol[3]  = " \r"; 
   
  uint8_t *psrc, *pdst;

  char *vfn = NULL;	      // i/o filenames
  char *afn = NULL;
  char *ofn = NULL;
  
  FifoList *vfl = NULL;	      // there are 2 pointers to the same object !
  FifoList *afl = NULL;
  
  VideoParser *vp;            // frame reader context
  AudioParser *ap;

  PackStream  *pack_stream;    
    
  int     a_shift        = 0;
  int     bytes_read     = 0;     // control counters
  double  bytes_needed   = 0.0;
  int32_t total_used;
  double  total_needed;
  double  samples_needed = 0.0;
  int     forceRestamp   = 0;
  int     muxType        = MX_TYPE_PS;
  
  //-- parse arguments --
  //---------------------
  for (i=1; i<argc; i++)
  {
    if (!strcmp(argv[i], "-1"))
      verbose = 1;
    else if (!strcmp(argv[i], "-2"))
    {
      verbose = 2;
      strcpy(eol,"\r\n");
    }  
    else if (!strcmp(argv[i], "-p"))
      rmode = IS_PIPE_MODE;
    else if (!strcmp(argv[i], "-pes"))
    {
      fprintf(stderr, "INFO: \'PES\' multiplex activated\n");
      muxType = MX_TYPE_PES;
    } 
    else if (!strcmp(argv[i], "-ts"))
    {
      fprintf(stderr, "INFO: \'TS\' multiplex activated\n");
      muxType   = MX_TYPE_TS;
    } 
    else if (!strcmp(argv[i], "-v"))
    {
      if (++i < argc) vfn = argv[i];
    }
    else if (!strcmp(argv[i], "-a"))
    {
      if (++i < argc) afn = argv[i];
    }
    else if (!strcmp(argv[i], "-o"))
    {
      if (++i < argc) ofn = argv[i];
    }
    else if (!strcmp(argv[i], "-pd"))
    {
    	pulldown=1;
    	fprintf(stderr, "INFO: Pulldown activated\n");
    }
    else if (!strcmp(argv[i], "-sh"))
    {
      if (++i < argc) a_shift = atoi(argv[i]);
    }
    else if (!strcmp(argv[i], "-r"))
    {
      forceRestamp = 1;
      fprintf(stderr, "INFO: GOP timestamps will be rebuild\n");
    }
  }
  
  if (!vfn || !afn || !ofn) 
  {
    fprintf(stderr, "usage: lvemux [options] -v <video stream> -a <audio stream> -o <output>\n"
                    "options: \"-p\" will switch to pipe mode (default is file mode)\n"
                    "         \"-pd\" must be set for 3:2 pulldown video\n"
                    "         \"-r\" to force rebuild of GOP timestamps\n"
                    "         \"-pes|-ts\" to force PES or TS packet muxing (new)\n"
                    "         \"-1|-2\" verbosity mode to print progress, how line ends\n"
                    "         with '\\r' for \"-1\" or '\\n' for \"-2\"\n" 
                    "         \"-sh <[+/-]time>\" shifts audio +/- <time> milliseconds\n"
                    "         against video.\n"
                    "         using a \"-\" char as filename redirects output to stdout\n");
    
    exit(0);
  }
  
  //-- get input streams into account --
  //------------------------------------
  vfl = open_video(vfn, rmode);
  afl = open_audio(afn, rmode);
  
  //-- check for errors --
  if (!vfl || !afl) 
  {
    fprintf(stderr, "ERROR: opening A/V streams (%d/%d)\n", 
           (afl!=NULL), (vfl!=NULL)); 
    error = 1;
  }

  if (!error)
  {
    //-- let main thread reach wait state !!! --
    usleep(100000); 
  
    //-- init A/V stream parser --
    //----------------------------
    vp = init_video_parser(vfl);
    ap = init_audio_parser(afl);
  
    //-- check for errors --  
    if (!vp || !ap)
    {
      fprintf(stderr, "ERROR: creating A/V parser (%d/%d)\n", 
           (ap!=NULL), (ap!=NULL)); 
      error = 1;
    }
  }      
  
  if (!error)  
  {
    //-- open output file including muxer initialisation. --
    //-- video bitrate set to 0 forces DVD fixed muxrate. --
    //------------------------------------------------------
    pack_stream  = mux_open(ofn, 0, vp->fr, ap->br, ap->sr, 
                            (ap->id==AUDIO_ID_MP2)? AUDIO_ID_MP2:AUDIO_ID_AC3 );
                         
    pack_stream->audio_encoded_fs = ap->size_pf;                     
    pack_stream->forceRestamp     = forceRestamp;           
    pack_stream->mux_type         = muxType;
 
    //-- any sync correction needed ? --                       
    //----------------------------------
    if (a_shift != 0)
    {
      int d;
      
      pack_stream->a_pts_ofs = A_PTS_MIN;
      pack_stream->v_pts_ofs = A_PTS_MIN + (V_PTS_DFLT-A_PTS_DFLT) - a_shift;

      d = pack_stream->v_pts_ofs - V_PTS_MIN;
      
      if (d < pack_stream->audio_delay)
      {
        pack_stream->v_pts_ofs  = V_PTS_MIN;
        pack_stream->a_pts_ofs -= d;
      }
      else 
      {
        pack_stream->v_pts_ofs -= pack_stream->audio_delay;
        pack_stream->a_pts_ofs -= pack_stream->audio_delay;
      }
    
      fprintf(stderr, "\naudio shift (%d) ms, A/V start (%d/%d) ms\n", 
              a_shift, pack_stream->a_pts_ofs + pack_stream->audio_delay,
              pack_stream->v_pts_ofs);
    }
  
    //-- calc needed audio samples corresponding to duration of one    --
    //-- video frame and the appropriate number of encoded audio bytes --
    //-------------------------------------------------------------------
    samples_needed = ap->sr / vp->fr + 0.5;
    bytes_needed   = samples_needed * ap->size_pf / ap->samples_pf;
  
    fprintf(stderr, "\n");
  
    //-- lets go --
    //-------------
    total_used   = 0;
    total_needed = 0.0;
    if (pack_stream)
    {
      do
      {
        //-- read one complete video frame or  --
        //-- 2 fields (if interlaced video)and --
        //-- deliver this data to the muxer.   --
        //---------------------------------------
        pdst = parse_video(vp);
        if (pdst && vp->frame_len) 
        {
          cnt++;
          mux_write_packet(pack_stream, vp->id, pdst, vp->frame_len);
        }
 
        //-- switch to audio (needs extra collect buffer) --
        //--------------------------------------------------
        pdst = a_buf;

        //-- read complete(!) audio frames at least "total_needed" is reached --
        //----------------------------------------------------------------------
        total_needed += bytes_needed;
        bytes_read    = 0;
        while (total_used < total_needed)
        {
          psrc = parse_audio(ap);
          if (psrc)
          { 
            int l = ap->frame_len;
          
            //-- collect audio frames --
            memcpy(pdst, psrc, l);
            pdst         += l;
            total_used   += l;
            bytes_read   += l;
          }
          else
            break;
        }
      
        //-- deliver only audio portion matching --
        //-- to duration of one video frame.     --
        //----------------------------------------- 
        mux_write_packet(pack_stream, ap->id, a_buf, bytes_read);
      
        if (verbose) fprintf(stderr, "(%d)%s", cnt, eol);
      
      } while (!vp->eof && !ap->eof);
    
      fprintf(stderr, "(%d) frames processed - done !\n", cnt);
    
      mux_close(pack_stream); // commit output
    }
  }
  
  if (vfl) close_video(vfl);
  if (afl) close_audio(afl);
  
  return 0;  
} 
//#eof
