/*
 *  scenelst.c
 *
 *  Copyright (C) Gerhard Monzel - October 2003
 *
 *  This file is part of lve, a linux mpeg video editor
 *      
 *  lve tools are free software; you can redistribute it and/or modify
 *  it under the terms of the GNU General Public License as published by
 *  the Free Software Foundation; either version 2, or (at your option)
 *  any later version.
 *   
 *  lve is distributed in the hope that it will be useful,
 *  but WITHOUT ANY WARRANTY; without even the implied warranty of
 *  MERCHANTABILITY or FITNESS FOR A PARTICULAR PURPOSE.  See the
 *  GNU General Public License for more details.
 *   
 *  You should have received a copy of the GNU General Public License
 *  along with GNU Make; see the file COPYING.  If not, write to
 *  the Free Software Foundation, 675 Mass Ave, Cambridge, MA 02139, USA. 
 *
 */
#include "scenelst.h"
#include "yuv2rgb16.h"
#include "display.h"

T_MB_SCENE::T_MB_SCENE(T_LVE_CTX *ctx, int num)  : T_MEMBER(ctx, num)
{
  mdf = NULL;
}

T_MB_SCENE::~T_MB_SCENE()
{
  if (mdf) TMediaFile::releaseMDF(mdf);
}

  
static char *timecode(char *c_str, int n, double f = 25.0)
{
  int    h, m, s;
  
  f = f * 3600;
  
  //-- calculate hours ... --  
  h  = (int)(n/f);
  n -= (uint32_t)(h*f);
  f /= 60;
  
  //-- ... minutes ... --
  m  = (int)(n/f);
  n -= (uint32_t)(m*f);
  f /= 60;
  
  //-- ... seconds and pictures rests --
  s  = (int)(n/f);
  n -= (uint32_t)(s*f);

  sprintf(c_str, "%02d:%02d:%02d.%02d", h, m, s, n);

  return c_str;
}

int T_MB_SCENE::GetOutputSize(void)
{
  T_FRAME_SRV *frmSrv = (T_FRAME_SRV *)mdf->frmSrv;
  return (frmSrv->filePos(end_real) - frmSrv->filePos(start_real));
}

void T_MB_SCENE::ShowInfo(T_LST_SCENE *lst)
{
  char t[81];
  char t1[16];
  
  sprintf (t, "media[%d] %s", GetMDFId(), GetMDFName());
  display_OvText(ctx, t, 0x05);
  display_OvText(ctx, " ", 0x00);

  sprintf(t, "  duration  %s [%d]", timecode(t1, dist), dist);
  display_OvText(ctx, t, 0x00);
  if (lst) 
  {
    sprintf(t, "  of total  %s [%d], (%d MB)", 
            timecode(t1, lst->FramesCnt()), lst->FramesCnt(),
            lst->GetOutputSize()/(1024*1024));
    display_OvText(ctx, t, 0x00);
  }
  display_OvText(ctx, " ", 0x00);

  sprintf(t, "     start  %s [%d]", timecode(t1, start_real+1), start_real); 
  display_OvText(ctx, t, 0x00);

  sprintf(t, "       end  %s [%d]   (in media)", timecode(t1, end_real+1), end_real);
  display_OvText(ctx, t, 0x00);
  display_OvText(ctx, " ", 0x00);

  sprintf(t, "     start  %s [%d]", timecode(t1, start+1), start); 
  display_OvText(ctx, t, 0x00);

  sprintf(t, "       end  %s [%d]   (in list)", timecode(t1, end+1), end);  
  display_OvText(ctx, t, 0x00);
  display_OvText(ctx, " ", 0x00);

  sprintf(t, "  Act. pos  %s [%d] (in scene)", timecode(t1, ctx->act_frame-start+1), ctx->act_frame-start);  
  display_OvText(ctx, t, 0x00);

  sprintf(t, "            %s [%d] (at all)", timecode(t1, ctx->act_frame+1), ctx->act_frame);  
  display_OvText(ctx, t, 0x00);
  
  sprintf(t, "            scene size (%d KB)", GetOutputSize()/1024);  
  display_OvText(ctx, t, 0x02);
}

int T_MB_SCENE::last_mdf_id = -1;


int T_MB_SCENE::Setup(TMediaFile *_mdf, int _astream, int _vstream) 
{
  int sav_as = ctx->a_stream;
  int sav_vs = ctx->v_stream;
  mdf        = _mdf;
  
  astream = _astream;
  vstream = _vstream;
   
  ctx->a_stream = astream;
  ctx->v_stream = vstream;

  if (!lve_load_mdf(ctx, mdf)) 
  {
    ctx->a_stream = sav_as;
    ctx->v_stream = sav_vs;
    return 0; 
  }
  
  //fprintf(stderr, "S: (%d/%d)\n", GetMDFId(), last_mdf_id);
  
  dist        = ((T_FRAME_SRV *)mdf->frmSrv)->nFrames() - 3;  
  start_real  = 0;
  end_real    = dist - 1;
  avs_dp      = 0;
  //last_mdf_id = GetMDFId();
  
  return 1; 
}


T_MEMBER *T_MB_SCENE::OnSelect(void)
{
  char avs_str[16];
  
  //-- reminder for video action --
  int action_sav = ctx->action;
  int iv_sav     = ctx->iv;

  //-- media file must be valid everytime --
  if (mdf == NULL)
  {
    fprintf(stderr, "ERROR: T_MB_SCENE::OnSelect() - no valid media file\n");
    return (T_MEMBER *)this;
  }  

  //fprintf(stderr, "O: (%d/%d)\n", GetMDFId(), last_mdf_id);

  //-- detect media-change --
  if (GetMDFId() != last_mdf_id)
  {
    ctx->action = A_NOMEDIA;
    ctx->iv     = IV_START;
  }      
  last_mdf_id = GetMDFId();

  //-- if new media needed ... --    
  if (ctx->action == A_NOMEDIA)
  { 
    //-- ...get media into access --
    if ( !lve_load_mdf(ctx, mdf) ) return NULL; 

    //-- request to adjust video overlay --
    display_reset_ov(ctx);
    
    //-- restore video action --
    ctx->action = action_sav;
    ctx->iv     = iv_sav;
  }
  
  ctx->file = (T_FRAME_SRV *)mdf->frmSrv;  // !!!

  switch (ctx->adj_pos)
  {
    //-- seek to start of scene --
    case 3:
      ctx->act_frame = start;
      ctx->act_real  = start_real;
      ctx->adj_pos   = 1;
      break;
      
    //-- no seek needed --  
    case 2:
      ctx->adj_pos = 0;
      break;
      
    //-- seek to new "act_frame" --  
    default:
      ctx->adj_pos = 1;
  }
    
  //-- if actual frame doesn't belong to this    --
  //-- scene, set it to start-frame of this scene --   
  if ( (ctx->act_frame < start) || (ctx->act_frame > end) )
  {
    ctx->act_frame = start;
  }  
  
  //-- adjust scene boundaries in video structure --
  ctx->start      = start;
  ctx->end        = end;
  ctx->start_real = start_real;
  ctx->end_real   = end_real; 

  //-- set stream numbers --
  ctx->v_stream = vstream;
  //ctx->a_stream = astream;
  astream = ctx->a_stream;      // new behavier !!!

  //-- new --
  ctx->avs_dp = avs_dp;
  lve_set_avs(avs_dp);
  
  //-- on first load -> automatic switch to stop action --
  if (ctx->action == A_NOMEDIA)
  {
    ctx->iv       = IV_NORM;
    ctx->action   = A_STOP;
  }
  return (T_MEMBER *)this;
}

int T_MB_SCENE::GetRealFrame(void)
{
  int p = (ctx->act_frame - start) + start_real;
  return p;         
}

int T_MB_SCENE::GetStartFrame() 
{ 
  return start; 
}

int T_MB_SCENE::GetEndFrame() 
{ 
  return end; 
}

void T_MB_SCENE::SetStartFrame(int p) 
{ 
  start = p; 
};

void T_MB_SCENE::SetEndFrame(int p) 
{ 
  end = p; 
}

int T_MB_SCENE::FramesCnt() 
{ 
  return dist; 
}
    
int T_MB_SCENE::GetMDFId() 
{ 
  return mdf->getId(); 
}
    
char *T_MB_SCENE::GetMDFName() 
{ 
  return mdf->getName(); 
}

T_MB_SCENE *T_MB_SCENE::Dup(void) 
{ 
  return NULL; 
}


static Uint8 page0[BMP_TH_W*BMP_TH_H];
static Uint8 page1[BMP_TH_W2*BMP_TH_H2];
static Uint8 page2[BMP_TH_W2*BMP_TH_H2];

void T_MB_SCENE::GenThumb(void)
{
  int   x, y, i, dx, dy, dx2, dy2;
  int   ofs_y, ofs_y2, ofs_x, ofs_x2;
  Uint8 *src, *dst;
  Uint8 t, t1, t2;

  switch (ctx->ov_fmt)
  {
    //-- YUV planar --
    //----------------
    case IMGFMT_YV12:
    case IMGFMT_I420:
    case IMGFMT_IYUV:

      //-- Y --
      dx = ctx->w / BMP_TH_W;
      dy = ctx->h / BMP_TH_H;
      dx2 = dx>>1;
      dy2 = dy>>1;
    
      //-- GMO: attention!!! on interlaced video --
      //-- we get only one field (we use +1) !!! --
      ofs_y  = (ctx->h - (dy * BMP_TH_H))/2 + 1;   
      ofs_y2 = (ofs_y>>1) + 1;
  
      ofs_x  = (ctx->w - (dx * BMP_TH_W))/2;
      ofs_x2 = ofs_x>>1; 
    
      dst = page0;
      src = ctx->page[0] + (ctx->w * ofs_y) + ofs_x;

      for (y=0; y<BMP_TH_H; y++)
      {
        i = 0;
        for (x=0; x<BMP_TH_W; x++)
        {
          dst[x] = src[i];
          i += dx;    
        }
        dst += BMP_TH_W;
        src += (dy * ctx->w);
      }

      //-- U --
      dst = page1;
      src = ctx->page[2] + (int)( (ctx->w/2 * ofs_y2) + ofs_x2 );
  
      for (y=0; y<BMP_TH_H2; y++)
      {
        i = 0;
        for (x=0; x<BMP_TH_W2; x++)
        {
          dst[x] = src[i];
          i += dx;    
        }
        dst += BMP_TH_W2;
        src += (dy2 * ctx->w);
      }
  
      //-- V --
      dst = page2;
      src = ctx->page[1] + (int)( (ctx->w/2 * ofs_y2) + ofs_x2 );
  
      for (y=0; y<BMP_TH_H2; y++)
      {
        i = 0;
        for (x=0; x<BMP_TH_W2; x++)
        {
          dst[x] = src[i];
          i += dx;    
        }
        dst += BMP_TH_W2;
        src += (dy2 * ctx->w);
      }

      yuv2rgb16(page0, page1, page2, bmp_thumb, BMP_TH_W, BMP_TH_H);
      *bmp_thumb=127;
      break;
      
    //-- YUV packed, flipped --
    //-------------------------
    case IMGFMT_YUY2:
    case IMGFMT_UYVY:
    case IMGFMT_YVYU:  
      fprintf(stderr, "WARNING: format not supported for thumbnails\n"); 
      break;
  }  
}

void T_MB_SCENE::BlitThumb(SDL_Rect *dst_rect)
{
  if (*bmp_thumb != 0)
  {
    SDL_Surface *s;
    char        id_str[8];
   
    s = SDL_CreateRGBSurfaceFrom(bmp_thumb, 
                                 BMP_TH_W, BMP_TH_H, 
                                 16, BMP_TH_W*2, 
                                 0, 0, 0, 0);
    if (s)
    { 
      SDL_Rect dr;
    
      dr.x = dst_rect->x + ((dst_rect->w - BMP_TH_W)>>1);
      dr.y = dst_rect->y + ((dst_rect->h - BMP_TH_H)>>1); 

      SDL_BlitSurface(s, NULL, ctx->screen, &dr);   
      SDL_FreeSurface(s);
    }
    
    sprintf(id_str, "%d", GetMDFId());
    DT_DrawText(id_str, ctx->screen, 0, 
                dst_rect->x+dst_rect->w/2-8, 
                dst_rect->y+dst_rect->h-18, 0);
  }  
}

//==========================

void T_LST_SCENE::resetLastMDFId(void)
{
  T_MB_SCENE::last_mdf_id = -1;
}

T_MEMBER *T_LST_SCENE::InsertMember(T_MEMBER *mb_new, T_MEMBER *mb)
{
  T_MB_SCENE *mb1 = (T_MB_SCENE *)mb;
  T_MB_SCENE *mb_new1;

  if ( (mb_new1 = (T_MB_SCENE *)T_LIST::InsertMember(mb_new, mb)) != NULL )
    ctx->dist_sum = dist_sum;
  return mb_new1; 
}


T_MEMBER *T_LST_SCENE::AppendMember(T_MEMBER *mb_new)
{ 
  T_MB_SCENE *mb      = NULL;
  T_MB_SCENE *mb_new1 = (T_MB_SCENE *)mb_new;
  
  if ( (mb = (T_MB_SCENE *)T_LIST::AppendMember(mb_new)) != NULL)
  {
    //-- adjust frames counts of list and new member --
    //-- further adjust start/end pos of new member  --
    mb_new1->start = dist_sum;    
    dist_sum += mb_new1->dist;
    mb_new1->end = dist_sum - 1;
  
    ctx->dist_sum = dist_sum;
  }
  return mb_new;
}    

void T_LST_SCENE::UpdateMember(T_MEMBER *mb, int all)
{
  T_MB_SCENE *mb1 = (T_MB_SCENE *)mb;
  int d, id;
  
  if (!mb1) mb1 = (T_MB_SCENE *)GetFocusMember();
  if (!mb1) return;

  id = mb1->GetMDFId();
  d  = ctx->avs_dp - mb1->avs_dp;
  
  //-- process only all marked with same ID --
  //------------------------------------------ 
  do
  {
    if (id != mb1->GetMDFId()) break;
    if (mb1->HasMark()) mb1->avs_dp += d;
  } while ( (mb1 = (T_MB_SCENE *)mb1->GetNext()) != NULL );
}
    
T_MEMBER *T_LST_SCENE::OnSelectMember(int num)
{
  return (T_LIST::OnSelectMember(num));
}

off_t T_LST_SCENE::GetOutputSize(void)
{
  T_MB_SCENE *mb  = (T_MB_SCENE *)mb_first;
  off_t      size = 0;  
    
  while (mb) 
  {  
    size += mb->GetOutputSize();
    mb = (T_MB_SCENE *)mb->GetNext();
  } 
  
  return size; 
}
